@@ -5,8 +5,8 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE            @(#) %derived_by: guidod %
- *  @version %version: bln_mpt1!5.29 %
- *    (%date_modified: Tue Oct 09 11:28:07 2001 %)
+ *  @version %version: bln_mpt1!5.36 %
+ *    (%date_modified: Wed Jul 17 18:02:23 2002 %)
  *  @description
  *         Subroutines for the Forth Core System - especially the
  *         general input/output routines like ACCEPT/QUERY/WORD/PARSE 
@@ -15,7 +15,7 @@
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  core-sub.c~bln_mpt1!5.29:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  core-sub.c~bln_mpt1!5.36:csrc:bln_12xx!1 % $";
 #endif
 
 #define _P4_SOURCE 1
@@ -49,6 +49,8 @@ static char* id __attribute__((unused)) =
 
 #include <pfe/logging.h>
 
+/***********************************************************************/
+
 /* removed this one from general def-types */
 #define SPAN    p4_SPAN
 
@@ -161,7 +163,10 @@ _export void
 p4_lower (char *p, int n)
 {
     while (--n >= 0)
-        *p = tolower (*p), p++;
+    {
+        *p = tolower (*p);
+        p++;
+    }
 }
 
 /** _upper_ ( str* str# -- )
@@ -172,7 +177,10 @@ _export void
 p4_upper (char *p, int n)
 {
     while (--n >= 0)
-        *p = toupper (*p), p++;
+    {
+        *p = toupper (*p); 
+        p++;
+    }
 }
 
 /** _zplaced_ ( str* str# dst* max# -- dst* ) [alias] _store_c_string_
@@ -182,6 +190,7 @@ p4_upper (char *p, int n)
 _export char *
 p4_store_c_string (const char *src, int n, char *dst, int max)
 {
+    /* RENAME: p4_zplaced */
     if (n >= max)
         n = max - 1;
     memcpy (dst, src, n);
@@ -196,6 +205,7 @@ p4_store_c_string (const char *src, int n, char *dst, int max)
 _export char*
 p4_pocket_c_string (const char* src, int n)
 {
+    /* RENAME: p4_pocket_zplace */ /* REQUIRE: p4_pocket_place */
     return p4_store_c_string (src, n, p4_pocket (), P4_POCKET_SIZE);
 }
 
@@ -207,9 +217,12 @@ p4_pocket_c_string (const char* src, int n)
 _export char* 
 p4_store_filename (const char *src, int n, char* dst, int max)
 {
+    /* RENAME: p4_zplace_filename */
     int s = 0;
     int d;
     char* p;
+
+    if (!src || !n) { *dst = '\0'; return dst; }
     
 #  if PFE_DIR_DELIMITER == '\\'
 #   define PFE_ANTI_DELIMITER '/'
@@ -262,6 +275,7 @@ p4_store_filename (const char *src, int n, char* dst, int max)
 _export char*
 p4_pocket_filename (const char* src, int n)
 {
+    /* RENAME: p4_pocket_zplace_filename */
     return p4_store_filename (src, n, p4_pocket (), P4_POCKET_SIZE);
 }
 
@@ -427,26 +441,27 @@ do_match (const short *pattern, const char *string, int ic)
 
     for (;;)
     {
+        --ic;
         switch (c = *pattern++)
         {
          case '\0':
              return *string == '\0';
          case -'*':
              while (*string && !do_match (pattern, string, ic))
-                 string++;
+             { --ic; string++; }
              continue;
          case -'?':
              if (*string++)
                  continue;
              return 0;
          default:
-             if (!ic) 
+             if (ic < 0) 
              {
                  if (*string++ == c)
                      continue;
              }else{
-                 if (tolower(*string++) == tolower(c))
-                     continue;
+                 if (*string == c || *string == toupper(c))
+                 { string++; continue; }
              }
              return 0;
         }
@@ -461,6 +476,7 @@ do_match (const short *pattern, const char *string, int ic)
 _export int
 p4_match (const char *pattern, const char *string, int ic)
 {
+    /* RENAME: p4_wild_match - move near p4_wild_words - possibly export */
     short buf[0x100], *p = buf;
 
     /* preprocess pattern, remove `\' */
@@ -493,6 +509,7 @@ p4_match (const char *pattern, const char *string, int ic)
         break;
     }
     /* match with preprocessed pattern */
+    if (ic) ic = 31;
     return do_match (buf, string, ic);
 }
 
@@ -581,7 +598,7 @@ p4_dig2num (p4char c, p4ucell *n, p4ucell base)
         c -= '0';
     else
     {
-        if (LOWER_CASE)
+        if (UPPER_CASE)
 	c = toupper (c);
         if (c < 'A')
             return P4_FALSE;
@@ -589,7 +606,7 @@ p4_dig2num (p4char c, p4ucell *n, p4ucell base)
             c -= 'A' - ('9' - '0' + 1);
         else
 	{
-            if (LOWER_CASE || c < 'a')
+            if (UPPER_CASE || c < 'a')
                 return P4_FALSE;
             c -= 'a' - ('9' - '0' + 1) - ('Z' - 'A' + 1);
 	}
@@ -914,6 +931,7 @@ p4_type (const char *s, p4cell n)
 _export void
 p4_type_on_line (const char *s, p4cell n)
 {
+    /* RENAME: ... might need p4_Q_cr variant... make macro from this? */
     if (p4_OUT + n >= p4_COLS)
         FX (p4_cr);
     p4_type (s, n);
@@ -1111,8 +1129,9 @@ FCode (p4_query)
     TO_IN = 0;
     TIB = PFE.tib;
     NUMBER_TIB = p4_accept (TIB, TIB_SIZE);
-    if (PFE.query_hook) 
-        NUMBER_TIB = (*PFE.query_hook)(NUMBER_TIB); 
+    /* if (PFE.query_hook) // please use lined.h:lined->intercept now 
+     *     NUMBER_TIB = (*PFE.query_hook)(NUMBER_TIB); 
+     */
     SPAN = NUMBER_TIB;
 }
 
@@ -1166,6 +1185,28 @@ p4_source (char **p, int *n)
     }
 }
 
+/** _size_saved_input_ ( -- iframe-size )
+ */
+_export p4ucell
+p4_size_saved_input (void)
+{
+    return sizeof (Iframe);
+}
+
+/** _link_saved_input_ ( iframe* -- )
+ * see => SAVE-INPUT
+ */
+_export void
+p4_link_saved_input (void *p)
+{
+    Iframe *iframe = (Iframe *) p;
+    
+    iframe->magic = P4_INPUT_MAGIC;
+    iframe->input = PFE.input;
+    iframe->prev = PFE.saved_input;
+    PFE.saved_input = iframe;
+}
+
 /** _save_input_ ( iframe-stack* -- iframe-stack*' )
  * see => SAVE-INPUT
  */
@@ -1173,16 +1214,25 @@ _export void *
 p4_save_input (void *p)
 {
     Iframe *iframe = (Iframe *) p;
-    
     --iframe;
-    iframe->magic = P4_INPUT_MAGIC;
-    iframe->input = PFE.input;
-    iframe->prev = PFE.saved_input;
-    PFE.saved_input = iframe;
-    
+    p4_link_saved_input (iframe);
     return ((void*) iframe);
 }
 
+/** _unlink_saved_input_ ( iframe* -- )
+ * see => RESTORE-INPUT
+ */
+_export void
+p4_unlink_saved_input (void *p)
+{
+    Iframe *iframe = (Iframe *) p;
+
+    if (iframe->magic != P4_INPUT_MAGIC)
+        p4_throw (P4_ON_ARG_TYPE);
+    PFE.input = iframe->input;
+    PFE.saved_input = iframe->prev;
+}
+
 /** _restore_input_ ( iframe-stack* -- iframe-stack*' )
  * see => RESTORE-INPUT
  */
@@ -1190,16 +1240,11 @@ _export void *
 p4_restore_input (void *p)
 {
     Iframe *iframe = (Iframe *) p;
-
-    if (iframe->magic != P4_INPUT_MAGIC)
-        p4_throw (P4_ON_ARG_TYPE);
-    PFE.input = iframe->input;
-    PFE.saved_input = iframe->prev;
+    p4_unlink_saved_input (p);
     ++iframe;
-
-    p = (void *) iframe;
-    return p;
+    return ((void *) iframe);
 }
+
 /** _refill_ ( -- flag )
  * see => REFILL
  */
