@@ -5,8 +5,8 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE             @(#) %derived_by: guidod %
- *  @version %version: bln_mpt1!5.21 %
- *    (%date_modified: Tue Oct 02 12:43:09 2001 %)
+ *  @version %version: bln_mpt1!5.36 %
+ *    (%date_modified: Tue Jun 18 14:49:54 2002 %)
  *
  *  @description
  *	The Portable Forth Environment provides a decompiler for
@@ -96,7 +96,6 @@ static char* id __attribute__((unused)) =
 #include <ctype.h>
 #include <string.h>
 
-
 #include <pfe/_missing.h>
 /************************************************************************/
 /* decompiler                                                           */
@@ -142,9 +141,15 @@ p4_literal_SEE (p4xt* ip, char* p, p4_Semant* s)
 {
     char buf[80];
     if (s) 
-        sprintf (p, "( %.*s) 0x%p ", NFACNT(*s->name), s->name+1, *(void**)ip);
-    else
+    {
+        if (s->name && ! memcmp (s->name+1, "LITERAL", 7)) /* 'bit fuzzy... */
+            sprintf (p, "0x%X ", *(p4cell*)ip);
+        else
+            sprintf (p, "( %.*s) 0x%X ", 
+              NFACNT(*s->name), s->name+1, *(p4cell*)ip);
+    }else{
         strcpy (p, p4_str_dot (*(p4cell *) ip, buf + sizeof buf, BASE));
+    }
     return ++ip;
 }
 
@@ -153,7 +158,7 @@ p4_lit_to_token_SEE (p4xt* ip, char* p, p4_Semant* s)
 {
     register p4char* nfa;
     register p4xt xt = ip[-1];
-    if (*xt == s->exec[0])
+    if (*P4_TO_CODE(xt) == s->exec[0])
     {
         xt = *ip++;
         nfa = p4_to_name (xt);
@@ -223,11 +228,11 @@ p4_decompile_word (p4xt* ip, char *p, p4_Decomp *d)
     memcpy (d, ((s) ? (& s->decomp) : (& default_style)), sizeof(*d));
 
     /* some tokens are (still) compiled without a semant-definition */
-    if (*xt == PFX (p4_literal_execution))
+    if (*P4_TO_CODE(xt) == PFX (p4_literal_execution))
         return p4_literal_SEE (ip, p, s);
-    if (*xt == PFX (p4_locals_bar_execution))
+    if (*P4_TO_CODE(xt) == PFX (p4_locals_bar_execution))
         return p4_locals_bar_SEE (ip, p, s);
-    if (*xt == PFX (p4_local_execution))
+    if (*P4_TO_CODE(xt) == PFX (p4_local_execution))
         return p4_local_SEE (ip, p, s);
 
     if (d->skips == P4_SKIPS_CELL 
@@ -305,7 +310,7 @@ p4_decompile_rest (p4xt *ip, int nl, int indent)
     }
 }
 
-static P4_CODE_RUN(p4_var_RT_SEE)
+static P4_CODE_RUN(p4_variable_RT_SEE)
 {
     strcat (p, "VARIABLE ");
     strncat (p, nfa+1, NFACNT(*nfa));
@@ -380,7 +385,7 @@ static P4_CODE_RUN(p4_colon_RT_SEE)
     return (p4xt*) p4_to_body (xt);
 }
 
-static P4_CODE_RUN(p4_does_defined_RT_SEE)
+static P4_CODE_RUN(p4_does_RT_SEE)
 {
     strcat (p, "<BUILDS ");
     strncat (p, nfa+1, NFACNT(*nfa));
@@ -396,22 +401,33 @@ p4_decompile (char *nfa, p4xt xt)
     *buf = '\0';
 
     FX (p4_cr);
-    if (*xt == p4_var_RT_) p4_var_RT_SEE (buf, xt, nfa);
-    else if (*xt == p4_builds_RT_) p4_builds_RT_SEE (buf, xt, nfa);
-    else if (*xt == p4_constant_RT_) p4_constant_RT_SEE (buf, xt, nfa);
-    else if (*xt == p4_value_RT_) p4_value_RT_SEE (buf, xt, nfa);
-    else if (*xt == p4_two_constant_RT_) p4_two_constant_RT_SEE (buf, xt, nfa);
+    if (*P4_TO_CODE(xt) == p4_variable_RT_) 
+	p4_variable_RT_SEE (buf, xt, nfa);
+    else if (*P4_TO_CODE(xt) == p4_builds_RT_) 
+	p4_builds_RT_SEE (buf, xt, nfa);
+    else if (*P4_TO_CODE(xt) == p4_constant_RT_) 
+	p4_constant_RT_SEE (buf, xt, nfa);
+    else if (*P4_TO_CODE(xt) == p4_value_RT_) 
+	p4_value_RT_SEE (buf, xt, nfa);
+    else if (*P4_TO_CODE(xt) == p4_two_constant_RT_) 
+	p4_two_constant_RT_SEE (buf, xt, nfa);
     else if (PFE.decompile[0] && PFE.decompile[0](nfa,xt)) { /* (unused) */ }
     else if (PFE.decompile[1] && PFE.decompile[1](nfa,xt)) { /* (unused) */ }
     else if (PFE.decompile[2] && PFE.decompile[2](nfa,xt)) { /* floating */ }
-    else if (*xt == p4_marker_RT_) p4_marker_RT_SEE (buf, xt, nfa);
-    else if (*xt == p4_defer_RT_ ) p4_defer_RT_SEE (buf, xt, nfa);
-    else if (*xt == p4_offset_RT_) p4_offset_RT_SEE (buf, xt, nfa);
-    else if (*xt == p4_vocabulary_RT_) p4_vocabulary_RT_SEE (buf, xt, nfa);
-    else if (*xt == p4_colon_RT_ || *xt == p4_debug_colon_RT_)
+    else if (*P4_TO_CODE(xt) == p4_marker_RT_) 
+	p4_marker_RT_SEE (buf, xt, nfa);
+    else if (*P4_TO_CODE(xt) == p4_defer_RT_ ) 
+	p4_defer_RT_SEE (buf, xt, nfa);
+    else if (*P4_TO_CODE(xt) == p4_offset_RT_) 
+	p4_offset_RT_SEE (buf, xt, nfa);
+    else if (*P4_TO_CODE(xt) == p4_vocabulary_RT_) 
+	p4_vocabulary_RT_SEE (buf, xt, nfa);
+    else if (*P4_TO_CODE(xt) == p4_colon_RT_ || 
+	     *P4_TO_CODE(xt) == p4_debug_colon_RT_)
 	rest = p4_colon_RT_SEE(buf,xt,nfa);
-    else if (*xt == p4_does_defined_RT_ || *xt == p4_debug_does_defined_RT_)
-        rest = p4_does_defined_RT_SEE(buf,xt,nfa);
+    else if (*P4_TO_CODE(xt) == p4_does_RT_ || 
+	     *P4_TO_CODE(xt) == p4_debug_does_RT_)
+        rest = p4_does_RT_SEE(buf,xt,nfa);
 
     if (*buf)
     {
@@ -426,6 +442,17 @@ p4_decompile (char *nfa, p4xt xt)
 	    p4_outs ("is IMMEDIATE ");
 	else
 	    p4_outs ("is prim CODE ");
+	if (P4xISxRUNTIME)
+	    if (*_FFA(nfa) & P4xISxRUNTIME)
+		p4_outs ("RUNTIME ");
+#     ifdef PFE_HAVE_GNU_DLADDR
+	{ 
+	    extern char* p4_dladdr (void*, int*);
+	    register char* name = p4_dladdr (*P4_TO_CODE(xt), 0); 
+	    if (name) p4_outs(name); else p4_outc('.');
+	    p4_outc(' ');
+	}
+#     endif
     }
 }
 
@@ -438,13 +465,13 @@ p4_category (p4code p)
 {
     if (p == p4_colon_RT_ || p == p4_debug_colon_RT_)
         return ':';
-    if (p == p4_var_RT_ || p == p4_value_RT_ || p == p4_builds_RT_)
+    if (p == p4_variable_RT_ || p == p4_value_RT_ || p == p4_builds_RT_)
         return 'V';
     if (p == p4_constant_RT_ || p == p4_two_constant_RT_)
         return 'C';
     if (p == p4_vocabulary_RT_)
         return 'W';
-    if (p == p4_does_defined_RT_ || p == p4_debug_does_defined_RT_)
+    if (p == p4_does_RT_ || p == p4_debug_does_RT_)
         return 'D';
     if (p == p4_marker_RT_)
         return 'M';
@@ -478,7 +505,9 @@ display (p4xt *ip)
     }
     FX (p4_cr);
     p4_decompile_word (ip, buf, &style);
+# ifndef PFE_CALL_THREADING /* FIXME */
     p4_outf ("%*s%c %s", indent, "", p4_category (**ip), buf);
+# endif
 }
 
 static void
@@ -524,6 +553,7 @@ interaction (p4xt *ip)
              PFE.debugging = 0;
              p4_throw (P4_ON_QUIT);
          case ' ':
+#          ifndef PFE_CALL_THREADING /*FIXME*/
              switch (p4_category (**ip))
              {
               default:
@@ -535,9 +565,12 @@ interaction (p4xt *ip)
                   break;
               case 'd':
                   p4_outs ("\nDOES>");
+#               ifndef PFE_CALL_THREADING /*FIXME*/
                   p4_decompile_rest ((p4xt *) (*ip)[-1], 0, 4);
+#               endif
                   break;
              }
+#          endif
              FX (p4_cr);
              continue;
          case 'r':
@@ -566,13 +599,13 @@ interaction (p4xt *ip)
 static void
 do_adjust_level (p4xt xt)
 {
-    if (*xt == p4_colon_RT_ 
-      || *xt == p4_debug_colon_RT_ 
-      || *xt == p4_does_defined_RT_ 
-      || *xt == p4_debug_does_defined_RT_)
+    if (*P4_TO_CODE(xt) == p4_colon_RT_ 
+      || *P4_TO_CODE(xt) == p4_debug_colon_RT_ 
+      || *P4_TO_CODE(xt) == p4_does_RT_ 
+      || *P4_TO_CODE(xt) == p4_debug_does_RT_)
         PFE.level++;
-    else if (*xt == PFX (p4_semicolon_execution) 
-      || *xt == PFX (p4_locals_exit_execution))
+    else if (*P4_TO_CODE(xt) == PFX (p4_semicolon_execution) 
+      || *P4_TO_CODE(xt) == PFX (p4_locals_exit_execution))
         PFE.level--;
 }
 
@@ -603,6 +636,7 @@ p4_debug_off (void)
 static void			/* modified inner interpreter for */
 do_single_step (void)		/* single stepping */
 {
+# ifndef PFE_SBR_CALL_THREADING /* FIXME: disable */
     while (PFE.level >= 0)
     {
         if (PFE.level <= PFE.maxlevel)
@@ -613,16 +647,20 @@ do_single_step (void)		/* single stepping */
         do_adjust_level (*IP);
         PFE.opcounter++;
         {
-#ifdef P4_WP_VIA_IP
-            p4xt w = *IP++;	/* ip is register but W isn't */
+#         if defined PFE_CALL_THREADING
+	    p4xcode w = *IP++;
+	    w ();
+#         elif defined P4_WP_VIA_IP
+            p4xcode w = *IP++;	/* ip is register but W isn't */
             
             (*w) ();
-#else
+#         else
             p4WP = *IP++;	/* ip and W are same: register or not */
             (*p4WP) ();
-#endif
+#         endif
         }
     }
+# endif
 }
 
 FCode (p4_debug_colon_RT)
@@ -635,10 +673,12 @@ FCode (p4_debug_colon_RT)
         p4_debug_off ();
     }
 }
+static FCode (p4_debug_colon) { /* dummy */ }
+P4RUNTIME1(p4_debug_colon, p4_debug_colon_RT);
 
-FCode (p4_debug_does_defined_RT)
+FCode (p4_debug_does_RT)
 {
-    FX (p4_does_defined_RT);
+    FX (p4_does_RT);
     if (!PFE.debugging)
     {
         p4_debug_on ();
@@ -646,6 +686,8 @@ FCode (p4_debug_does_defined_RT)
         p4_debug_off ();
     }
 }
+static FCode (p4_debug_does) { /* dummy */ }
+P4RUNTIME1(p4_debug_does, p4_debug_does_RT);
 
 /** DEBUG ( 'word' -- )
  * this word will place an debug-runtime into
@@ -661,13 +703,13 @@ FCode (p4_debug)
     p4xt xt;
 
     xt = p4_tick_cfa (FX_VOID);
-    if (*xt == p4_debug_colon_RT_ 
-      || *xt == p4_debug_does_defined_RT_)
+    if (P4_XT_VALUE(xt) == FX_GET_RT (p4_debug_colon) 
+      || P4_XT_VALUE(xt) == FX_GET_RT (p4_debug_does))
         return;
-    else if (*xt == p4_colon_RT_)
-        *xt = p4_debug_colon_RT_;
-    else if (*xt == p4_does_defined_RT_)
-        *xt = p4_debug_does_defined_RT_;
+    else if (P4_XT_VALUE(xt) == FX_GET_RT (p4_colon))
+        P4_XT_VALUE(xt) = FX_GET_RT (p4_debug_colon);
+    else if (P4_XT_VALUE(xt) == FX_GET_RT (p4_does))
+        P4_XT_VALUE(xt) = FX_GET_RT (p4_debug_does);
     else
         p4_throw (P4_ON_ARG_TYPE);
 }
@@ -680,10 +722,10 @@ FCode (p4_no_debug)
     p4xt xt;
 
     xt = p4_tick_cfa (FX_VOID);
-    if (*xt == p4_debug_colon_RT_)
-        *xt = p4_colon_RT_;
-    else if (*xt == p4_debug_does_defined_RT_)
-        *xt = p4_does_defined_RT_;
+    if (P4_XT_VALUE(xt) == FX_GET_RT (p4_debug_colon))
+        P4_XT_VALUE(xt) = FX_GET_RT (p4_colon);
+    else if (P4_XT_VALUE(xt) == FX_GET_RT (p4_debug_does))
+        P4_XT_VALUE(xt) = FX_GET_RT (p4_does);
     else
         p4_throw (P4_ON_ARG_TYPE);
 }
@@ -702,7 +744,7 @@ FCode (p4_paren_see)
  * is next too. If it is not in the base-dictionary, then
  * just return 0 as not-found.
  */
-const p4char *
+_export const p4char *
 p4_addr_to_name (const p4char* addr)
 {
     Wordl* wl;
@@ -750,17 +792,21 @@ FCode (p4_addr_to_name)
  */
 FCode (p4_come_back)
 {
+# ifdef PFE_SBR_CALL_THREADING
+    p4_outs ("come_back not implemented in sbr-threaded mode\n");
+# else
     char const * nfa;
-    p4xt** rp = (p4xt**) p4_CSP;
+    p4xcode** rp = (p4xcode**) p4_CSP;
 
     if (PFE.rstack < rp && rp < PFE.r0)
     {
         if (PFE.dict < (p4char*) *rp && (p4char*) *rp < PFE.dp
           && (nfa = p4_addr_to_name ((void*)((*rp)[-1]))))
         {
-            p4_outf (" --> in %.*s \n", NFACNT(*nfa), nfa+1);
+            p4_outf ("[at] %08p ' %.*s (%+d) \n", *rp, NFACNT(*nfa), nfa+1,
+              ((p4xt) *rp) - (p4_name_from(nfa)));
         }else{
-            p4_outs (" --> (???) \n");
+            p4_outf ("[at] %08p (???) \n", *rp);
         }
 
         while (rp < RP)
@@ -768,11 +814,11 @@ FCode (p4_come_back)
             nfa = p4_addr_to_name ((void*)(*rp));
             if (nfa)
             {
-                p4_outf ("[%d] %08p ' %.*s (%+d) \n", 
+                p4_outf ("[%02d] %08p ' %.*s (%+d) \n", 
                   RP-rp, *rp, NFACNT(*nfa), nfa+1, 
                   ((p4xt) *rp) - (p4_name_from(nfa)));
             }else{
-                p4_outf ("[%d] %08p   %+ld \n", 
+                p4_outf ("[%02d] %08p   %+ld \n", 
                   RP-rp, *rp, (long) *rp);
             }
             rp++;
@@ -780,6 +826,7 @@ FCode (p4_come_back)
     }else{
         p4_outs (" come_back csp trashed, sorry \n");
     }
+# endif
 }
 
 P4_LISTWORDS (debug) =
