@@ -5,13 +5,13 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE              @(#) %derived_by: guidod %
- *  @version %version: bln_mpt1!5.33 %
- *    (%date_modified: Tue Oct 02 15:46:38 2001 %)
+ *  @version %version: bln_mpt1!5.55 %
+ *    (%date_modified: Wed Sep 11 13:17:53 2002 %)
  */
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  dict-sub.c~bln_mpt1!5.33:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  dict-sub.c~bln_mpt1!5.55:csrc:bln_12xx!1 % $";
 #endif
 
 #define _P4_SOURCE 1
@@ -30,6 +30,14 @@ static char* id __attribute__((unused)) =
 #include <pfe/_missing.h>
 #include <pfe/logging.h>
 
+#define UPPERMAX 32
+#define UPPERCOPY(_upper_, _nm_, _l_) \
+        if ((_l_) < UPPERMAX)         \
+        { memcpy ((_upper_), (_nm_), (_l_)); p4_upper ((_upper_), (_l_)); }  \
+        else                          \
+        { *(int*)(_upper_) = 0; }
+
+
 /*
  * A vocabulary is organized as a mixture between hash-table and
  * linked list. (This is a practice you can observe in several
@@ -45,14 +53,14 @@ _export int
 p4_wl_hash (const char *s, int l)
 /* s string, l length of string, returns hash-code for that name */
 {
-#  if 0 /*GD* original by DUZ */
+#  if 0 /*GUD* original by DUZ */
     int n = *s++ - '@';
 
     while (--l > 0)
         n = n * 37 + *s++ - '@';        /* a maybe-stupid hash function :-) */
     return n & (THREADS - 1);   /* i.e. modulo threads */
 
-#  else /*GD* more simpler, avoiding multiply */
+#  else /*GUD* more simpler, avoiding multiply */
     register char c = *s;
 
     while(--l > 0)
@@ -124,8 +132,10 @@ p4_make_wordlist (p4char* nfa)
 _export p4_Wordl *
 p4_find_wordlist (const char* nm, int nmlen)
 {
+    auto p4char upper[UPPERMAX];
     p4_Wordl* wl;
     char* nfa;
+    UPPERCOPY(upper, nm, nmlen)
 
     /* a special, since FORTH has no vocabulary_RT anymore */
     if (nmlen == 5 && ! memcmp (nm, "FORTH", 5))
@@ -137,7 +147,8 @@ p4_find_wordlist (const char* nm, int nmlen)
     {
         if (! wl->nfa) continue;
         nfa = wl->nfa;
-        if (NFACNT(*nfa) == nmlen && !p4_strncmpi (nfa+1, nm, nmlen))
+        if (NFACNT(*nfa) != nmlen) continue;
+        if (!memcmp (nfa+1, nm, nmlen) || !memcmp (nfa+1, upper, nmlen))
             return wl;
     }
     return 0;
@@ -157,7 +168,10 @@ FCode (p4_forget_dp)
     {
         p4char **p = wl->thread;
         int i;
-        
+       
+	if (0) if (wl->nfa) 
+	    fprintf(stderr,"\"%.*s\"", NFACNT(*wl->nfa), wl->nfa+1);
+
         for (i = THREADS; --i >= 0; p++)
         {  /* unchain words in thread: */
             while (*p >= new_dp) 
@@ -179,7 +193,7 @@ FCode (p4_forget_dp)
     {   
         {   /* delete from search-order */   
             int i;
-            for (i=0; i < ORDER_LEN; i++) 
+            for (i=0; i < PFE_set.wordlists; i++) 
             {
                 if (CONTEXT[i] == VOC_LINK) 
                 {
@@ -214,17 +228,17 @@ FCode (p4_forget_dp)
     
     /* compact search-order */
     { register int i, j;
-      for (i=0, j=0; i < ORDER_LEN; i++)
+      for (i=0, j=0; i < PFE_set.wordlists; i++)
       {
         if (CONTEXT[i]) CONTEXT[j++] = CONTEXT[i];
       }
-      while (j < ORDER_LEN) CONTEXT[j++] = NULL;
+      while (j < PFE_set.wordlists) CONTEXT[j++] = NULL;
 
-      for (i=0, j=0; i < ORDER_LEN; i++)
+      for (i=0, j=0; i < PFE_set.wordlists; i++)
       {
         if (PFE.dforder[i]) PFE.dforder[j++] = PFE.dforder[i];
       }
-      while (j < ORDER_LEN) PFE.dforder[j++] = NULL;
+      while (j < PFE_set.wordlists) PFE.dforder[j++] = NULL;
     }
     
     /* free dictionary space: */
@@ -263,13 +277,24 @@ p4_forget (char* above)
     }
 }
 
-FCode (p4_destroyer_RT)
-{
-    /* put the argument for the destroyer-call in PFA[0] and
-     * the call-address in PFA[1]. And here we call it.
+FCode_RT (p4_destroyer_RT)
+{   
+    /* this code is a trampoline for ITC code not using an FFA flag.
+     * we just expect the a prior p4_call in p4_forget has setup an
+     * appropriate BODY pointer - either it goes through a p4WP or
+     * indirectly through p4IP. All we have to do is, to fetch the
+     * actual ccode from PFA[1] and branch down into the target code.
      */
-    p4_call ((p4xt) &(WP_PFA [1]));
+#   if ! defined PFE_CALL_THREADING
+    ((p4code*)(P4_WP_PFA))[1](FX_VOID);
+#   elif ! defined PFE_SBR_CALL_THREADING /* not actually ever compiled */
+    ((p4code*)(*IP))[1](FX_VOID); /* CTC, (*IP) would have the BODY-pointer */
+#   else
+    P4_fail (
+        "you are not supposed to have a destroyer_RT in this configuration");
+#   endif
 }
+P4RUNTIME1_RT(p4_destroyer);
 
 /**
  * create a destroyer word. Upon =>'FORGET' the code will will be
@@ -278,24 +303,37 @@ FCode (p4_destroyer_RT)
 _export char*
 p4_forget_word (const char *name, p4cell id, p4code ccode, p4cell what)
 {
-    char* nfa;
-    char nm[255];
-
+    char nm[255]; /* better as POCKET ? */
     sprintf (nm, name, id);
 
-#  if defined PFE_WITH_FFA  
-    nfa = p4_make_header (ccode, 0, nm, strlen(nm), PFE.atexit_wl);
-    *_FFA(nfa) |= (P4xIMMEDIATE|P4xONxDESTROY);
-    FX_VCOMMA (what);  /*pfa*/
-#  else
-    p4_make_header (PFX(p4_destroyer_RT, 0, 
-                        nm, strlen(nm), PFE.atexit_wl)
-    *_FFA(nfa) |= P4xIMMEDIATE;
-    FX_VCOMMA (what); /*pfa*/
-    FX_RCOMMA (ccode); /*pfa+1*/
-#endif    
+#  if defined PFE_WITH_FFA  && ! defined PFE_CALL_THREADING
+    {
+	p4_header_comma (nm, strlen(nm), PFE.atexit_wl);
+	*_FFA(LAST) |= (P4xIMMEDIATE|P4xONxDESTROY);
+	FX_RCOMMA (ccode); /*cfa*/
+	FX_VCOMMA (what);  /*pfa*/
+    }
+#  elif ! defined PFE_CALL_THREADING
+    {
+	p4_header_comma (nm, strlen(nm), PFE.atexit_wl);
+	*_FFA(LAST) |= (P4xIMMEDIATE|P4xONxDESTROY);
+	FX_RUNTIME1_RT (p4_destroyer);
+	FX_VCOMMA (what); /*pfa*/
+	FX_RCOMMA (ccode); /*pfa+1*/
+    }
+#  else /* PFE_CALL_THREADING and up, p4_call uses INFO-block */
+    {
+	p4Word* w = (p4xt) p4_DP;
+	FX_RCOMMA ("~"); /* trampoline with body */
+	FX_RCOMMA (ccode);
+	p4_header_comma (nm, strlen(nm), PFE.atexit_wl);
+	*_FFA(LAST) |= (P4xIMMEDIATE|P4xONxDESTROY);
+	FX_PCOMMA (w);    /* cfa = word-comp-info */
+	FX_VCOMMA (what); /* pfa */
+    }
+#  endif    
     
-    return nfa;
+    return LAST;
 }
 
 /* ------------------------------ 
@@ -305,6 +343,7 @@ p4_forget_word (const char *name, p4cell id, p4code ccode, p4cell what)
 static char *
 search_thread (const char *nm, int l, char *t, p4cell wl_flag )
 {
+    auto char upper[UPPERMAX];
     if (l > NFACNTMAX)
         return NULL;
 
@@ -318,23 +357,53 @@ search_thread (const char *nm, int l, char *t, p4cell wl_flag )
 # endif
 
     if( LOWER_CASE && (wl_flag & WORDL_NOCASE) )
-    {
-        /* this thread does only contain upper-case defs 
-           AND lower-case input shall match definitions */
+    {   /* pfe 33.x will have this mode removed: */
         while (t)
         {
-            if ( !(*_FFA(t) & P4xSMUDGED) 
-              && NFACNT(*t) == l && !p4_strncmpi (nm, t+1, l))
-                break;
+            if ( !(*_FFA(t) & P4xSMUDGED) && NFACNT(*t) == l)
+            {
+                if (!p4_strncmpi (nm, t+1, l))  break;
+            }
+            t = *p4_name_to_link (t);
+        }
+        if (t && (wl_flag & WORDL_UPPER_CASE) && memcmp (nm, t+1, l)) 
+        {   /* does not match case-sensitive */
+            UPPERCOPY (upper, nm, l);
+            if (memcmp (upper, t+1, l)) 
+            {  /* and neither the upper-case form */
+#             if defined _K12_SOURCE && defined VXWORKS
+                p4_outf ("(warning: input '%.*s' hits '%.*s': bad spelling?)", 
+                          l, nm, l, t+1);
+#             endif
+                P4_warn4 ("oops, input '%.*s' hits '%.*s': bad spelling?", 
+                          l, nm, l, t+1);
+                P4_info4 ("- the input word '%.*s' might not match '%.*s'"
+                          "in the future - please fix it now.", l, nm, l, t+1);
+            }
+        }
+    }else if( UPPER_CASE && (wl_flag & WORDL_UPPER_CASE) )
+    {   /* note: p4_match/p4_search_incomplete */
+        UPPERCOPY (upper, nm, l);
+
+        /* this thread does contain some upper-case defs 
+           AND lower-case input shall match those definitions */
+        while (t)
+        {
+            if ( !(*_FFA(t) & P4xSMUDGED) && NFACNT(*t) == l)
+            {
+                if (!memcmp (nm, t+1, l))  break;
+                if (!memcmp (upper, t+1, l)) break;
+            }
             t = *p4_name_to_link (t);
         }
     }else{
         /* input is case-sensitive OR vocabulary contains mixed-case defs */
         while (t)
         {
-            if ( !(*_FFA(t) & P4xSMUDGED) 
-              && NFACNT(*t) == l && !strncmp (nm, t+1, l))
-                break;
+            if ( !(*_FFA(t) & P4xSMUDGED) && NFACNT(*t) == l)
+            {
+                if (!memcmp (nm, t+1, l))  break;
+            }
             t = *p4_name_to_link (t);
         }
     }
@@ -451,20 +520,14 @@ p4_string_comma (const char *s, int len)
     return p;
 }
 
-#if 0
-_export char*
-/*use FX(p4_parse_comma_quote)*/   p4_parse_comma(char del)
-{
-    p4_word_parse (del); /* PARSE-WORD-NOTHROW */
-    return p4_string_comma (PFE.word.ptr, (int) PFE.word.len);
-}
-#endif
-
-static void                     /* written to cfa following make_header() */
-illegal_xt (void)               /* to give an error msg when calling */
-{                               /* a word without execution semantics */
-    p4_throw (P4_ON_INVALID_NAME);
-}
+/*
+ *  char*                       |* use FX(p4_parse_comma_quote) !!! *|
+ *  p4_parse_comma(char del)
+ * {
+ *    p4_word_parse (del); |* PARSE-WORD-NOTHROW *|
+ *    return p4_string_comma (PFE.word.ptr, (int) PFE.word.len);
+ * }
+ */
 
 /* ----------------------
  * words with wildcards 
@@ -478,8 +541,9 @@ illegal_xt (void)               /* to give an error msg when calling */
 _export void
 p4_wild_words (const p4_Wordl *wl, const char *pattern, const char *categories)
 {
-    Wordl wcopy = *wl;          /* clobbered while following it */
     p4char **t;
+    /* Wordl wcopy = *wl;          // clobbered while following it */
+    Wordl wcopy; memcpy (&wcopy, wl, sizeof(wcopy));
 
 # ifndef WILD_TAB
 # define WILD_TAB 26 /* traditional would be 20 (26*4=80), now 26*3=78 */
@@ -495,11 +559,10 @@ p4_wild_words (const p4_Wordl *wl, const char *pattern, const char *categories)
         p4char *w = *t;
         p4char **s = p4_name_to_link (w);
         int l = NFACNT(*w++);
-        
         p4_store_c_string (w, l, wbuf, sizeof wbuf);
-        if (p4_match (pattern, wbuf, wl->flag & WORDL_NOCASE))
+        if (p4_match (pattern, wbuf, wl->flag & P4_UPPER_CASE_FLAGS))
         {
-            char c = p4_category (*P4_LINK_FROM (s));
+	    char c = p4_category (*P4_TO_CODE(P4_LINK_FROM (s)));
             if (! categories || strchr (categories, c))
             {
                 if (p4_OUT+WILD_TAB - p4_OUT%WILD_TAB + 2 + l > p4_COLS ||
@@ -518,26 +581,35 @@ p4_wild_words (const p4_Wordl *wl, const char *pattern, const char *categories)
     }
 }
 
+#ifdef _use_pfe_old_style_search_incomplete
+
 /* completion of word against dictionary */
 static p4char *
-search_incomplete (const char *name, int len, Wordl *w)
+search_incomplete (const char *nm, int l, Wordl *w)
 /*
  * traverses the entire given wordlist to find a matching word
  * caution: clobbers *w. This is needed to be able to continue the search.
  */
 {
+    auto char upper[UPPERMAX];
     p4char **t, *s;
+
+    if (w->flag & P4_UPPER_CASE_FLAGS)
+    {
+        UPPERCOPY(upper, nm, l);
+    }
     
     for (t = p4_topmost (w); *t; t = p4_topmost (w))
     {
         s = *t;
         *t = *p4_name_to_link (*t);
-        if (NFACNT(*s) >= len 
-          && ( (w->flag & WORDL_NOCASE) 
-            ? p4_strncmpi (s + 1, name, len) 
-            : strncmp  (s + 1, name, len) 
-               ) == 0)
-            return s;
+
+        if ( !(*_FFA(t) & P4xSMUDGED) && NFACNT(*s) >= l)
+        {
+            if (NFACNT(*s) <= UPPERMAX) 
+            { if (!memcmp (upper, s+1, l)) return s; }
+            if (! memcmp (nm, s+1, l))  return s;
+        }
     }
     return NULL;
 }
@@ -576,7 +648,7 @@ p4_complete_word (const char *in, int len, char *out, int display)
                     FX (p4_space);
                     p4_type_on_line (t + 1, NFACNT(*t));
                 }
-                if (cnt == 0) 
+                if (! s) 
                 {
                     s = t + 1;
                     m = NFACNT(*t);
@@ -597,6 +669,137 @@ p4_complete_word (const char *in, int len, char *out, int display)
     return cnt;
 }
 
+#else
+
+static char *
+search_thread_inco (const char *nm, int l, char *t, p4cell wl_flag )
+{  /* compare with p4_search_thread */
+    auto char upper[UPPERMAX];
+    if (l > NFACNTMAX)
+        return NULL;
+
+    if( UPPER_CASE && (wl_flag & WORDL_UPPER_CASE) )
+    {   /* note: p4_match/p4_search_incomplete */
+        UPPERCOPY (upper, nm, l);
+
+        /* this thread does contain some upper-case defs 
+           AND lower-case input shall match those definitions */
+        while (t)
+        {
+            if ( !(*_FFA(t) & P4xSMUDGED) && NFACNT(*t) >= l)
+            {
+                if (!memcmp (nm, t+1, l))  break;
+                if (!memcmp (upper, t+1, l)) break;
+            }
+            t = *p4_name_to_link (t);
+        }
+    }else{
+        /* input is case-sensitive OR vocabulary contains mixed-case defs */
+        while (t)
+        {
+            if ( !(*_FFA(t) & P4xSMUDGED) && NFACNT(*t) >= l)
+            {
+                if (!memcmp (nm, t+1, l))  break;
+            }
+            t = *p4_name_to_link (t);
+        }
+    }
+    return t;
+}
+
+static char *
+find_next_incomplete (const char *nm, int l, char* old)
+{   /* compare with p4_find */
+    register Wordl **p, **q;
+    register Wordl *wordl;
+    register char *w = NULL;
+    register int n;
+    register p4ucell searched = 0;
+    
+    for (p = CONTEXT; p <= &ONLY; p++)
+    {
+        for (q = CONTEXT; q < p; q++)
+        {
+            if (!*q) continue;
+            if (*q == *p) goto continue_outer;
+        }
+
+        for (wordl = *p; wordl ; wordl=wordl->also)
+	{
+            /* we _must_ ensure that no world-thread is searched twice unlike
+             * p4_find, so just skip any wordl that does not have an id for
+             * the search_also case. Generally, we do not miss any word thus.
+             */
+            if (! wordl->id) 
+                continue;  
+	    if (searched & wordl->id)
+		continue;
+	    searched |= wordl->id;
+            
+            for (n = 0; n < THREADS; n++)
+            {
+                w = wordl->thread[n];
+                while (w)
+                {
+                    w = search_thread_inco (nm, l, w, wordl->flag );
+
+                    if (w) 
+                    {
+                        if (! old)               /* search all wordl-threads */
+                            return w;            /* until "old" is found     */
+                        if (old == w)            /* then return the next nfa */
+                            old = 0;             /* found after that one -   */
+                        w = *p4_name_to_link(w); /* and if old-arg was null  */
+                    }                            /* it returns the first w   */
+                } /* while (w) */
+            }
+        }
+    continue_outer:
+        continue;
+    }
+    return w; /*0*/
+}
+
+/*
+ * Try to complete string in/len from dictionary.
+ * Store completion in out (asciiz), return number of possible completions.
+ * If display is true, display alternatives.
+ * (if (display && !len) { don't print 200 words, just the number })
+ */
+static int
+p4_complete_word (const char *in, int len, char *out, int display)
+{
+    char *s = NULL, *t = NULL;  
+    int n = 0, m = 0, cnt = 0;
+
+    while ((t = find_next_incomplete(in, len, t)))
+    {
+        cnt ++;
+        if (display && len) 
+        {
+            FX (p4_space);
+            p4_type_on_line (t + 1, NFACNT(*t));
+        }
+        if (! s) 
+        {
+            s = t + 1;
+            m = NFACNT(*t);
+        }else{
+            for (n = 0; n < m; n++)
+                if (s[n] != (t+1)[n])
+                    break;
+            m = n;
+        }
+    }
+    if (cnt)
+        p4_store_c_string (s, m, out, NFACNTMAX+1);
+    if (display && !len)
+    { p4_outf (" %i words ", cnt); }
+    return cnt;
+}
+
+#endif
+
 _export int
 p4_complete_dictionary (char *in, char *out, int display)
 {
@@ -614,299 +817,9 @@ p4_complete_dictionary (char *in, char *out, int display)
     return n;
 }
 
-/* ---------------------------------------------------------------------- *
- * initial dictionary setup                                             
- */
-
-FCode (p4_forget_wordset_RT)
-{
-    /* do nothing so far, forget_wordset_RT_ is just a type-marker */
-}
-
-extern int p4_slot_use (int*); /* FIXME: move to header file ? */
-extern int p4_slot_unuse (int*); /* FIXME: move to header file ? */
-
-static FCode (p4_forget_slot_RT)
-{
-    int* slot = (int*)(WP_PFA[0]);
-    P4_info1 ("unuse load-slot '%i'", *slot);
-   
-    if (slot && *slot && PFE.p[*slot]) 
-    {
-        p4_xfree (PFE.p[*slot]); PFE.p[*slot] = 0;
-    }
-   
-    p4_slot_unuse (slot);
-}
-
-static void
-p4_load_slot_open (int* slot)
-{
-    int e;
-    if (!slot) return;
-                
-    if ((e=p4_slot_use (slot))) 
-    {
-        P4_fail2 ("load-slot %i failed : %s", *slot, strerror(-e));
-        return; 
-    }
-}
-
-static void
-p4_load_slot_init (int* slot, p4ucell size)
-{
-    if (!slot || !*slot || size < 4)
-        return;
-
-    if (!(PFE.p)[*slot]) 
-    {
-        (PFE.p)[*slot] = p4_calloc (1, size);
-        P4_info3 ("load-slot %i size %lu alloc (%p)", 
-                  *slot, (unsigned long)size, (PFE.p)[*slot]);
-    }else{ 
-        P4_warn2 ("load-slot %i already allocated (%p)", 
-                  *slot, (PFE.p)[*slot]);
-    }
-    
-    p4_forget_word ("(load-slot: %i)", *slot, 
-                    p4_forget_slot_RT_, (p4cell) slot);
-}
-
-static void
-p4_load_into (const char* vocname)
-{
-    Wordl* voc;
-    if (! vocname) return;
-
-    voc = p4_find_wordlist (vocname, strlen(vocname));
-    if (voc) 
-    {
-        {    
-            register int i;
-            for (i=ORDER_LEN; --i > 0; )
-                if (CONTEXT[i] == voc) 
-                {
-                    P4_info1 ("search also '%s' : already there", 
-                              vocname);
-                    return;
-                }
-        }
-        FX (p4_also);    /* the top-of-order (CONTEXT) isn't changed */
-        CONTEXT [1] = voc; /* instead we place it under-the-top */
-        P4_info1 ("search also '%s' : done", vocname);
-    }else{
-        P4_warn2 ("search also failed: no '%s' vocabulary (%lu)", 
-                  vocname, (unsigned long) strlen(vocname));
-    }
-} 
-
-static void p4_exception_string (const char* name, p4cell id)
-{
-    /* FIXME: instead of compiling to the forth-dict we should better
-       create a way to let functions search the loaded wordset tables
-    */
-    p4_Exception* expt = (void*) DP; DP += sizeof(*expt);
-    if (id < PFE.next_exception) PFE.next_exception = id - 1;
-    expt->next = PFE.exception_link; PFE.exception_link = expt;
-    expt->name = name; expt->id = id;
-}
-
-_export void
-p4_load_words (const p4Words* ws, p4_Wordl* wid, int unused)
-{
-    Wordl* save_current = CURRENT;
-    int k = ws->n;
-    const p4Word* w = ws->w;
-    char dictname[NFACNTMAX+1]; char* dn;
-    int* slot = 0;
-
-    if (!wid) wid = CURRENT;
-    
-    if (ws->name) 
-    {  
-        P4_info1 ("load '%s'", (ws->name));
-        strncpy (dictname, ws->name, NFACNTMAX);
-        dictname[NFACNTMAX] = '\0';
-        if ((dn= strchr (dictname, ' '))
-            ||  (dn= strchr (dictname, '(')))
-            *dn = '\0';
-    }else{
-        sprintf (dictname, "%p", DP);
-    }
-    
-    p4_forget_word ("wordset:%s", (p4cell) dictname,
-                    p4_forget_wordset_RT_, 
-                    (p4cell) (ws));
-    
-    for ( ; --k >= 0; w++)
-    {
-        wid = CURRENT;
-        if (w)
-        {
-            /* the C-name is really type-byte + count-byte away */
-            char type = *w->name;
-            const char* name = w->name+2;
-            int len = strlen (w->name+2);
-            void* ptr = w->ptr;
-            Wordl* wid = CURRENT;
-
-            /* and paste over make_word inherited from pre 0.30.28 times */
-            p4xt  cfa;
-            char *nfa;
-
-            /* part 1: specials... */
-  
-            switch (type)
-            {
-            case p4_LOAD:
-                if (ptr)
-                    p4_load_words ((p4Words*) ptr, 0, 0); /* RECURSION !! */
-                continue;
-            case p4_INTO:
-            {
-                register void* p;
-                p = p4_find_wordlist (name, strlen (name));
-                if (p) 
-                {   
-                    P4_debug1 (13, "load into old '%s'", name);
-                    CURRENT = p;
-                }else{
-                    Wordl* current = 0;
-                    if (ptr) {
-                        current = p4_find_wordlist (ptr, strlen(ptr));
-                        if (! current) 
-                            P4_warn1 ("could not find also-voc %s", 
-				      (char*)(ptr));
-                    }
-                    if (! current) current = CURRENT;
-                    P4_info1 ("load into new '%s'", name);
-                    CURRENT = p4_make_wordlist (p4_make_header (
-                        PFX(p4_vocabulary_RT), P4xIMMEDIATE,
-                        name, strlen(name), current) );
-                }
-
-                if (ptr) 
-                {
-                    if (! CURRENT->also)
-                        CURRENT->also = p4_find_wordlist (ptr, strlen(ptr));
-
-                    p4_load_into (name); /* search-also */
-                }
-            } continue;
-            case p4_SLOT:
-                slot = (int*) ptr;
-                p4_load_slot_open (slot);
-                continue;
-            case p4_SSIZ:
-                p4_load_slot_init (slot, (p4ucell)(ptr));
-                continue;
-            case p4_EXPT:
-                p4_exception_string(name, (p4cell)(ptr));
-                continue;
-            case p4_XXCO: /* constructors are registered in => LOADED */
-                wid = PFE.atexit_wl;
-                break;
-            } /*switch*/
-
-            /* part 2: general... CREATE a name and setup its CFA field */
-
-            nfa = p4_make_header (illegal_xt, 0, name, len, wid);
-            if ('A' <= type && type <= 'Z')
-                *_FFA(nfa) |= P4xIMMEDIATE;
-            cfa = P4_BODY_FROM(DP);
-            switch (type)
-            {
-            case p4_SXCO:
-#             ifndef HOST_WIN32
-                *cfa = ((p4_Semant *) ptr) ->comp;
-                if (! ((p4_Semant *)ptr) ->name)
-                    (char*) ((p4_Semant *)ptr) ->name = name-1; 
-                /* discard const */
-                /* BEWARE: the arg' name must come from a wordset entry to
-                   be both static and have byte in front that could be 
-                   a maxlen
-                */
-#             else
-		/* on WIN32, the ptr is a function that returns a SemantP */
-		*cfa = ((p4_Semant*(*)())ptr) () -> comp;
-#             endif
-                continue;
-            case p4_RTCO:
-                *cfa = ((p4_Runtime2 *) ptr) ->comp;
-		/* and start registering the runtimes centrally FIXME:
-		   FX_COMMA(PFE.runtime); PFE.runtime = p4_HERE;
-		   FX_COMMA(ptr);
-		   but that sys-link should be honoured in p4_forget too
-		*/
-                continue;
-            case p4_IXCO:
-            case p4_FXCO:
-                *cfa = (p4code) ptr;
-                continue;
-            case p4_XXCO:
-                *cfa = (p4code) ptr;
-                ((p4code)ptr) ();     /* runs *now* !! no checks !! */
-                continue;
-            case p4_IVOC:
-            case p4_OVOC:
-                /* creating a VO before IN will make sure that the */
-                /* other words will go in there. Nice stuff, eh ;-) */
-                *cfa = p4_vocabulary_RT_ ;
-                /* (((WList*) ptr)->wid = p4_make_wordlist (nfa)); */
-                continue;
-            case p4_DVAR:
-                *cfa = p4_dictvar_RT_ ;
-                break;
-            case p4_DCON:
-                *cfa = p4_dictget_RT_ ;
-                break;
-            case p4_DSET:
-                *cfa = p4_dictset_RT_ ;
-                break;
-            case p4_OVAR:
-            case p4_IVAR:
-                *cfa = p4_var_RT_ ;
-                break;
-            case p4_OVAL:
-            case p4_IVAL:
-                *cfa = p4_value_RT_ ;
-                break;
-            case p4_OFFS:
-                *cfa = p4_offset_RT_ ;
-                break;
-	    case p4_iOLD:
-	    case p4_xOLD:
-		*cfa = p4_obsoleted_RT_;
-		if (p4_LogMask && p4_LogMask^P4_LOG_FATAL) goto synonym;
-            case p4_SNYM:
-            case p4_FNYM:
-                *cfa = p4_synonym_RT_ ;
-	    synonym:
-		ptr = p4_find (ptr, strlen(ptr));
-		if (ptr) ptr = p4_name_from (ptr);
-		else P4_fail3 ("could not resolve SYNONYM %.*s %s",
-			       NFACNT(*nfa), nfa+1, (char*)w->ptr);
-		break;
-            default:
-                P4_fail3 ("unknown typecode for loadlist entry: "
-                          "0x%x -> \"%.*s\"", 
-                          type, len, name);
-            case p4_OCON:
-            case p4_ICON:
-                *cfa = p4_constant_RT_ ;
-                break;
-            }
-            FX_VCOMMA (ptr);
-            continue;
-        } /* if(w) */
-    } /* for w in ws->w */
-
-    CURRENT = save_current; /* should save_current moved to the caller? */
-}
-
 /* ------------------------------------------------------------------- */
 
-/** ((ONLY)) ( -- )
+/** ONLY ( -- )
  * the only-vocabulary is special. Calling it will erase
  * the search => ORDER of vocabularies and only allows
  * to name some very basic vocabularies. Even => ALSO
@@ -916,51 +829,55 @@ p4_load_words (const p4Words* ws, p4_Wordl* wid, int unused)
  */
 FCode (p4_only_RT)
 {
-    ZERO (CONTEXT);
+    /* NO BODY_ADDR */
+    memset(CONTEXT, 0, PFE_set.wordlists*sizeof(p4_Wordl*));
     CONTEXT[0] = CURRENT = ONLY;
 }
+P4RUNTIMES1_RT(p4_only, P4_ONLY_CODE1);
 
 /** FORTH ( -- )
  : FORTH FORTH-WORDLIST CONTEXT ! ;
  */
 FCode (p4_forth_RT)
 {
+    /* NO BODY_ADDR */
     CONTEXT[0] = PFE.forth_wl;
 }
+P4RUNTIMES1_RT(p4_forth, P4_ONLY_CODE1);
 
 _export void
 p4_preload_only (void)
 {
-    Wordl only;                   /* scratch ONLY word list */
-    char* nfa;
+    auto p4_Wordl only;                   /* scratch ONLY word list */
     
-    DP = (p4char *) &PFE.dict[1];
+    DP = (p4char *) & PFE.dict[1];
   
     /* Load the ONLY word list to the scratch ONLY: */
     memset (&only, 0, sizeof only);
     /* # only.flag |= WORDL_NOHASH; */
-    nfa = p4_make_header (p4_only_RT_, 0, "ONLY", 4, &only );
-    ONLY = p4_make_wordlist (nfa);
+    p4_header_comma ("ONLY", 4, &only ); FX_RUNTIME1_RT(p4_only);
+    ONLY = p4_make_wordlist (LAST);
     /* # ONLY->flag |= WORDL_NOHASH; */
     COPY (ONLY->thread, only.thread);   /* Copy scratch ONLY to real ONLY */
     CURRENT = ONLY;
 
     /* FORTH -> [ANS] -> ONLY */
-    nfa = p4_make_header (p4_forth_RT_ , 0, "FORTH", 5, ONLY);
-    PFE.forth_wl = p4_make_wordlist (nfa); 
-    nfa = p4_make_header (p4_vocabulary_RT_, P4xIMMEDIATE, "[ANS]", 5, ONLY);
-    PFE.forth_wl->also = p4_make_wordlist (nfa);
+    p4_header_comma ("FORTH", 5, ONLY); FX_RUNTIME1_RT(p4_forth);
+    PFE.forth_wl = p4_make_wordlist (LAST); 
+    p4_header_comma ("[ANS]", 5, ONLY); FX_RUNTIME1(p4_vocabulary);
+    FX_IMMEDIATE;
+    PFE.forth_wl->also = p4_make_wordlist (LAST);
     PFE.forth_wl->also->also = ONLY;
 
     /* destroyers :: LOADED */
-    nfa = p4_make_header (p4_vocabulary_RT_ , 0, "LOADED", 6, ONLY);
-    PFE.atexit_wl = p4_make_wordlist (nfa); 
+    p4_header_comma ("LOADED", 6, ONLY); FX_RUNTIME1(p4_vocabulary);
+    PFE.atexit_wl = p4_make_wordlist (LAST); 
     PFE.atexit_wl->flag |= WORDL_NOHASH; /* see environment_dump in core.c */
 
     /* ENVIRONMENT -> LOADED */
-    nfa = p4_make_header (p4_vocabulary_RT_ , P4xIMMEDIATE, 
-                          "ENVIRONMENT", 11, ONLY);
-    PFE.environ_wl = p4_make_wordlist (nfa);
+    p4_header_comma ("ENVIRONMENT",11, ONLY); FX_RUNTIME1(p4_vocabulary);
+    FX_IMMEDIATE;
+    PFE.environ_wl = p4_make_wordlist (LAST);
     PFE.environ_wl->also = PFE.atexit_wl;
     PFE.environ_wl->flag |= WORDL_NOHASH;          /* for option-ext */
     PFE.environ_wl->thread[0] = PFE.set->opt.link; /* that goes here */
