@@ -5,8 +5,8 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE                %derived_by: guidod %
- *  @version %version: 2.16 %
- *    (%date_modified: Tue Jul 17 17:41:41 2001 %)
+ *  @version %version: bln_mpt1!2.25 %
+ *    (%date_modified: Mon Apr 08 20:16:52 2002 %)
  *
  *  @description
  *		This file exports a set of system words for 
@@ -15,7 +15,7 @@
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  dl-ext.c~2.16:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  dl-ext.c~bln_mpt1!2.25:csrc:bln_12xx!1 % $";
 #endif
 
 #define _P4_SOURCE 1
@@ -496,8 +496,9 @@ p4_dlslot_open (const char* nameptr, int namelen)
     if (! strchr (p4_dlslot_table[slot].name, '/')
 	&& ! strchr (p4_dlslot_table[slot].name, ':'))
     {
-	char named[255] = PFE_MODULE_DIR;  /* system module path */
-	int len = strlen(named);
+	/* char named[255] = PFE_MODULE_DIR;  // system module path */
+	char named[255]; int len;
+	strcpy (named, PFE_MODULE_DIR);	len = strlen(named); 
 	if (named[len-1] != '/' && named[len-1] != '\\' && named[len-1] != ':')
 	    strncat (named, PFE_DIR_DELIMSTR, 255);
 	strncat (named, p4_dlslot_table[slot].name, 255);
@@ -505,6 +506,10 @@ p4_dlslot_open (const char* nameptr, int namelen)
 	if (memchr (named, '\0', 255))
 	{
 	    dll = p4_dlopenext (named);
+#         ifdef _CHECKDLL
+	    if (! dll)
+		P4_warn1 ("%s", p4_dlerror ());
+#         endif
 	    if (systemonly)
 		goto skipdirectpath;
 	}
@@ -514,7 +519,11 @@ p4_dlslot_open (const char* nameptr, int namelen)
  skipdirectpath:
     if (! dll) 
     { 
-        p4_dlerror ();
+#     ifdef _CHECKDLL
+        P4_warn1 ("%s", p4_dlerror ());
+#     else
+	p4_dlerror ();
+#     endif
         p4_dlslot_remove (slot);
         mutexGive ();
         return -ENOENT; 
@@ -564,21 +573,21 @@ p4_dlslot_close (int slot)
 }
 
 void
-p4_forget_loadm (void)
-{
-    int slot = WP_PFA[0];
+p4_forget_loadm (void) /* FCode_RT */
+{   FX_USE_BODY_ADDR {
+    int slot = FX_POP_BODY_ADDR[0];
   
     P4_note1 ("unuse dlslot %i", slot);
     
     p4_dlslot_close (slot);
-}
+}}
 
 void
-p4_forget_loadm_internal (void)
-{
-    int slot = WP_PFA[0];
-    P4_note1 ("unuse internal %i", slot);
-}
+p4_forget_loadm_prelinked (void) /* FCode_RT */
+{   FX_USE_BODY_ADDR {
+    int slot = FX_POP_BODY_ADDR[0];
+    P4_note1 ("unuse prelinked %i", slot);
+}}
 
 void*
 p4_loadm (const p4char* nm, int l)
@@ -610,11 +619,17 @@ p4_loadm (const p4char* nm, int l)
             return 0;
         }
     }
-    else
-    {   /* no external module loaded, try to find an internal module */
-        extern const p4Words P4WORDS(internal);
-        int k = P4WORDS(internal).n;
-        const p4Word* w = P4WORDS(internal).w;
+
+    /* no external module loaded, try to find a prelinked module */
+    if (PFE.set->modules)
+    {
+	/*  extern const p4Words P4WORDS(internal);
+	 *  register int k  = P4WORDS(internal).n;
+	 *  const p4Word* w = P4WORDS(internal).w;
+	 */
+	register int k  = ((p4Words*)PFE.set->modules)->n;
+	const p4Word* w = ((p4Words*)PFE.set->modules)->w;
+	
         for ( ; --k >= 0; w++ )
         {
             const char* name;
@@ -631,13 +646,13 @@ p4_loadm (const p4char* nm, int l)
 
             /* found it ! */
             p4_forget_word ("%s.O", (p4cell) name, 
-                            p4_forget_loadm_internal, (p4cell) -k);
+                            p4_forget_loadm_prelinked, (p4cell) -k);
             p4_load_words (ptr, ONLY, 1); /* load the words */
             return ptr;
         }
-
-        return 0;  /* no luck either */
     }
+
+    return 0;  /* no luck in either case */
 }
 
 /** (LOADM) ( bstring -- sucess )
@@ -723,6 +738,58 @@ FCode (p4_local_dlcall)
     }
 }
 
+/* ----------------------- direct calls into lower layer ----------------- */
+
+/** lt_dlinit ( -- ior )
+ * initialiize library, usually open the program itself so that its
+ * handles can be found under "0"
+ */
+FCode (p4_lt_dlinit)
+{
+    FX_PUSH (p4_dlinit ());
+}
+
+/** lt_dlopenext ( name-ptr,len -- handle-ptr|0 )
+ * walk the searchpath for dlopen and try to open a binary module
+ * under the given name with the usual file extension for the 
+ * current system.
+ */
+FCode (p4_lt_dlopenext)
+{
+    SP[1] = (p4ucell) 
+        p4_dlopenext (p4_pocket_filename ((char*)(SP[1]), SP[0]));
+    FX_DROP;
+}
+
+/** lt_dlsym ( name-ptr,len handle -- symbol-addr|0)
+ * try to find the name in the binary module denoted by its handle
+ * .. if handle is null, use the main body of the program
+ */
+FCode (p4_lt_dlsym)
+{
+    SP[2] = (p4ucell)
+        p4_dlsym ((void*)(SP[0]), 
+                  p4_pocket_filename ((char*)(SP[2]), SP[1]));
+    FX_2DROP;
+}
+
+/** lt_dlclose ( handle -- ior )
+ * close handle that was returned by => lt_dlopenext
+ */
+FCode (p4_lt_dlcose)
+{
+    *SP = p4_dlclose ((void*)(*SP));
+}
+
+/** lt_dlerror ( -- z-string* )
+ * returns string describing the last dlerror as for => lt_dlopenext
+ * and => lt_dlsym
+ */
+FCode (p4_lt_dlerror)
+{
+    FX_PUSH (p4_dlerror ());
+}
+
 P4_LISTWORDS (dlfcn) =
 {
     P4_INTO ("FORTH", 0),
@@ -744,6 +811,11 @@ P4_LISTWORDS (dlfcn) =
     P4_FXco ("USELIBRARY",	p4_uselibrary),
 # endif
 
+    P4_FXco ("lt_dlinit",          p4_lt_dlinit),
+    P4_FXco ("lt_dlopenext",       p4_lt_dlopenext),
+    P4_FXco ("lt_dlsym",           p4_lt_dlsym),
+    P4_FXco ("lt_dlclose",         p4_lt_dlcose),
+    P4_FXco ("lt_dlerror",         p4_lt_dlerror),
 };
 P4_COUNTWORDS (dlfcn, "Dynamic-Loading of code modules");
 
