@@ -5,8 +5,8 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE            @(#) %derived_by: guidod %
- *  @version %version: bln_mpt1!5.20 %
- *    (%date_modified: Tue Oct 09 20:29:19 2001 %)
+ *  @version %version: bln_mpt1!5.22 %
+ *    (%date_modified: Mon Apr 08 20:17:44 2002 %)
  *
  *  @description
  *       This is a simple fullscreen FORTH block editor.
@@ -17,7 +17,7 @@
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  edit-ext.c~bln_mpt1!5.20:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  edit-ext.c~bln_mpt1!5.22:csrc:bln_12xx!1 % $";
 #endif
 
 #define _P4_SOURCE 1
@@ -159,12 +159,12 @@ static void edit_init (struct edit* set)
    }
 }
 
-FCode (p4_edit_forget)
-{
-    register struct edit* set = (struct edit*) WP_PFA[0];
+FCode_RT (p4_edit_forget)
+{   FX_USE_BODY_ADDR {
+    register struct edit* set = (struct edit*) FX_POP_BODY_ADDR[0];
     P4_note ("clean edit area");
     if (set->heap.editor) { free ((char*) set->editor); set->heap.editor = 0; }
-}
+}}
 
 FCode (p4_edit_init)
 {
