@@ -5,13 +5,13 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE            @(#) %derived_by: guidod %
- *  @version %version: bln_mpt1!1.19 %
- *    (%date_modified: Mon Sep 24 19:29:36 2001 %)
+ *  @version %version: bln_mpt1!1.48 %
+ *    (%date_modified: Wed Jul 24 11:37:14 2002 %)
  */
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  engine-sub.c~bln_mpt1!1.19:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  engine-sub.c~bln_mpt1!1.48:csrc:bln_12xx!1 % $";
 #endif
 
 #define _P4_SOURCE 1
@@ -38,14 +38,22 @@ static char* id __attribute__((unused)) =
 #include <pfe/term-sub.h>
 #include <pfe/_missing.h>
 #include <pfe/exception-sub.h>
+#include <pfe/chainlist-ext.h>
 
 #include <pfe/logging.h>
 
+FCode(p4_noop)
+{
+    /* well, nothing... */
+}
+
 
 /* ********************************************************************** 
  * inner and outer interpreter						  
  */
 
+#ifndef PFE_SBR_CALL_THREADING
+
 /* jumpbuf is a jmp_buf enanced by space to preserve variables that
  * may be stored in cpu register, would be the Forth VM
  */
@@ -53,7 +61,7 @@ typedef struct
 {
     jmp_buf jmp;
 # ifdef P4_REGRP
-    p4xt **rp;
+    p4xcode **rp;
 # endif
 # ifdef P4_REGSP
     p4cell *sp;
@@ -71,8 +79,8 @@ jumpbuf;
  * longjmp via (jumpbuf*) following inline
  * - purpose: stop the inner interpreter
  */
-FCode (p4_jump)	
-{
+FCode_XE (p4_jump)	
+{   FX_USE_CODE_ADDR {
     jumpbuf *buf = (jumpbuf *) *IP;
 
 # ifdef P4_REGRP		/* save global register variables */
@@ -90,7 +98,10 @@ FCode (p4_jump)
 # endif
 # endif
     longjmp (buf->jmp, 1);
-}
+    /*FX_USE_CODE_EXIT;*/
+}}
+
+#endif /* ! SBR_THREADING */
 
 /**
  * Run a forth word from within C-code
@@ -99,17 +110,33 @@ FCode (p4_jump)
 _export void
 p4_run_forth (p4xt xt)
 {
-    static p4code jump_p = PFX (p4_jump);
-    jumpbuf stop;
-    p4xt list[3];
+#if defined PFE_SBR_CALL_THREADING
+    p4_sbr_call (xt);
+    return;
+#else
 
+    jumpbuf stop;
+
+#  if ! defined PFE_CALL_THREADING
+    static p4code jump_p = PFX (p4_jump);
+    p4xcode list[3];
     list[0] = xt;
     list[1] = &jump_p;
-    list[2] = (p4xt) &stop;
+    list[2] = (p4xcode) &stop;
+#  else
+    /* sbr-stub, xt-code, xt-data, sbr-stub, jump-code, jump-data */
+    p4xcode list[6] /* = { 0,0,0,0,0,0 } */ ;
+    *p4_compile_xcode(
+	p4_compile_comma(
+	    list,
+	    xt),
+	PFX(p4_jump)) = (p4xcode) &stop;
+#  endif
+
     IP = list;
-# ifndef P4_WP_VIA_IP
+#  if !defined P4_WP_VIA_IP && !defined PFE_CALL_THREADING
     p4WP = *IP;
-# endif
+#  endif
 
     if (setjmp (stop.jmp))
     {
@@ -133,14 +160,19 @@ p4_run_forth (p4xt xt)
     /* next_loop */
     for (;;)
     {
-#     ifdef P4_WP_VIA_IP
-        p4xt w;
-#     define NEXT w = *IP++, (*w) ()	/* ip is register but p4WP isn't */
-#     else
-#     define NEXT p4WP = *IP++, (*p4WP) ()	
+#    ifdef PFE_CALL_THREADING
+#      define NVAR register p4code c;
+#      define NEXT c = *IP++, (c)()
+#    elif defined P4_WP_VIA_IP
+#      define NVAR register p4xt w;
+#      define NEXT w = *IP++, (*w) ()	/* ip is register but p4WP isn't */
+#    else
+#      define NVAR
+#      define NEXT p4WP = *IP++, (*p4WP) ()	
 				/* ip and p4WP are same: register or not */
-#     endif
+#    endif
 
+        NVAR;
 #     ifdef UNROLL_NEXT		/* USER-CONFIG: if it helps */
         NEXT; NEXT; NEXT; NEXT;	/* do a little loop unrolling for speed */
         NEXT; NEXT; NEXT; NEXT;
@@ -148,6 +180,8 @@ p4_run_forth (p4xt xt)
         NEXT;			/* on some machines it doesn't do any good */
 #     endif
     }
+
+#endif /* ! PFE_SBR_CALL_THREADING */
 }
 
 /**
@@ -164,11 +198,12 @@ p4_call (p4xt xt)
     printf ("%X/%X\n", IP, saved_ip);
     IP = saved_ip;
     printf ("%X/%X\n\n", IP, saved_ip);
-# else
-    p4xt *saved_ip = IP;
-    
+# elif !defined PFE_SBR_CALL_THREADING
+    p4xcode *saved_ip = IP;
     p4_run_forth (xt);
     IP = saved_ip;
+# else
+    p4_sbr_call (xt);
 # endif
 }
 
@@ -179,12 +214,42 @@ p4_call (p4xt xt)
 _export void
 p4_normal_execute (p4xt xt)
 {
-# ifdef P4_WP_VIA_IP
-    p4_call (xt);
-# else
-    p4WP = xt;
-    (*p4WP) ();
-# endif
+#  if defined PFE_SBR_CALL_THREADING /*FIXME: BODY / CODE ADDR needed? */
+    (*p4_to_code(xt))();
+#  else
+    p4_call(xt);
+#  endif
+}
+
+/**
+ * quick execute - unsafe and slow and simple
+ *
+ * use this routine for callbacks that might go through some forth
+ * colon-routines - that code shall not THROW or do some other nifty
+ * tricks with the return-stack or the inner interpreter. 
+ * Just simple things - use only for primitives or colon-routines,
+ * nothing curried with a DOES part in SBR-threading or sth. like that.
+ */
+_export void
+p4_simple_execute (p4xt xt)
+{
+#  if defined PFE_SBR_CALL_THREADING /*FIXME: BODY / CODE ADDR needed? */
+    (*p4_to_code(xt))();
+#  elif defined PFE_CALL_THREADING 
+    P4_REGIP_T ip = IP;
+    P4_REGRP_T rp = RP; 
+    p4xcode body = (p4xcode) P4_TO_BODY(xt);
+    IP = &body; /* fake the body-field, just in case it is needed */
+    (*p4_to_code(xt))();
+    while (RP < rp) { NVAR; NEXT; }
+    IP = ip;
+#  else /* ITC: */
+    P4_REGIP_T ip = IP;
+    P4_REGRP_T rp = RP; 
+    IP = & xt;
+    do { NVAR; NEXT; }  while (RP < rp);
+    IP = ip;
+#  endif
 }
 
 /* ================= INTERPRET =================== */
@@ -192,7 +257,7 @@ p4_normal_execute (p4xt xt)
 static p4ucell
 FXCode (p4_interpret_find) /* hereclean */
 {
-    register char *s;
+    register p4char *s;
     register p4xt xt;
 
     /* WORD-string is at HERE and at PFE.word.ptr / PFE.word.len */
@@ -202,10 +267,10 @@ FXCode (p4_interpret_find) /* hereclean */
     xt = p4_name_from (s);
     if (! STATE || *_FFA(s) & P4xIMMEDIATE)
     {
-	p4_call (xt);         /* execute it now */
-	FX (p4_Q_stack);      /* check stack */
+	p4_call (xt);           /* execute it now */
+	FX (p4_Q_stack);        /* check stack */
     }else{
-	FX_XCOMMA (xt);        /* comma token */
+	FX_COMPILE_COMMA (xt);  /* comma token */
     }
     return 1;
 }
@@ -259,6 +324,7 @@ FCode (p4_interpret)
 	/* PFE.interpret[1] = PFX (p4_interpret_smart); */
     }
 
+    PFE.last_here = PFE.dp;
     for (;;)
     {
     again:
@@ -275,7 +341,10 @@ FCode (p4_interpret)
             {
              default:
                  if (p4_next_line ())
+                 {
+                     PFE.last_here = PFE.dp;
                      continue;
+                 }
              case 0:
              case -1:
                  return;
@@ -298,14 +367,23 @@ FCode (p4_interpret)
 _export void
 p4_evaluate (char *p, int n)
 {
-    RP = (p4xt **) p4_save_input (RP);
+#  if !defined P4_RP_IN_VM
+    Iframe saved;
+    p4_link_saved_input (&saved);
+#  else
+    RP = (p4xcode **) p4_save_input (RP);
+#  endif
     SOURCE_ID = -1;
     BLK = 0;
     TIB = p;
     NUMBER_TIB = n;
     TO_IN = 0;
     FX (p4_interpret);
-    RP = (p4xt **) p4_restore_input (RP);
+#  if defined P4_RP_IN_VM
+    RP = (p4xcode **) p4_restore_input (RP);
+#  else
+    p4_unlink_saved_input (&saved);
+#  endif
 }
 
 /**
@@ -315,12 +393,24 @@ p4_include_file (p4_File *fid)
 {
     if (fid == NULL || fid->f == NULL)
         p4_throws (P4_ON_FILE_NEX, fid->name, 0);
-    RP = (p4xt **) p4_save_input (RP);
-    SOURCE_ID = (p4cell) fid;
-    BLK = 0;
-    TO_IN = 0;
-    FX (p4_interpret);
-    RP = (p4xt **) p4_restore_input (RP);
+    else
+    {
+#      if !defined P4_RP_IN_VM
+	Iframe saved;
+	p4_link_saved_input (&saved);
+#      else
+	RP = (p4xcode **) p4_save_input (RP);
+#      endif
+	SOURCE_ID = (p4cell) fid;
+	BLK = 0;
+	TO_IN = 0;
+	FX (p4_interpret);
+#      if defined P4_RP_IN_VM
+	RP = (p4xcode **) p4_restore_input (RP);
+#      else
+	p4_unlink_saved_input (&saved);
+#      endif
+    }
 }
 
 /**
@@ -383,7 +473,11 @@ p4_unnest_input (p4_Iframe *p)
          default:
              p4_close_file (SOURCE_FILE);
 	}
-        RP = (p4xt **) p4_restore_input (PFE.saved_input);
+#     if defined P4_RP_IN_VM
+        RP = (p4xcode **) p4_restore_input (PFE.saved_input);
+#     else
+	p4_unlink_saved_input (PFE.saved_input);
+#     endif
     }
 }
 
@@ -435,8 +529,10 @@ FCode (p4_ok)
 static void
 quit_system (P4_VOID)
 {
+#  ifdef P4_RP_IN_VM
     CSP = (p4cell*) RP;         /* come_back marker */
-    RP = p4_R0;			/* return stack */
+    RP = p4_R0;			/* return stack to its bottom */
+#  endif
     LP = NULL;			/* including all local variables */
     STATE = P4_FALSE;		/* interpreting now */
     PFE.cAtch = NULL;		/* and no exceptions caught */
@@ -452,11 +548,19 @@ abort_system (P4_VOID)
     SP = p4_S0;				/* stacks */
     if (PFE.abort[2]) { (PFE.abort[2]) (FX_VOID); } /* -> floating */
     if (PFE.abort[3]) { (PFE.abort[3]) (FX_VOID); } /* -> dstrings */
-    FX (p4_reset_order);		/* reset search order */
-    CURRENT = PFE.forth_wl;             /* initial CURRENT is FORTH-WORDLIST */
+    if (p4_RESET_ORDER)  { FX (p4_reset_order); }   /* reset search order */
     FX (p4_decimal);			/* number i/o base */
     FX (p4_standard_io);		/* disable i/o redirection */
     FX (p4_closeall_files);             /* close open filedescriptors */
+    if (PFE.dictlimit - PFE_MINIMAL_UNUSED > PFE.dp)
+        return;
+    else
+    {
+        P4_fail2 ("DICT OVER - reset HERE from %+i to %+i",
+                  PFE.dp - PFE.dict, PFE.last_here - PFE.dict);
+
+        PFE.dp = PFE.last_here;
+    }
 }
 
 FCode (p4_paren_abort)
@@ -478,6 +582,7 @@ p4_interpret_loop (P4_VOID)
          /* initialize */
      case 'A': /* do abort */
          abort_system (FX_VOID);
+         p4_redo_all_words (PFE.abort_wl);
          /* -> do quit */
      case 'Q': /* do quit */
          quit_system (FX_VOID);
@@ -487,8 +592,9 @@ p4_interpret_loop (P4_VOID)
          p4_unnest_input (NULL);
          for (;;)
          {
+             p4_do_all_words (PFE.prompt_wl);
              FX (p4_ok);
-             FX (p4_cr);		
+             FX (p4_cr);	
              FX (p4_query);		
              FX (p4_interpret);		
              FX (p4_Q_stack);	
@@ -516,13 +622,20 @@ cold_system (void)
 #  ifndef P4_NO_FP
     FP = p4_F0;
 #  endif
+#  ifdef P4_RP_IN_VM
     RP = p4_R0;
+#  endif
     TIB = PFE.tib;
     BASE = 10;
     p4_DPL = -1;
-    PRECISION = 6;
-    WORDL_FLAG = P4_opt.lower_case_on; /* implicitly enables HASHing */
-    LOWER_CASE_FN = P4_opt.lower_case_fn;
+    PRECISION = p4_search_option_value("precision",9, 6, PFE.set);
+    WORDL_FLAG = 0; /* implicitly enables HASHing */
+    if (p4_search_option_value("source-any-case",15, PFE_set.find_any_case,
+          PFE.set)) WORDL_FLAG |= WORDL_NOCASE;
+    if (p4_search_option_value("source-upper-case",17, PFE_set.upper_case_on,
+          PFE.set)) WORDL_FLAG |= WORDL_UPPER_CASE;
+    LOWER_CASE_FN = p4_search_option_value("lower-case-filenames",20, 
+      PFE_set.lower_case_fn, PFE.set);
     FLOAT_INPUT = P4_opt.float_input;
     PFE.local = (char (*)[32]) PFE.stack;
 
@@ -551,6 +664,8 @@ cold_system (void)
     /* Wipe the dictionary: */
     memset (PFE.dict, 0, (PFE.dictlimit - PFE.dict));
     p4_preload_only ();
+    if (! PFE.abort_wl)  PFE.abort_wl  = p4_new_wordlist (0);
+    if (! PFE.prompt_wl) PFE.prompt_wl = p4_new_wordlist (0);
     FX (p4_only_RT);
     {
         /* Defines the following default search order:
@@ -611,6 +726,7 @@ p4_boot_system (void)
     if (PFE.nr) { printf (" CPU%i ", PFE.nr); }
 
     /* Action of COLD ABORT and QUIT, but don't enter the interactive QUIT */
+    RESET_ORDER = P4_TRUE;
     cold_system ();
     abort_system ();
     quit_system ();
@@ -647,6 +763,21 @@ p4_boot_system (void)
 
     /*  read_help_index (PFE_PKGHELPDIR, "index"); */
 
+    /* According to the ANS Forth description, the order after BOOT must
+     * include the FORTH-WORDLIST, and the CURRENT definition-wordlist
+     * must be the FORTH-WORDLIST. Here we assume that the various LOADs
+     * before have kept atleast one occurence of FORTH-WORDLIST in the
+     * search-order but we explicitly set the CURRENT definition-wordlist
+     * Then we do DEFAULT-ORDER so it can pop up in a RESET-ORDER on ABORT
+     * BEWARE: a bootscript can arrange the items in the search-order but
+     * it can not arrange to set the CURRENT definitions-wordlist as well.
+     * Note that ONLY is always searched, so one can always get back at FORTH
+     * OTOH, in main-sub, the first include-file is loaded after boot_system
+     * so it can arrange for a different the DEFAULT-ORDER incl. CURRENT.
+     */
+    CURRENT = PFE.forth_wl;
+    FX (p4_default_order);
+
     FENCE = DP;
     LAST  = NULL;
 
