@@ -5,19 +5,19 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE             @(#) %derived_by: guidod %
- *  @version %version: bln_mpt1!5.14 %
- *    (%date_modified: Tue Oct 09 13:05:51 2001 %)
+ *  @version %version: bln_mpt1!5.25 %
+ *    (%date_modified: Mon Mar 04 8:30:00 2003 %)
  *
  *  @description
  *         The Optional Floating-Point Wordset is not usually
  *         used on embedded platforms. The PFE can be configured
- *         to even not allocate the seperate floating-point stack
+ *         to even not allocate the separate floating-point stack
  *         that most of the floating-point words refer to.
  */
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  floating-ext.c~bln_mpt1!5.14:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  floating-ext.c~bln_mpt1!5.25:csrc:bln_12xx!1 % $";
 #endif
 
 #define _P4_SOURCE 1
@@ -37,10 +37,59 @@ static char* id __attribute__((unused)) =
 #include <pfe/option-ext.h>
 #include <pfe/def-comp.h>
 #include <pfe/double-sub.h>
-#include <pfe/_missing.h>
+// #include <pfe/_missing.h>
 
 #define CELLBITS	BITSOF (p4cell)
 
+
+/* ------------------------------------------------------------------ 
+ * Exact comparison of raw floats.  The following is intended to
+ * capture the sizes listed for IEEE 754 by William Kahan,
+ * "Fclass: a Proposed Classification of Standard Floating-Point
+ * Operands", March 2, 2002.  He lists 4, >=6, 8, 10, 12, or 16
+ * 8-bit bytes.  We assume that 4, 8, 10, 12, or 16 might
+ * correspond to a double, and probably 4 could be omitted.  
+ * --KM&DNW 2Mar03
+ */
+#if PFE_SIZEOF_DOUBLE == PFE_SIZEOF_INT
+#  define EXACTLY_EQUAL(A,B)  ( *((int*) &(A)) == *((int*) &(B)) )
+#elif PFE_SIZEOF_DOUBLE == 2 * PFE_SIZEOF_INT
+#  define EXACTLY_EQUAL(A,B) \
+        ( *((int*) &(A)) == *((int*) &(B)) \
+       && *(((int*) &(A)) + 1) == *(((int*) &(B)) + 1) )
+#elif PFE_SIZEOF_DOUBLE == 2 * PFE_SIZEOF_INT + PFE_SIZEOF_SHORT
+#  define EXACTLY_EQUAL(A,B) \
+        ( *((int*) &(A)) == *((int*) &(B)) \
+       && *(((int*) &(A)) + 1) == *(((int*) &(B)) + 1) \
+       && (short)*(((int*) &(A)) + 2) == (short)*(((int*) &(B)) + 2) )
+#elif PFE_SIZEOF_DOUBLE == 3 * PFE_SIZEOF_INT
+#  define EXACTLY_EQUAL(A,B) \
+        ( *((int*) &(A)) == *((int*) &(B)) \
+       && *(((int*) &(A)) + 1) == *(((int*) &(B)) + 1) \
+       && *(((int*) &(A)) + 2) == *(((int*) &(B)) + 2) )
+#elif PFE_SIZEOF_DOUBLE == 4 * PFE_SIZEOF_INT
+#  define EXACTLY_EQUAL(A,B) \
+        ( *((int*) &(A)) == *((int*) &(B)) \
+       && *(((int*) &(A)) + 1) == *(((int*) &(B)) + 1) \
+       && *(((int*) &(A)) + 2) == *(((int*) &(B)) + 2) \
+       && *(((int*) &(A)) + 3) == *(((int*) &(B)) + 3) )
+#else
+#  define EXACTLY_EQUAL(A,B)  (memcmp (&(A), &(B), sizeof (double)) == 0)
+#  ifdef __GNUC__
+#  warning using memcmp() in p4_f_proximate()
+#  elif !defined _PFE_FLOATING_USING_MEMCMP
+#  error   using memcmp() in p4_f_proximate()
+#  endif
+#endif
+
+#ifndef USE_STRTOD	/* USER-CONFIG: */
+#define USE_STRTOD 1	/* most systems have good strtod */
+#endif
+
+#ifndef USE_SSCANF	/* USER-CONFIG: */
+#define USE_SSCANF 1	/* define this if you fully trust your scanf */
+#endif
+
 /* ------------------------------------------------------------------ 
  * static helper routines for missing functionality.
  */
@@ -94,7 +143,7 @@ p4_dfaligned (p4cell n)
 _export int
 p4_to_float (char *p, p4cell n, double *r)
 {
-# if defined USE_STRTOD		/* most systems have good strtod */
+# if USE_STRTOD		/* most systems have good strtod */
 
     char buf[80], *q;
   
@@ -244,11 +293,9 @@ p4_to_float (char *p, p4cell n, double *r)
 }
 
 
-/**
- */
-FCode (p4_d_f_align);
+FCode (p4_d_f_align); /* forward */
 
-#if defined USE_SSCANF		/* define this if you fully trust your scanf */
+#if USE_SSCANF		/* define this if you fully trust your scanf */
 
 /*
  * This is a working solution on most machines.
@@ -258,7 +305,7 @@ FCode (p4_d_f_align);
 FCode (p4_to_float)		
 {
     char *p, buf[80];
-    static char *fmt[] =
+    static const char *fmt[] =
     {
         "%lf%n %n%d%n$",
         "%lf%*1[DdEe]%n %n%d%n$",
@@ -541,6 +588,12 @@ FCode (p4_f_less_than)
     FP += 2;
 }
 
+/** ( f: a -- n,n )
+ * b is the integer representation of a
+ * 
+ * we use truncation towards zero.
+ * compare with =>"F>S" and its => "FROUND>S" / => "FTRUNC>S"
+ */
 FCode (p4_f_to_d)
 {
     double a, hi, lo;
@@ -563,14 +616,15 @@ FCode (p4_f_fetch)
     *--FP = *(double *) *SP++;
 }
 
-FCode (p4_f_constant_RT)
+FCode_RT (p4_f_constant_RT)
 {
-    *--FP = *(double *) p4_dfaligned ((p4cell) WP_PFA);
+    FX_USE_BODY_ADDR;
+    *--FP = *(double *) p4_dfaligned ((p4cell) FX_POP_BODY_ADDR);
 }
 
 FCode (p4_f_constant)
 {
-    FX_HEADER;
+    FX_RUNTIME_HEADER;
     FX_RUNTIME1 (p4_f_constant);
     FX (p4_d_f_align);
     FX_FCOMMA (*FP++);
@@ -607,9 +661,11 @@ p4_lit_float_SEE (p4xt* ip, char* p, p4_Semant* s)
     return ip;
 }
 
-FCode (p4_f_literal_execution)
+FCode_XE (p4_f_literal_execution)
 {
+    FX_USE_CODE_ADDR;
     *--FP= P4_POP_ (double, IP);
+    FX_USE_CODE_EXIT;
 }
 
 FCode (p4_f_literal)
@@ -667,9 +723,37 @@ FCode (p4_f_rot)
     FP[0] = h;
 }
 
+#ifndef FROUND_FLOOR                /* USER-CONFIG: */
+#define FROUND_FLOOR 0              /* FROUND identical with floor(fp+0.5) ? */
+#endif
+
 FCode (p4_f_round)
 {
-    *FP = floor (*FP + 0.5);
+#  if defined HAVE_RINT || defined PFE_HAVE_RINT
+    /* correct and fast */
+    *FP = rint (*FP);
+#  elif FROUND_FLOOR
+    /* incorrect but fast */
+    *FP = floor (*FP + 0.5); 
+#  else
+    /* correct but slow */
+    double whole, frac, offset;
+ 
+    frac = fabs(modf(*FP, &whole));
+    *FP = whole;
+    FX(p4_f_to_d);  /* execute F>D */
+    offset = (*SP < 0) ? -1. : 1.;
+    
+    if (*(SP+1) & 1)  /* check even or odd */
+    {
+	if (frac >= 0.5) whole += offset;
+    }
+    else
+    {
+	if (frac > 0.5) whole += offset;
+    }
+    *--FP = whole; SP += 2;  
+#  endif
 }
 
 FCode (p4_f_swap)
@@ -680,14 +764,15 @@ FCode (p4_f_swap)
     FP[0] = h;
 }
 
-FCode (p4_f_variable_RT)
+FCode_RT (p4_f_variable_RT)
 {
-    *--SP = p4_dfaligned ((p4cell) WP_PFA);
+    FX_USE_BODY_ADDR;
+    FX_PUSH_SP = p4_dfaligned ((p4cell) FX_POP_BODY_ADDR);
 }
 
 FCode (p4_f_variable)
 {
-    FX_HEADER;
+    FX_RUNTIME_HEADER;
     FX_RUNTIME1 (p4_f_variable);
     FX (p4_d_f_align);
     FX_FCOMMA (0.);
@@ -764,8 +849,7 @@ FCode (p4_f_dot)
 
 FCode (p4_f_abs)
 {
-    if (*FP < 0)
-        *FP = -*FP;
+  *FP = fabs (*FP);
 }
 
 FCode (p4_f_e_dot)			/* with help from Lennart Benshop */
@@ -801,23 +885,12 @@ FCode (p4_f_proximate)
     b = FP[1];
     c = FP[0];
     FP += 3;
-# if 0
-    SP--;
-    if (c > 0)
-        *SP = P4_FLAG (fabs (a - b) < c);
-    else if (c < 0)
-        *SP = P4_FLAG (fabs (a - b) < -c * (fabs (a) + fabs (b)));
-    else
-        *SP = P4_FLAG (memcmp (&a, &b, sizeof (double)) == 0);
-    
-# else
     *--SP = P4_FLAG
         (c > 0 
           ? fabs (a - b) < c 
           : c < 0 
           ? fabs (a - b) < -c * (fabs (a) + fabs (b))
-          : a == b);
-# endif
+          : EXACTLY_EQUAL (a, b));
 }
 
 FCode (p4_set_precision)
@@ -858,9 +931,17 @@ FCode (p4_f_atanh)	{ *FP = atanh (*FP); }
 FCode (p4_f_cos)	{ *FP = cos (*FP); }
 FCode (p4_f_cosh)	{ *FP = cosh (*FP); }
 FCode (p4_f_exp)	{ *FP = exp (*FP); }
+#if 1  /* ante C99 */
 FCode (p4_f_expm1)	{ *FP = exp (*FP) - 1.0; }
-FCode (p4_f_ln)	{ *FP = log (*FP); }
+#else  /* post C99 */
+FCode (p4_f_expm1)	{ *FP = expm1 (*FP); }
+#endif
+FCode (p4_f_ln)		{ *FP = log (*FP); }
+#if 1  /* ante C99 */
 FCode (p4_f_lnp1)	{ *FP = log (*FP + 1.0); }
+#else  /* post C99 */
+FCode (p4_f_lnp1)	{ *FP = log1p (*FP); }
+#endif
 FCode (p4_f_log)	{ *FP = log10 (*FP); }
 FCode (p4_f_sin)	{ *FP = sin (*FP); }
 FCode (p4_f_sincos)	{ --FP; FP [0] = cos (FP [1]); FP [1] = sin (FP [1]); }
@@ -922,14 +1003,14 @@ static FCode(abort_float)
 
 static int decompile_floating (char* nfa, p4xt xt)
 {
-    if (*xt == PFX (p4_f_constant_RT))          
+    if (*P4_TO_CODE(xt) == PFX (p4_f_constant_RT))          
     {
         p4_outf ("%g FCONSTANT ", 
           *(double *) p4_dfaligned ((p4cell) P4_TO_BODY (xt)));
         p4_dot_name (nfa);
         return 1;
     }
-    else if (*xt == PFX (p4_f_variable_RT))
+    else if (*P4_TO_CODE(xt) == PFX (p4_f_variable_RT))
     {
         p4_outf ("%g FVARIABLE ", 
           *(double *) p4_dfaligned ((p4cell) P4_TO_BODY (xt)));
@@ -939,17 +1020,27 @@ static int decompile_floating (char* nfa, p4xt xt)
     return 0;
 }
 
-static FCode(floating_deinit)
+#ifndef FLOATING_INTERPRET_SLOT       /* USER-CONFIG: */
+#define FLOATING_INTERPRET_SLOT 2     /* 1 == smart-ext / 2 == floating-ext */
+#endif
+
+static FCode_RT(floating_deinit)
 {
-    PFE.decompile[2] = 0;
-    PFE.interpret[2] = 0;
-    PFE.abort[2] = 0;
+    FX_USE_BODY_ADDR; 
+    FX_POP_BODY_ADDR_UNUSED;
+    PFE.decompile[FLOATING_INTERPRET_SLOT] = 0;
+    PFE.interpret[FLOATING_INTERPRET_SLOT] = 0;
+    PFE.abort[FLOATING_INTERPRET_SLOT] = 0;
 }
 
 #ifndef FLT_STACK_SIZE          /* USER-CONFIG: --fp-stack-size */
 #define	FLT_STACK_SIZE	0	/* 0 -> P4_KB*1024 / 16 */
 #endif
 
+#ifndef FLOATING_HEADROOM       /* USER-CONFIG: */
+#define FLOATING_HEADROOM 2     /* F-stack underflow does no harm for these */
+#endif
+
 static FCode(floating_init)
 {
     p4ucell flt_stack_size =
@@ -967,14 +1058,18 @@ static FCode(floating_init)
 			    (void**) &PFE.fstack, (void**) &PFE.f0)
 	) p4_throw (P4_ON_DICT_OVER); /** FIXME: no good idea to throw here */
 
+    PFE.f0 -= FLOATING_HEADROOM;
+
     FP = PFE.f0; /* same as abort_float above. Is that a rule? askmee */
 
-    PFE.interpret[2] = PFX (interpret_float);
-    PFE.abort[2] = PFX(abort_float);
-    PFE.decompile[2] = decompile_floating;
-    p4_forget_word ("deinit:floating:%i", 2, PFX(floating_deinit), 2);
+    PFE.interpret[FLOATING_INTERPRET_SLOT] = PFX (interpret_float);
+    PFE.abort[FLOATING_INTERPRET_SLOT] = PFX(abort_float);
+    PFE.decompile[FLOATING_INTERPRET_SLOT] = decompile_floating;
+    p4_forget_word ("deinit:floating:%i", FLOATING_INTERPRET_SLOT, 
+		    PFX(floating_deinit), FLOATING_INTERPRET_SLOT);
 }
 
+
 extern p4Words P4WORDS(floating_misc);
 
 P4_LISTWORDS (floating) =
@@ -1072,4 +1167,3 @@ P4_COUNTWORDS (floating, "Floating point + extensions");
  * c-file-style: "stroustrup"
  * End:
  */
-
