@@ -5,8 +5,8 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE              @(#) %derived_by: guidod %
- *  @version %version:  1.11 %
- *    (%date_modified:  Tue Oct 02 17:17:15 2001 %)
+ *  @version %version:  1.24 %
+ *    (%date_modified:  Mon Feb 24 20:24:13 2003 %)
  *  @description
  *    Implements header creation and navigation words including the
  *    ones from the forth-83 "experimental annex" targetting definition
@@ -17,7 +17,7 @@
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  header-ext.c~1.11:ascii:bln_mpt1!1 % $";
+"@(#) $Id: %full_filespec:  header-ext.c~1.24:ascii:bln_mpt1!1 % $";
 #endif
 
 #define _P4_SOURCE 1
@@ -108,7 +108,7 @@ FCode (p4_link_from)
  * converts a pointer to the link-field (LFA) to point
  * then to the corresponding name-field (CFA) - this one is one of
  * the slowest operation available. One should always use the inverse
- * operation => "N>LINK" and cache an older value if that is needed.
+ * operation =>"N>LINK" and cache an older value if that is needed.
  * Some words might be linked but they do not have a name-field (just
  * the other fields) but this word can not detect that and will try to look
  * into the bits of the dictionary anway in the assumption that there is 
@@ -180,7 +180,7 @@ FCode (p4_ffa_from)
 
 /** NAME>STRING        ( name-token -- str-ptr str-len )
  * convert a name-token into a string-span, used to detect the
- * name for a word and print it. The word => ID. can be defined as
+ * name for a word and print it. The word =>"ID." can be defined as
  : ID. NAME>STRING TYPE ;
  * the implementation of => NAME>STRING depends on the header layout
  * that is defined during the configuration of the forth system.
@@ -195,7 +195,7 @@ FCode (p4_ffa_from)
 FCode (p4_name_to_string)
 {
     FX_1ROOM;
-    SP[0] = NFACNT(*((p4char*)SP[1])++);
+    SP[0] = NFACNT(*P4_VAR(p4char*,SP[1])++);
 }
 
 /** HEADER, ( str-ptr str-len -- )
@@ -204,21 +204,20 @@ FCode (p4_name_to_string)
  */
 FCode (p4_header_comma)  
 {
-    extern FCode (p4_var_RT);
     p4_header_comma ((p4char*)SP[1], (int)SP[0], CURRENT);
     FX_2DROP;
 }
 
 /** $HEADER ( bstring -- )
  * => CREATE a new header in the dictionary from the given string
- * with the variable runtime (see =>"HEADER,")
+ * with the variable runtime (see =>"HEADER," and =>"CREATE:")
  usage: : VARIABLE  BL WORD $HEADER ;
  */
 FCode (p4_str_header)  
 {
-    extern FCode (p4_var_RT);
-    p4_make_header (PFX(p4_var_RT), 0, (p4char*)SP[1], (int)SP[0], CURRENT);
-    FX_2DROP;
+    p4_header_comma (1+ *(p4char**)SP,  **(p4char**)SP, CURRENT);
+    FX_RUNTIME1 (p4_variable);
+    FX_DROP;
 }
 
 /** LATEST ( -- nfa )
@@ -342,16 +341,18 @@ FCode (p4_defer_RT)
  */
 FCode (p4_defer)
 {
-    FX_HEADER;
+    FX_RUNTIME_HEADER;
     FX_RUNTIME1 (p4_defer);
     FX_XCOMMA (0); /* <-- leave it blank (may become chain-link later) */
     FX_XCOMMA (0); /* <-- put XT here in fig-mode */
 }
 P4RUNTIME1(p4_defer, p4_defer_RT);
     
-FCode (p4_is_execution)
+FCode_XE (p4_is_execution)
 {
+    FX_USE_CODE_ADDR;
     * (p4xt*) P4_TO_DOES_BODY(*IP++) = (p4xt) FX_POP;
+    FX_USE_CODE_EXIT;
 }
 
 /** IS ( xt-value [word] -- )
@@ -427,7 +428,7 @@ FCode (p4_obsoleted_RT)
  * even an execution token), instead it gets resolved at compile time.
  * In theory, you can try to => FIND the name of the => SYNONYM but as
  * soon as you apply =>"NAME>" the execution token of the end-point is
- * returned. This has also the effect that using the inverse => ">NAME"
+ * returned. This has also the effect that using the inverse =>">NAME"
  * operation will result in the name-token of the other name. 
 
    SYNONYM CREATE <BUILDS ( like it is in ANS Forth )
@@ -443,13 +444,13 @@ FCode (p4_obsoleted_RT)
  */
 FCode (p4_synonym)
 {
-    FX_HEADER;
+    FX_RUNTIME_HEADER;
     FX_RUNTIME1 (p4_synonym);
     {
 	register p4char* nfa = p4_tick_nfa ();
 	if (*NFA2FF(nfa) & P4xIMMEDIATE) 
 	    FX_IMMEDIATE;
-	FX_COMMA (p4_name_from (nfa));
+	FX_XCOMMA (p4_name_from (nfa));
     }
 }
 P4RUNTIME1(p4_synonym, p4_synonym_RT);
@@ -458,12 +459,12 @@ P4RUNTIME1(p4_synonym, p4_synonym_RT);
  * same as => SYNONYM but on the first use an error message will be
  * displayed on both the screen and the sys-log.
  */
-FCode (p4_synonym_obsoleted)
+FCode (p4_obsoleted)
 {
     FX (p4_synonym);
-    *p4_name_from(LAST) = p4_obsoleted_RT_;
+    P4_XT_VALUE(p4_name_from(LAST)) = FX_GET_RT (p4_obsoleted);
 }
-P4RUNTIME1(p4_synonym_obsoleted, p4_obsoleted_RT);
+P4RUNTIME1(p4_obsoleted, p4_obsoleted_RT);
 
 P4_LISTWORDS (header) =
 {
@@ -489,6 +490,8 @@ P4_LISTWORDS (header) =
     P4_FXco ("NAME-FLAGS!",		p4_name_flags_store),
 
     P4_FXco ("HEADER,",			p4_header_comma),
+    P4_FXco ("$HEADER",		        p4_str_header),
+    P4_xOLD ("HEADER",                  "$HEADER"),
     P4_FXco ("SMUDGE",			p4_smudge),
     P4_FXco ("REVEAL",			p4_reveal),
     P4_IXco ("RECURSIVE",		p4_reveal),
@@ -503,7 +506,7 @@ P4_LISTWORDS (header) =
     P4_SXco ("IS",			p4_is),
     P4_FXco ("BEHAVIOR",		p4_behavior),
     P4_RTco ("SYNONYM",			p4_synonym),
-    P4_RTco ("SYNONYM-OBSOLETED",	p4_synonym_obsoleted),
+    P4_RTco ("SYNONYM-OBSOLETED",	p4_obsoleted),
 
     P4_INTO ("ENVIRONMENT", 0 ),
     P4_OCON ("HEADER-EXT", 1983),
