@@ -5,15 +5,23 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE              @(#) %derived_by: guidod %
- *  @version %version: 1.11 %
- *    (%date_modified: Fri Mar 27 12:00:00 2001 %)
+ *  @version %version: bln_mpt1!1.13 %
+ *    (%date_modified: Mon Mar 25 17:02:44 2002 %)
  *
  *  @description
  *  Process command line, get memory and start up the interpret loop of PFE
  */
 /*@{*/
 
+#include "pfe-config.h"
+
+#if defined PFE_WITH_STATIC_REGS
+#include "main-alloc.c"
+#elif defined PFE_WITH_STATIC_DICT
+#include "main-static.c"
+#else
 #include "main-stdc.c"
+#endif
 
 /*@}*/
 
