@@ -1,21 +1,23 @@
 #ifndef __PFE_MAIN_K12_H
-#define __PFE_MAIN_K12_H "%full_filespec: main-k12.h~5.3:incl:bln_12xx!2 %"
+#define __PFE_MAIN_K12_H "%full_filespec: main-k12.h~bln_mpt1!32.55:incl:bln_12xx!2 %"
 /** @name main-k12.h
  *
  *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
  *
  *  @author Tektronix CTE               @(#) %derived_by: guidod %
- *  @version %version: 5.3 %
- *    (%date_modified: Mon Mar 12 10:33:39 2001 %)
+ *  @version %version: bln_mpt1!32.55 %
+ *    (%date_modified: Tue Oct 22 15:13:21 2002 %)
  */
 /*@{ */
 
 #include <pfe/term-k12.h>
 
+typedef char _p4_msc_string_buffer_t[256];
+
 typedef struct                  /* FIXME: Later deletion planned! */
 {
     int msc_ints[16];
-    char msc_strings[81][16];
+    _p4_msc_string_buffer_t msc_strings[16];
     int msc_timers[64];
 } _p4_msc_var_buf_t;
 
@@ -35,6 +37,7 @@ extern k12_emu_table_t   p4_emu_table;
 
 #define P4_K12_EMUL(P) ((p4_emu_t*)((P)->priv))
 
+extern char* k12_pfe_sap_table[];
+
 /*@}*/
 #endif
-
