@@ -5,8 +5,8 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE              @(#) %derived_by: guidod %
- *  @version %version: bln_mpt1!5.32 %
- *    (%date_modified: Tue Oct 02 21:48:21 2001 %)
+ *  @version %version: bln_mpt1!5.49 %
+ *    (%date_modified: Wed Aug 14 16:10:36 2002 %)
  *
  *  @description
  *  Process command line, get memory and start up the interpret loop of PFE
@@ -14,7 +14,7 @@
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  main-sub.c~bln_mpt1!5.32:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  main-sub.c~bln_mpt1!5.49:csrc:bln_12xx!1 % $";
 #endif
 
 #define	_P4_SOURCE 1
@@ -51,175 +51,9 @@ static char* id __attribute__((unused)) =
 #define _export
 # include <pfe/incl-sub.h>
 # include <pfe/def-types.h>
-
-  /* and a few internals */
-# define _p4_thread_save() (p4_thread_save(p4_main_threadP))
-# define _p4_thread_load() (p4_thread_load(p4_main_threadP))
-
-# ifdef P4_REGTH
-# define _p4_thread_local() (p4TH = p4_main_threadP)
-# define _p4_thread_ENTER_(arg_threadP) \
-   	register p4_threadP _p4_thread_SAVED_ = p4TH; \
-                        	p4TH = arg_threadP;
-# define _p4_thread_LEAVE_() p4TH = _p4_thread_SAVED_;
-# else
-# define _p4_thread_local()
-# define _p4_thread_ENTER_(ignored)
-# define _p4_thread_LEAVE_()
-# endif
 #endif
 
-/************************************************************************/
-/* physical instance of the global system variable:                     */
-/************************************************************************/
-
-#ifndef P4_REGTH
-p4_threadP p4TH;
-#endif
-
-/** two helper functions (thread_save and thead_load)
- * that can be called from non-pfe compiled context
- */
-_export void 
-p4_thread_save (p4_threadP p)
-{
-# ifdef P4_REGIP
-    p->ip = p4IP;
-# endif
-  
-# ifdef P4_REGW
-    p->wp = p4WP;
-# endif
-  
-# ifdef P4_REGSP
-    p->sp = p4SP;
-# endif
-  
-# ifdef P4_REGRP
-    p->rp = p4RP;
-# endif
-  
-# ifdef P4_REGLP
-    p->lp = p4LP;
-# endif
-
-# ifndef P4_NO_FP  
-# ifdef P4_REGFP
-    p->fp = p4FP;
-# endif
-# endif
-}
-
-_export void 
-p4_thread_load (p4_threadP p)
-{
-    p4TH = p; /* either a global variable or a register */
-
-# ifdef P4_REGIP
-    p4IP = p->ip;
-# endif
-  
-# ifdef P4_REGW
-    p4WP = p->wp;
-# endif
-  
-# ifdef P4_REGSP
-    p4SP = p->sp;
-# endif
-  
-# ifdef P4_REGRP
-    p4RP = p->rp;
-# endif
-  
-# ifdef P4_REGLP
-    p4LP = p->lp;
-# endif
-
-# ifndef P4_NO_FP  
-# ifdef P4_REGFP
-    p4FP = p->fp;
-# endif
-# endif
-}
-
-typedef struct p4_thread_VM p4_thread_VM;
-struct p4_thread_VM
-{
-    p4xt* ip;    /* the instruction pointer */
-    p4xt wp;     /* the word pointer */
-    p4cell* sp;  /* the parameter stack */
-    p4xt** rp;   /* the return stack */
-    p4cell* lp;  /* the locals pointer */
-    double* fp;  /* the floating stack */
-    p4_threadP th; /* the thread pointer */
-};
-
-inline void p4_thread_save_VM (p4_thread_VM* p)
-{
-# ifdef P4_REGTH
-    p->th = p4TH;
-# endif
-  
-# ifdef P4_REGIP
-    p->ip = p4IP;
-# endif
-  
-# ifdef P4_REGW
-    p->wp = p4WP;
-# endif
-  
-# ifdef P4_REGSP
-    p->sp = p4SP;
-# endif
-  
-# ifdef P4_REGRP
-    p->rp = p4RP;
-# endif
-  
-# ifdef P4_REGLP
-    p->lp = p4LP;
-# endif
-
-# ifndef P4_NO_FP  
-# ifdef P4_REGFP
-    p->fp = p4FP;
-# endif
-# endif
-}
-
-inline void p4_thread_load_VM (p4_thread_VM* p)
-{
-# ifdef P4_REGTH
-    p4TH = p->th;
-# endif
-
-# ifdef P4_REGIP
-    p4IP = p->ip;
-# endif
-  
-# ifdef P4_REGW
-    p4WP = p->wp;
-# endif
-  
-# ifdef P4_REGSP
-    p4SP = p->sp;
-# endif
-  
-# ifdef P4_REGRP
-    p4RP = p->rp;
-# endif
-  
-# ifdef P4_REGLP
-    p4LP = p->lp;
-# endif
-
-# ifndef P4_NO_FP  
-# ifdef P4_REGFP
-    p4FP = p->fp;
-# endif
-# endif
-}
-
+#include <pfe/def-restore.h>
 
 /************************************************************************/
 /* Analyze command line options:                                        */
@@ -228,8 +62,11 @@ inline void p4_thread_load_VM (p4_thread_VM* p)
 #ifndef CAPS_ON                 /* USER-CONFIG: */
 #define	CAPS_ON		0	/* do you like (faked) caps lock by default? */
 #endif
+#ifndef UPPER_CASE_ON           /* USER-CONFIG: */
+#define	UPPER_CASE_ON	1	/* allow "dup" to find "DUP" by default */
+#endif
 #ifndef LOWER_CASE_ON           /* USER-CONFIG: */
-#define	LOWER_CASE_ON	1	/* allow "dup" to find "DUP" by default */
+#define	LOWER_CASE_ON	1	/* allow "Dup" to find "dup" by default */
 #endif
 #ifndef LWRCASE_FN_ON           /* USER-CONFIG: */
 #define LWRCASE_FN_ON	1	/* convert file names to lower case? */
@@ -254,6 +91,18 @@ inline void p4_thread_load_VM (p4_thread_VM* p)
 #define	RET_STACK_SIZE	0	/* 0 -> P4_KB*1024 / 64 + 256 */
 #endif
 
+#ifndef ORDER_LEN               /* USER-CONFIG: */
+#define ORDER_LEN 64            /* maximum wordlists in search order */
+#endif
+#ifndef HISTORY_SIZE	        /* USER-CONFIG: */
+#define HISTORY_SIZE	0x1000  /* size of command line history buffer */
+#endif
+
+#ifdef _K12_SOURCE
+#undef  LOWER_CASE_ON
+#define LOWER_CASE_ON 1
+#endif
+
 /**
  * fill the session struct with precompiled options
  */
@@ -269,8 +118,9 @@ p4_SetOptionsDefault(p4_sessionP set, int len)
         set->isnotatty = 0;
         set->stdio = 0;
         set->caps_on = CAPS_ON;
-        set->lower_case_on = LOWER_CASE_ON;
+        set->find_any_case = LOWER_CASE_ON;
         set->lower_case_fn = LWRCASE_FN_ON;
+        set->upper_case_on = UPPER_CASE_ON;
 #       ifndef P4_NO_FP
         set->float_input = FLOAT_INPUT_ON;
 #       else
@@ -352,6 +202,7 @@ p4_SetOptionsDefault(p4_sessionP set, int len)
 static const char* help_options[] = {
     ">> Usage: %s [-#bcdefhklrsv] [file] [args..]",
     "-B --prefix DIR       \t prefix installation path to be used",  
+    "-C --case-sensitive   \t turn on case-sensitive matching (no upper)",
     "-c --caps-lock        \t turn on CAPS lock",
     "-e --evaluate         \t bootcommand to evaluate",
     "-f --max-files        \t maximum N simultaneously open files",
@@ -427,7 +278,7 @@ help_print (p4_sessionP set, FILE* f)
 	    break;
 	case 'B': fprintf(f, "%s [%s]", *p, set->prefix ? set->prefix : "." );
 	    break;
-	case 'C': fprintf(f, "%s [%i]", *p, set->cpus); 
+	case 'C': fprintf(f, "%s [%s]", *p, set->upper_case_on ? "OFF":"ON"); 
 	    break;
 	case 'c': fprintf(f, "%s [%s]", *p, set->caps_on ? "ON":"OFF"); 
 	    break;
@@ -439,7 +290,7 @@ help_print (p4_sessionP set, FILE* f)
 	    break;
 	case 'k': fprintf(f, "%s [%d K]", *p, (int) set->total_size >> 10);
 	    break;
-	case 'l': fprintf(f, "%s [%s]", *p, set->lower_case_on ? "ON":"OFF");
+	case 'l': fprintf(f, "%s [%s]", *p, set->find_any_case ? "ON":"OFF");
              break;
 	case 'p': fprintf(f, "%s [%d]", *p, (int) set->pockets);
 	    break;
@@ -549,17 +400,18 @@ p4_AddOptions (p4_sessionP set, int argc, char* argv[])
 	     * Simple flag options can be -x or -x- to turn them off.
 	     * these can be combined into a single option.
 	     */
-	case 'c': set->caps_on = flag;       continue;
-	case 'l': set->lower_case_on = flag; continue;
-	case 'F': set->lower_case_fn = flag; continue;
-	case 'G': set->float_input = flag;   continue;
-	case 'L': set->license = flag;	      continue;
-	case 'W': set->warranty = flag;      continue;
-	case 'q': set->quiet = flag;         continue;
-	case 'v': set->verbose = flag;       continue;
-	case 'P': set->stdio = flag;         continue;
-	case 'y': set->bye = flag;           continue;
-	case '!': set->debug = flag;         continue;
+	case 'c': set->caps_on = flag;         continue;
+	case 'C': set->upper_case_on = ! flag; continue;
+	case 'l': set->find_any_case = flag;   continue; /* depracated */
+	case 'F': set->lower_case_fn = flag;   continue;
+	case 'G': set->float_input = flag;     continue;
+	case 'L': set->license = flag;	       continue;
+	case 'W': set->warranty = flag;        continue;
+	case 'q': set->quiet = flag;           continue;
+	case 'v': set->verbose = flag;         continue;
+	case 'P': set->stdio = flag;           continue;
+	case 'y': set->bye = flag;             continue;
+	case '!': set->debug = flag;           continue;
 
              /*
               * Other options have values either following 
@@ -588,7 +440,7 @@ p4_AddOptions (p4_sessionP set, int argc, char* argv[])
 	    char* p;
 	    static const char delimstr[2] = { PFE_PATH_DELIMITER, '\0' };
 
-	    p = malloc (strlen(set->incpaths) + 1 + strlen(val));
+	    p = malloc (strlen(set->incpaths) + 1 + strlen(val) + 1);
 	    if (p) { 
 		strcpy (p, set->incpaths);
 		strcat (p, delimstr);
@@ -597,7 +449,7 @@ p4_AddOptions (p4_sessionP set, int argc, char* argv[])
 		set->incpaths = p; set->heap.incpaths = 1;
 	    }
 
-	    p = malloc (strlen(set->blkpaths) + 1 + strlen(val));
+	    p = malloc (strlen(set->blkpaths) + 1 + strlen(val) + 1);
 	    if (p) {
 		strcpy (p, set->blkpaths);
 		strcat (p, delimstr);
@@ -608,6 +460,7 @@ p4_AddOptions (p4_sessionP set, int argc, char* argv[])
 		
 	    i+=s; continue;
 	}
+#       ifdef __move_cpus_code_to_forth_vm_init
 	case 'C':
 	{  
 	    register int cpus = atoi(val);
@@ -618,6 +471,7 @@ p4_AddOptions (p4_sessionP set, int argc, char* argv[])
 	    }
 	    i+=s; continue;
 	}
+#       endif 
         case '?': help_print (set, stdout);  return 1; continue;
 	default:  
         {
@@ -760,6 +614,110 @@ p4_FreeOptions (int returncode, p4_sessionP set)
     return returncode;
 }
 
+/**
+ * set prelinked-modules-table
+ */
+_export int
+p4_SetModules (p4_sessionP set, p4Words* modules)
+{
+    set->modules = modules;
+    return 0;
+}
+
+/************************************************************************/
+/* physical instance of the global system variable:                     */
+/************************************************************************/
+
+#ifndef _export
+#define _export
+# ifndef P4_REGTH
+#  ifndef PFE_WITH_STATIC_REGS
+/* */ extern p4_threadP p4TH;
+#  else
+/* */ extern struct p4_Thread  p4_reg;
+/* */ extern struct p4_Session p4_opt;
+#  endif
+# endif
+#endif
+
+#ifndef P4_REGTH
+# ifndef PFE_WITH_STATIC_REGS
+/*export*/ p4_threadP p4TH;
+# else
+/*export*/ struct p4_Thread  p4_reg;
+/*export*/ struct p4_Session p4_opt;
+static  char allocated_p4_reg = 0;
+static  char allocated_p4_opt = 0;
+# endif
+#endif
+
+_export p4_sessionP
+p4_NewSessionOptions (int extra)
+{
+#  ifdef PFE_WITH_STATIC_REGS
+    if (allocated_p4_opt)
+        return 0;
+    p4_SetOptionsDefault (&p4_opt, sizeof(p4_opt));
+    allocated_p4_opt = 1;
+    return &p4_opt;
+#  else
+    p4_sessionP ptr = malloc (sizeof(*ptr)+extra);
+    p4_SetOptionsDefault (ptr, sizeof(*ptr)+extra);
+    return ptr;
+#  endif
+}
+
+
+_export p4_threadP
+p4_NewThreadOptions (p4_sessionP set)
+{
+#  ifdef PFE_WITH_STATIC_REGS
+    if (allocated_p4_reg)
+        return 0;
+    p4_reg.set = set;
+    allocated_p4_reg = 1;
+    return &p4_reg;
+#  else
+    p4_threadP ptr = malloc (sizeof(*ptr));
+    memset (ptr, 0, sizeof(*ptr));
+    ptr->set = set;
+    return ptr;
+#  endif
+}
+
+_export p4_threadP
+p4_SetThreadOf(p4_threadP ptr, p4_sessionP set)
+{
+    if (! ptr) return ptr;
+    memset (ptr, 0, sizeof (*ptr));
+    ptr->set = set;
+    return ptr;
+}
+
+_export void
+p4_FreeSessionPtr (p4_sessionP ptr)
+{
+#  ifdef PFE_WITH_STATIC_REGS
+    if (ptr != &p4_opt)
+        return 1;
+    return ((allocated_p4_opt = 0));
+#  else
+    if (ptr) free (ptr);
+#  endif
+}
+
+_export void
+p4_FreeThreadPtr (p4_threadP ptr)
+{
+#  ifdef PFE_WITH_STATIC_REGS
+    if (ptr != &p4_reg)
+        return 1;
+    return ((allocated_p4_reg = 0));
+#  else
+    if (ptr) free (ptr);
+#  endif
+}
+
 /************************************************************************/
 /* Initialize memory map:                                               */
 /************************************************************************/
@@ -793,55 +751,7 @@ init_accept_lined (void)
     PFE.accept_lined.caps = PFE_set.caps_on != 0;
 }
 
-/* 
- * Allocates all memory areas in a continuous buffer at p with given size. 
- */
-static void
-dict_allocate (void *p, p4ucell size)
-{
-    void *tmp;
-    typedef char pock_t[POCKET_SIZE];
-
-    PFE.dict = p;
-    PFE.dictlimit = PFE.dict + size;
-
-    if (! PFE_set.stack_size)
-    	PFE_set.stack_size = STACK_SIZE 
-            ? STACK_SIZE     
-	    : (PFE_set.total_size / 32 + 256)  / sizeof(p4cell);
-    if (! PFE_set.ret_stack_size)
-        PFE_set.ret_stack_size = RET_STACK_SIZE  
-            ? RET_STACK_SIZE 
-	    : (PFE_set.total_size / 64 + 256) / sizeof(p4cell);
-
-    p4_dict_allocate (PFE_set.max_files+3, sizeof(File), PFE_ALIGNOF_CELL,
-		      (void**) & PFE.files, (void**) & PFE.files_top);
-
-    p4_dict_allocate (TIB_SIZE, sizeof(char), sizeof(char),
-		      (void**) & PFE.tib, (void**) & PFE.tib_end);
-
-    p4_dict_allocate (HISTORY_SIZE, sizeof(char), sizeof(char),
-		      (void**) & PFE.history, (void**) & PFE.history_top);
-
-    p4_dict_allocate (PFE_set.ret_stack_size, sizeof(p4xt*), PFE_ALIGNOF_CELL,
-		      (void**) & PFE.rstack, (void**) & PFE.r0);
-    
-    p4_dict_allocate (PFE_set.stack_size, sizeof(p4cell),  PFE_ALIGNOF_CELL,
-		      (void**) & PFE.stack, (void**) & PFE.s0);
-
-    p4_dict_allocate (PFE_set.pockets, sizeof(pock_t), sizeof(char),
-		      (void**) & PFE.pockets, & tmp );
-
-    if (PFE.dictlimit < PFE.dict + MIN_PAD + MIN_HOLD + 0x4000)
-    {
-	P4_fatal ("impossible memory map");
-	PFE.exitcode = 3;
-	p4_longjmp_exit ();
-    }
-
-    init_accept_lined ();
-}
-
+typedef char pock_t[POCKET_SIZE];
 /************************************************************************/
 /* Here's main()                                                        */
 /************************************************************************/
@@ -864,7 +774,16 @@ p4_main (p4_threadP th)
     extern int taskIdSelf ();
     taskVarAdd (taskIdSelf (), (int*) &p4_main_threadP);
 #  endif
-    p4TH = p4_main_threadP = th;  
+    p4_main_threadP = th;  
+
+#  ifdef PFE_WITH_STATIC_REGS
+#  define p4_main_threadP_TO_p4TH 
+#  else
+#  define p4_main_threadP_TO_p4TH \
+   p4TH = p4_main_threadP
+#  endif
+
+    p4_main_threadP_TO_p4TH;
 
 #  ifdef PFE_HAVE_LOCALE_H
     setlocale (LC_ALL, "C");
@@ -878,6 +797,13 @@ p4_main (p4_threadP th)
     case 'A':
     case 'Q':
 	P4_fatal ("Fatal Run Error");
+        { 
+            extern FCode(p4_come_back); /*:debug-ext:*/ 
+#         ifdef P4_RP_IN_VM
+            if (p4_R0) RP = p4_R0; /* quit_system */
+            FX (p4_come_back); 
+#         endif
+        }
 	p4_atexit_cleanup ();
 	return -1;
     case 'X':
@@ -886,13 +812,15 @@ p4_main (p4_threadP th)
 	return PFE.exitcode;
     }
 
-    p4TH = p4_main_threadP; 
+    /* _______________ terminal settings _____________ */
+
+    p4_main_threadP_TO_p4TH; 
 
 #  if !defined __WATCOMC__
     if (! isatty (STDIN_FILENO))
         PFE_set.stdio = 1;
 #  endif
-
+    
     if (PFE_set.stdio)
         PFE_set.isnotatty = P4_TTY_ISPIPE;
     else
@@ -937,16 +865,16 @@ p4_main (p4_threadP th)
 	    p4_outs (p4_license_string ());
 	if (PFE_set.warranty)
 	    p4_outs (p4_warranty_string ());
-
+        
 	if (! PFE_set.bye)
 	{
 	    if (! PFE_set.license || ! PFE_set.warranty)
 		p4_outs ("\n\nPlease enter LICENSE and WARRANTY. ");
 	    else
 		p4_outs ("\n\nHi there, enjoy Forth! ");
-
+            
 #         ifndef _K12_SOURCE /* BYE does'nt make sense in an embedded system */
-		p4_outs ("- To quit say BYE.\n");
+                p4_outs ("- To quit say BYE.\n");
 #         else
 		p4_outs ("- To restart say COLD.\n");
 #         endif /* _K12_SOURCE */
@@ -956,11 +884,13 @@ p4_main (p4_threadP th)
         PFE.rows = PFE_set.rows;
     if (PFE.cols == 0)
         PFE.cols = PFE_set.cols;
-    
+
     p4TH->atexit_cleanup = &p4_atexit_cleanup;
     
+    /* _______________ dictionary block __________________ */
+    
 # ifdef USE_MMAP
-    if ((s = p4_search_option_string ("map-file", 8, 0, PFE.set))
+    if ((s = p4_search_option_string ("map-file", 8, 0, PFE.set)))
     {
         p4ucell l = p4_search_option_value ("map-base", 8, 0, PFE.set);
 	PFE.mapfile_fd = p4_mmap_creat (s, l, PFE_set.total_size);
@@ -999,22 +929,77 @@ p4_main (p4_threadP th)
             PFE_set.total_size = total_size;
         }
     }
+
+    /* ________________ initialize _____________ */
+
+    PFE.dict = PFE_MEM;
+    PFE.dictlimit = PFE.dict + PFE_set.total_size;
+
+    p4_dict_allocate (PFE_set.pockets, sizeof(pock_t), sizeof(char),
+                      (void**) & PFE.pockets, 0 );
+
+    PFE_set.history_size = 
+        p4_search_option_value ("/history", 8, HISTORY_SIZE, PFE.set);
+    p4_dict_allocate (PFE_set.history_size, sizeof(char), sizeof(char),
+                      (void**) & PFE.history, (void**) & PFE.history_top);
+
+    p4_dict_allocate (PFE_set.max_files+3, sizeof(File), PFE_ALIGNOF_CELL,
+                      (void**) & PFE.files, (void**) & PFE.files_top);
+
+    p4_dict_allocate (TIB_SIZE, sizeof(char), sizeof(char),
+                      (void**) & PFE.tib, (void**) & PFE.tib_end);
+
+    if (! PFE_set.ret_stack_size)
+        PFE_set.ret_stack_size = 
+            p4_search_option_value (
+                "return-stack-cells", 18,
+                RET_STACK_SIZE ? RET_STACK_SIZE 
+                : (PFE_set.total_size / 64 + 256) / sizeof(p4cell), PFE.set);
+    p4_dict_allocate (PFE_set.ret_stack_size, sizeof(p4xt*), 
+                      PFE_ALIGNOF_CELL,
+                      (void**) & PFE.rstack, (void**) & PFE.r0);
     
-    dict_allocate ((char *) PFE_MEM, PFE_set.total_size);
+    if (! PFE_set.stack_size)
+        PFE_set.stack_size = 
+            p4_search_option_value (
+                "stack-cells", 11,
+                STACK_SIZE ? STACK_SIZE
+                : (PFE_set.total_size / 32 + 256)  / sizeof(p4cell), PFE.set);
+    p4_dict_allocate (PFE_set.stack_size, sizeof(p4cell),  
+                      PFE_ALIGNOF_CELL,
+                      (void**) & PFE.stack, (void**) & PFE.s0);
+
+    PFE_set.wordlists = 
+        p4_search_option_value ("wordlists", 9, ORDER_LEN, PFE.set);
+    p4_dict_allocate (PFE_set.wordlists+1, sizeof(void*), sizeof(void*),
+                      (void**) & PFE.context, (void**) 0);
+    p4_dict_allocate (PFE_set.wordlists, sizeof(void*), sizeof(void*),
+                      (void**) & PFE.dforder, (void**) 0);
+
+    if (PFE.dictlimit < PFE.dict + MIN_PAD + MIN_HOLD + 0x4000)
+    {
+	P4_fatal ("impossible memory map");
+	PFE.exitcode = 3;
+	p4_longjmp_exit ();
+    }
+
+    init_accept_lined ();
+
+    /* should be splitted into boot-core exec-bootcommand boot-extensions */
     p4_boot_system ();
-    p4TH = p4_main_threadP; 
+
+    p4_main_threadP_TO_p4TH; 
 
     /* USER-CONF --load-image=<file>            (alias --image-file=<name>) */
     s = p4_search_option_string ("image-file", 10, 0, PFE.set); /* gforth's */
     s = p4_search_option_string ("load-image", 10, s, PFE.set); /* pfe's */
     if (s)
     {
-	if (! p4_load_dict_image (s))
-	{
-	    P4_fail2 ("[%p] load_image failed: %s", p4TH, s);
-	}
+        P4_fail2 ("[%p] load wordset image-file not implemented: %s", p4TH, s);
     }
 
+    /* _______________ evaluate ________________ */
+
     /* process the boot command: */
     if (PFE_set.bootcommand)
     {
@@ -1045,36 +1030,6 @@ p4_main (p4_threadP th)
     if (PFE_set.verbose)
         FX (p4_dot_memory);
     
-    /* USER-CONF: --dump-file=<filename> */
-    if ((s=p4_search_option_string ("dump-file", 9, 0, PFE.set)))
-    {  
-        int f = p4_creat_image(s, 0);
-        if (-1 != f) { 
-            p4_write_image (f, PFE.dict, PFE.dp-PFE.dict, 
-			    "dictbase", PFE.nr); 
-            p4_write_image (f, &PFE,     sizeof(PFE),      
-			    "thread",  PFE.nr); 
-            close (f); 
-        }else{
-            /* perror(PFE_set.dumpfile); */
-        }
-    }
-
-    /* USER-CONF: --make-image=<filename> */
-    if ((s = p4_search_option_string ("make-image", 10, 0, PFE.set)))
-    {  
-        int f = p4_creat_image(s, 0);
-        if (-1 != f) { 
-            p4_write_image (f, PFE.fence, PFE.dp-PFE.fence, 
-			    "dictfence", PFE.nr); 
-            p4_write_image (f, &PFE,     sizeof(PFE),     
-			    "thread",    PFE.nr); 
-            close (f); 
-        }else{
-            /* perror(PFE_set.make_image); */
-        }
-    }
-    
     if (! PFE_set.bye)
 	p4_interpret_loop (); /* will catch QUIT, ABORT, COLD .. and BYE */
     PFE.atexit_cleanup ();
@@ -1088,11 +1043,10 @@ p4_main (p4_threadP th)
 _export int 
 p4_Exec(p4_threadP th)
 {
-    auto p4_thread_VM vm;
     auto volatile int retval;
-    p4_thread_save_VM(&vm);
+    P4_CALLER_SAVEALL;
     retval = p4_main(th);
-    p4_thread_load_VM(&vm);
+    P4_CALLER_RESTORE;
     return retval;
 }
 
