@@ -5,8 +5,8 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE              @(#) %derived_by: guidod %
- *  @version %version: bln_mpt1!5.31 %
- *    (%date_modified: Tue Oct 02 16:07:46 2001 %)
+ *  @version %version: bln_mpt1!5.57 %
+ *    (%date_modified: Mon Feb 24 20:03:36 2003 %)
  *
  *  @description
  *      Compatiblity with former standards, miscellaneous useful words.
@@ -14,7 +14,7 @@
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  misc-ext.c~bln_mpt1!5.31:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  misc-ext.c~bln_mpt1!5.57:csrc:bln_12xx!1 % $";
 #endif
 
 #define _P4_SOURCE 1
@@ -208,9 +208,20 @@ FCode (p4_srand)
     srand (*SP++);
 #endif
 }
-/** UNDER+ ( n1 n2 -- n1+n2 n2 )
+
+/** UNDER+ ( n1 x n2 -- n1+n2 x ) TF
+ *     quicker than
+ : UNDER+  ROT + SWAP ;
+ */
+FCode (p4_plus_under)
+{
+    p4cell n = FX_POP;
+    SP[1] += n;
+}
+
+/** "(UNDER+)" ( n1 n2 -- n1+n2 n2 )
  * quicker than
- : UNDER+ TUCK + SWAP ;
+ : (UNDER+) TUCK + SWAP ; or : (UNDER+) DUP UNDER+ ;
  */
 FCode (p4_under_plus)
 {		
@@ -225,18 +236,22 @@ FCode (p4_under_plus)
  * execution compiled by => +TO
  * adds the stack-val to the lvalue compiled
  */
-FCode (p4_plus_to_execution)	
-{			
-    *p4_to_body (*IP++) += *SP++;
+FCode_XE (p4_plus_to_execution)	
+{	
+    FX_USE_CODE_ADDR;
+    *p4_to_body ((p4xt)(*IP++)) += *SP++;
+    FX_USE_CODE_EXIT;
 }
 
 /** ((+TO.local)) ( val -- )
  * same as => ((+TO)) when the lvalue is a => LOCALS| value
  * <br> compiled by => +TO
  */
-FCode (p4_plus_to_local_execution)
+FCode_XE (p4_plus_to_local_execution)
 {
+    FX_USE_CODE_ADDR;
     LP[(p4cell) *IP++] += *SP++;
+    FX_USE_CODE_EXIT;
 }
 
 /** +TO ( val [name] -- )
@@ -338,6 +353,27 @@ FCode (p4_source_line)
     }
 }
 
+/** SOURCE-NAME ( -- str-ptr str-len )
+ * if => SOURCE is from => INCLUDE then the result is the filename,
+ * otherwise a generic name for the SOURCE-ID is given.
+ */
+FCode (p4_source_name)
+{
+    switch (SOURCE_ID)
+    {
+     case 0:
+         if (BLK) FX_PUSH("*block#*");
+	 else FX_PUSH ("*query*"); /*correct?*/
+	 break;
+     case -1:	
+	 FX_PUSH ("*evaluate*");
+         break;
+     default:			/* source line from text file */
+         FX_PUSH (SOURCE_FILE->name);
+    }
+    FX_1ROOM; SP[0] = strlen((char*)(SP[1]));
+}
+
 /** TH'POCKET ( n -- addr u )
  * returns the specified pocket as a => S" string reference
  */
@@ -554,7 +590,7 @@ FCode (p4_dot_memory)
 	     (long) (PFE.dictlimit - PFE.dict),
 	     (long) (PFE.dp - PFE.dict), 
 	     (long) (PFE.s0 - PFE.stack),  /* sizeof (p4cell) */
-	     (PFE.dstrings ? "cells" : "cells, (extra dstrings stack)"),
+	     (PFE.dstrings ? "cells, (extra dstrings stack)" : "cells"),
 	     (long) (PFE.f0 - PFE.fstack), /* sizeof (double) */
 	     (PFE.f0 ? "floats" : "       (not used)"),
 	     (long) (PFE.r0 - PFE.rstack),  /* sizeof (p4xt**) */
@@ -589,35 +625,48 @@ FCode (p4_dot_status)
     p4_outf ("\nmaximum number of open files:     %u", P4_opt.max_files);
     p4_outf ("\nmaximum simultaneous S\" pockets   %u", P4_opt.pockets);
     p4_outf ("\ndictionary threads configured     %u", 1<<LD_THREADS);
-    p4_outf ("\nmaximum length of search order    %u", ORDER_LEN);
+    p4_outf ("\nmaximum wordlists in search order %u", P4_opt.wordlists);
     FX (p4_cr);
     p4_outf ("\nText screen size:                 %dx%d", PFE.rows, PFE.cols);
-    FX (p4_cr);
+
 #define flag(X) ((X) ? "ON " : "OFF")
-    p4_outf ("\nLOWER-CASE    %s", flag (LOWER_CASE));
+    p4_outf ("\n      CASELESS %s", flag (LOWER_CASE));
+    p4_outf ("\nUPPER-CASE-IS  %s", flag (UPPER_CASE));
 #ifdef P4_REGTH
     p4_outs ("           REGTH="P4_REGTH);
+#elif defined PFE_WITH_STATIC_REGS
+    p4_outs ("           (static regVM)");
+#else
+    p4_outs ("           (static regTH)");
 #endif
-    p4_outf ("\nLOWER-CASE-FN %s", flag (LOWER_CASE_FN));
+    p4_outf ("\nLOWER-CASE-FN  %s", flag (LOWER_CASE_FN));
 #ifdef P4_REGSP
     p4_outs ("           REGSP="P4_REGSP);
 #endif
-    p4_outf ("\nRESET-ORDER   %s", flag (RESET_ORDER));
+    p4_outf ("\nRESET-ORDER-IS %s", flag (RESET_ORDER));
 #ifdef P4_REGIP
     p4_outs ("           REGIP="P4_REGIP);
+#elif defined PFE_SBR_CALL_ARG_THREADING
+    p4_outs ("           (sbr-call-arg-threading)");
+#elif defined PFE_SBR_CALL_THREADING
+    p4_outs ("           (sbr-call-threading)");
 #endif
-    p4_outf ("\nREDEFINED-MSG %s", flag (REDEFINED_MSG));
+    p4_outf ("\nREDEFINED-MSG  %s", flag (REDEFINED_MSG));
 #ifdef P4_REGRP
     p4_outs ("           REGRP="P4_REGRP);
 #endif
-    p4_outf ("\nFLOAT-INPUT   %s", flag (FLOAT_INPUT));
+    p4_outf ("\nFLOAT-INPUT    %s", flag (FLOAT_INPUT));
 #ifdef P4_REGLP
     p4_outs ("           REGLP="P4_REGLP);
+#elif defined PFE_REGFP
+    p4_outs ("           REGFP="P4_REGFP);
 #endif
 #undef flag
-    p4_outf ("\nPRECISION     %d", PRECISION);
+    p4_outf ("\nPRECISION      %d", PRECISION);
 #ifdef P4_REGW
     p4_outs ("            REGW="P4_REGW);
+#elif defined PFE_CALL_THREADING && !defined PFE_SBR_CALL_THREADING
+    p4_outs ("            (call-threading)");
 #endif
     FX (p4_space);
 }
@@ -713,15 +762,32 @@ FCode (p4_paren_type)
  */
 FCode (p4_standard_io)
 {
-    static p4code paren_emit_cfa = PFX (p4_paren_emit);
-    static p4code paren_expect_cfa = PFX (p4_paren_expect);
-    static p4code paren_key_cfa = PFX (p4_paren_key);
-    static p4code paren_type_cfa = PFX (p4_paren_type);
+#  ifdef PFE_CALL_THREADING
+    static const p4Word paren_emit_w =   { "",  PFX (p4_paren_emit) };
+    static const p4Word paren_expect_w = { "", PFX (p4_paren_expect) };
+    static const p4Word paren_key_w =    { "", PFX (p4_paren_key) };
+    static const p4Word paren_type_w =   { "", PFX (p4_paren_type) };
 
-    PFE.emit = &paren_emit_cfa;
-    PFE.expect = &paren_expect_cfa;
-    PFE.key = &paren_key_cfa;
-    PFE.type = &paren_type_cfa;
+    static const p4Word* paren_emit_xt =   & paren_emit_w;
+    static const p4Word* paren_expect_xt = & paren_expect_w;
+    static const p4Word* paren_key_xt =    & paren_key_w;
+    static const p4Word* paren_type_xt =   & paren_type_w;
+
+    PFE.emit =   &paren_emit_xt;
+    PFE.expect = &paren_expect_xt;
+    PFE.key =    &paren_key_xt;
+    PFE.type =   &paren_type_xt;
+#  else
+    static p4code paren_emit_xt =   PFX (p4_paren_emit);
+    static p4code paren_expect_xt = PFX (p4_paren_expect);
+    static p4code paren_key_xt =    PFX (p4_paren_key);
+    static p4code paren_type_xt =   PFX (p4_paren_type);
+
+    PFE.emit =   &paren_emit_xt;
+    PFE.expect = &paren_expect_xt;
+    PFE.key =    &paren_key_xt;
+    PFE.type =   &paren_type_xt;
+#  endif
 }
 
 /************************************************************************/
@@ -746,9 +812,11 @@ p4_store_execution (p4xt xt, int key)
 /** ((EXECUTES)) ( n -- )
  * compiled by => EXECUTES
  */
-FCode (p4_executes_execution)
+FCode_XE (p4_executes_execution)
 {
-    p4_store_execution (*IP++, *SP++);
+    FX_USE_CODE_ADDR;
+    p4_store_execution ((p4xt)(*IP++), *SP++);
+    FX_USE_CODE_EXIT;
 }
 
 /** EXECUTES ( n [word] -- )
@@ -880,14 +948,24 @@ FCode (p4_expand_fn)
 
 /** ((LOAD")) ( -- ? )
  */
-FCode (p4_load_quote_execution)
-{
+FCode_XE (p4_load_quote_execution)
+{   FX_USE_CODE_ADDR {
+#  if !defined PFE_SBR_CALL_THREADING
     register char *p = (char *) IP;
     register int n = (p4char) *p++;
 
     FX_SKIP_STRING;
     p4_load_file (p, n, *SP++);
-}
+#  else
+    register char *p;
+    FX_NEW_IP_WORK;
+    p = FX_NEW_IP_CHAR;
+    FX_NEW_IP_SKIP_STRING;
+    FX_NEW_IP_DONE;
+    p4_load_file (p+1, *p, *SP++);
+#  endif
+    FX_USE_CODE_EXIT;
+}}
 
 /** LOAD"  ( [filename<">] -- ? ) obsolete 
  * load the specified file - this word can be compiled into a word-definition
@@ -921,13 +999,23 @@ FCode (p4_system)
 /** ((SYSTEM")) 
  * compiled by => SYSTEM" commandline"
  */
-FCode (p4_system_quote_execution)
-{
+FCode_XE (p4_system_quote_execution)
+{   FX_USE_CODE_ADDR {
+#  if !defined PFE_SBR_CALL_THREADING
     char *p = (char *) IP;
 
     FX_SKIP_STRING;
     *--SP = p4_systemf ("%.*s", *p, p + 1);
-}
+#  else
+    char *p;
+    FX_NEW_IP_WORK;
+    p = FX_NEW_IP_CHAR;
+    FX_NEW_IP_SKIP_STRING;
+    FX_NEW_IP_DONE;
+    *--SP = p4_systemf ("%.*s", *p, p + 1);
+#  endif
+    FX_USE_CODE_EXIT;
+}}
 
 /** SYSTEM" ( [commandline<">] -- ret-val ) obsolete
  * run a shell command (note:embedded systems have no shell)
@@ -982,7 +1070,7 @@ extern FCode (p4_ok);
  */
 FCode (p4_create_var)
 {
-    FX_HEADER;
+    FX_RUNTIME_HEADER;
     FX_RUNTIME1 (p4_variable);
 }
 
@@ -992,7 +1080,7 @@ FCode (p4_create_var)
  */
 FCode (p4_buffer_var)
 {
-    FX_HEADER;
+    FX_RUNTIME_HEADER;
     FX_RUNTIME1 (p4_variable);
     FX (p4_allot);
 }
@@ -1000,62 +1088,110 @@ FCode (p4_buffer_var)
 /** R'@ ( R: a b -- a R: a b )
  * fetch the next-under value from the returnstack.
  * used to interpret the returnstack to hold two => LOCALS| values.
- * ( => R@ / => 2R@ / => R>DROP / => R"@ )
+ * ( =>'R@' / =>'2R@' / =>'R>DROP' / =>'R"@')
  */
 FCode (p4_r_tick_fetch)
 {
-    FX_PUSH (FX_RP[1]);
+    FX_COMPILE (p4_r_tick_fetch);
 }
+FCode_XE (p4_r_tick_fetch_execution)
+{
+    FX_USE_CODE_ADDR;
+    FX_PUSH (FX_RP[1]);
+    FX_USE_CODE_EXIT;
+}
+P4COMPILES (p4_r_tick_fetch, p4_r_tick_fetch_execution,
+	    P4_SKIPS_NOTHING, P4_DEFAULT_STYLE);
 
 /** R'! ( x R: a b -- R: x b )
  * store the value into the next-under value in the returnstack.
  * used to interpret the returnstack to hold two => LOCALS| values.
- * see => R'@ for inverse operation
+ * see =>"R'@" for inverse operation
  */
 FCode (p4_r_tick_store)
 {
-    FX_RP[1] = FX_POP;
+    FX_COMPILE (p4_r_tick_store);
 }
+FCode_XE (p4_r_tick_store_execution)
+{
+    FX_USE_CODE_ADDR;
+    FX_RP[1] = FX_POP;
+    FX_USE_CODE_EXIT;
+}
+P4COMPILES (p4_r_tick_store, p4_r_tick_store_execution,
+	    P4_SKIPS_NOTHING, P4_DEFAULT_STYLE);
 
 /** R"@ ( R: a b c -- a R: a b c )
  * fetch the second-under value from the returnstack.
  * used to interpret the returnstack to hold three => LOCALS| values.
- * see => R"! for inverse operation ( => R'@ => R@ / => 2R@ / => R>DROP )
+ * see =>'R"!' for inverse operation ( =>"R'@" =>"R@" / =>"2R@" / =>"R>DROP" )
  */
 FCode (p4_r_quote_fetch)
 {
-    FX_PUSH (FX_RP[2]);
+    FX_COMPILE (p4_r_quote_fetch);
 }
+FCode_XE (p4_r_quote_fetch_execution)
+{
+    FX_USE_CODE_ADDR;
+    FX_PUSH (FX_RP[2]);
+    FX_USE_CODE_EXIT;
+}
+P4COMPILES (p4_r_quote_fetch, p4_r_quote_fetch_execution,
+	    P4_SKIPS_NOTHING, P4_DEFAULT_STYLE);
 
 /** R"! ( x R: a b c -- R: x b c )
  * store the value into the second-under value in the returnstack.
  * used to interpret the returnstack to hold three => LOCALS| values.
- * see => R"@ for inverse operation
+ * see =>'R"@' for inverse operation
  */
 FCode (p4_r_quote_store)
 {
-    FX_RP[2] = FX_POP;
+    FX_COMPILE (p4_r_quote_store);
 }
+FCode_XE (p4_r_quote_store_execution)
+{
+    FX_USE_CODE_ADDR;
+    FX_RP[2] = FX_POP;
+    FX_USE_CODE_EXIT;
+}
+P4COMPILES (p4_r_quote_store, p4_r_quote_store_execution,
+	    P4_SKIPS_NOTHING, P4_DEFAULT_STYLE);
 
 /** R! ( x R: a -- R: x )
  * store the value as the topmost value in the returnstack.
- * see => R@ for inverse operation ( => R'@ / => R"@ / => 2R@ / => 2R! )
+ * see =>"R@" for inverse operation ( =>"R'@" / =>'R"@' / =>'2R@' / =>'2R!')
  */
 FCode (p4_r_store)
 {
-    FX_RP[0] = FX_POP;
+    FX_COMPILE (p4_r_store);
 }
+FCode_XE (p4_r_store_execution)
+{
+    FX_USE_CODE_ADDR;
+    FX_RP[0] = FX_POP;
+    FX_USE_CODE_EXIT;
+}
+P4COMPILES (p4_r_store, p4_r_store_execution,
+	    P4_SKIPS_NOTHING, P4_DEFAULT_STYLE);
 
 /** 2R! ( x y R: a b -- R: x y )
  * store the value as the topmost value in the returnstack.
- * see => R@ for inverse operation ( => R'@ / => R"@ / => 2R@ / => 2R! )
+ * see =>"2R@" for inverse operation ( =>"R'@" / =>'R"@' / =>'2R@' / =>'2R!')
  */
 FCode (p4_two_r_store)
 {
+    FX_COMPILE (p4_two_r_store);
+}
+FCode_XE (p4_two_r_store_execution)
+{
+    FX_USE_CODE_ADDR;
     FX_RP[0] = SP[0];
     FX_RP[1] = SP[1];
     FX_2DROP;
+    FX_USE_CODE_EXIT;
 }
+P4COMPILES (p4_two_r_store, p4_two_r_store_execution,
+	    P4_SKIPS_NOTHING, P4_DEFAULT_STYLE);
 
 /** DUP>R ( val -- val )
  * shortcut, see => R>DROP
@@ -1064,8 +1200,22 @@ FCode (p4_two_r_store)
  */
 FCode (p4_dup_to_r)
 {
-    RP_PUSH (*SP);
+    FX_COMPILE (p4_dup_to_r);
 }
+FCode_XE (p4_dup_to_r_execution)
+{
+    FX_USE_CODE_ADDR;
+#  ifndef PFE_SBR_CALL_THREADING
+    RP_PUSH (*SP);
+#  else
+    FX_NEW_RP_WORK;
+    FX_NEW_RP_CELL_POSH = (p4xcode*) *SP;
+    FX_NEW_RP_DONE;
+#  endif
+    FX_USE_CODE_EXIT;
+}
+P4COMPILES (p4_dup_to_r, p4_dup_to_r_execution,
+	    P4_SKIPS_NOTHING, P4_DEFAULT_STYLE);
 
 /** R>DROP ( -- )
  * shortcut (e.g. in CSI-Forth)
@@ -1077,8 +1227,20 @@ FCode (p4_dup_to_r)
  */
 FCode (p4_r_from_drop)
 {
-    RP++;
+    FX_COMPILE (p4_r_from_drop);
 }
+FCode_XE (p4_r_from_drop_execution)
+{
+    FX_USE_CODE_ADDR;
+#  ifndef PFE_SBR_CALL_THREADING
+    RP++;
+#  else
+    FX_RP_DROP (1);
+#  endif
+    FX_USE_CODE_EXIT;
+}
+P4COMPILES (p4_r_from_drop, p4_r_from_drop_execution,
+	    P4_SKIPS_NOTHING, P4_DEFAULT_STYLE);
 
 /** 2R>2DROP ( -- )
  * this is two times => R>DROP but a bit quicker.
@@ -1086,8 +1248,16 @@ FCode (p4_r_from_drop)
  */
 FCode (p4_two_r_from_drop)
 {
-    RP += 2;
+    FX_COMPILE (p4_two_r_from_drop);
 }
+FCode_XE (p4_two_r_from_drop_execution)
+{
+    FX_USE_CODE_ADDR;
+    FX_RP_DROP (2);
+    FX_USE_CODE_EXIT;
+}
+P4COMPILES (p4_two_r_from_drop, p4_two_r_from_drop_execution,
+	    P4_SKIPS_NOTHING, P4_DEFAULT_STYLE);
 
 /** CLEARSTACK ( -- )
  * reset the parameter stack to be empty
@@ -1098,6 +1268,49 @@ FCode (p4_clearstack)
     p4SP = PFE.s0;
 }
 
+/** LOWER-CASE ( -- var* ) depracated
+ * Call to enable lower-case symbols as input
+ * Replace code of "LOWER-CASE OFF" with "NO TO UPPER-CASE?"
+ * see => UPPER-CASE?
+ */
+
+/** UPPER-CASE? ( -- flag ) 
+ * Call to check whether lower-case symbols in input can match 
+ * words defined in uppercase. Actually it sets the internal
+ * wordl-flag which has some bit-defs used when creating new
+ * vocabularies.
+ *
+ * This flag => VALUE is modifiable with => TO
+   YES TO UPPER-CASE?
+ */
+
+/** REDEFINED-MSG ( -- var* ) depracated
+ * Call to enable warnings if creating symbols being in the => ORDER 
+ * Replace code of "REDEFINED-MSG OFF" with "NO TO REDEFINED-MSG?"
+ * see => REDEFINED-MSG?
+ */
+
+/** REDEFINED-MSG? ( -- flag ) 
+ * Call to check whether the system will emit warnings if creating 
+ * symbols being already defined in the => CURRENT vocabulary.
+ *
+ * This flag => VALUE is modifiable with => TO
+   YES TO REDEFINED-MSG?
+ */
+
+/** QUOTED-PARSE ( -- var* ) depracated
+ * Call to enable quoted-parse extension in =>"PARSE" =>"WORD"s 
+ * Replace code of "QUOTED-PARSE OFF" with "NO TO QUOTED-PARSE?"
+ * see => QUOTED-PARSE?
+ */
+
+/** QUOTED-PARSE? ( -- flag ) 
+ * Call to check for quoted-parse extension in =>"PARSE" =>"WORD"s 
+ *
+ * This flag => VALUE is modifiable with => TO
+   YES TO QUOTED-PARSE?
+ */
+
 P4_LISTWORDS (misc) =
 {
     P4_INTO ("FORTH", 0),
@@ -1124,7 +1337,8 @@ P4_LISTWORDS (misc) =
     P4_FXco ("-ROLL",		p4_dash_roll),
     P4_FXco ("RANDOM",		p4_random),
     P4_FXco ("SRAND",		p4_srand),
-    P4_FXco ("UNDER+",		p4_under_plus),
+    P4_FXco ("(UNDER+)",	p4_under_plus),
+    P4_xOLD ("UNDER+",		"(UNDER+)"),
 
     /* more local variables */
     P4_SXco ("+TO",		p4_plus_to),
@@ -1136,23 +1350,17 @@ P4_LISTWORDS (misc) =
     P4_FXco (".STATUS",		p4_dot_status),
     P4_FNYM ("SHOW-STATUS",	".STATUS"),
 
-    /* Call to enable lower-case symbols as input, supersedes plain Vars */
-    P4_DVaL ("LOWER-CASE@",	wordl_flag),    
-    P4_DSeT ("LOWER-CASE!",	wordl_flag),    
-    P4_DVaR ("LOWER-CASE",	wordl_flag),    /* fixme: delete somewhen */
-    P4_DVaL ("LOWER-CASE-FN@",	lower_case_fn),
-    P4_DSeT ("LOWER-CASE-FN!",	lower_case_fn),
+    /* NOTE: the newer "TO"-implementation is able to set DVaLs */
+    P4_DVaL ("UPPER-CASE?",	wordl_flag), /* will be bool-VaL */
+    P4_DVaR ("LOWER-CASE",	wordl_flag), /* fixme: deleted somewhen */
+    P4_DVaL ("LOWER-CASE-FN?",	lower_case_fn), /* will be bool-VaL */
     P4_DVaR ("LOWER-CASE-FN",	lower_case_fn), /* fixme: delete somewhen */
-    /** Call to enable warnings if creating symbols being in the => ORDER */
-    P4_DVaL ("REDEFINED-MSG@",	redefined_msg),
-    P4_DSeT ("REDEFINED-MSG!",	redefined_msg),
+    P4_DVaL ("REDEFINED-MSG?",	redefined_msg), /* will be bool-VaL */
     P4_DVaR ("REDEFINED-MSG",	redefined_msg), /* fixme: delete somewhen */
-    /** Call to enable quoted-parse extension in =>"PARSE" =>"WORD"s */
-    P4_DVaR ("QUOTED-PARSE*",	quoted_parse),
-    P4_DVaL ("QUOTED-PARSE@",	quoted_parse),
-    P4_DSeT ("QUOTED-PARSE!",	quoted_parse),
+    P4_DVaL ("QUOTED-PARSE?",	quoted_parse),  /* will be bool-VaL */
   
     P4_FXco ("SOURCE-LINE",	p4_source_line),
+    P4_FXco ("SOURCE-NAME",	p4_source_name),
     P4_FXco ("TH'POCKET",	p4_th_pocket),
     P4_FXco ("POCKET-PAD",	p4_pocket_pad),
     P4_OCoN ("/CELL",		sizeof (p4cell)),
@@ -1221,17 +1429,20 @@ P4_LISTWORDS (misc) =
     P4_FXco ("BUFFER:",		p4_buffer_var),
 
 /* quite useful, esp. for non-locals code */
-    P4_FXco ("R'@",		p4_r_tick_fetch),
-    P4_FXco ("R'!",		p4_r_tick_store),
-    P4_FXco ("R\"@",		p4_r_quote_fetch),
-    P4_FXco ("R\"!",		p4_r_quote_store),
-    P4_FXco ("R!",		p4_r_store),
-    P4_FXco ("2R!",		p4_two_r_store),
-    P4_FXco ("DUP>R",		p4_dup_to_r),
-    P4_FXco ("R>DROP",		p4_r_from_drop),
-    P4_FXco ("2R>2DROP",	p4_two_r_from_drop),
+    P4_SXco ("R'@",		p4_r_tick_fetch),
+    P4_SXco ("R'!",		p4_r_tick_store),
+    P4_SXco ("R\"@",		p4_r_quote_fetch),
+    P4_SXco ("R\"!",		p4_r_quote_store),
+    P4_SXco ("R!",		p4_r_store),
+    P4_SXco ("2R!",		p4_two_r_store),
+    P4_SXco ("DUP>R",		p4_dup_to_r),
+    P4_SXco ("R>DROP",		p4_r_from_drop),
+    P4_SXco ("2R>2DROP",	p4_two_r_from_drop),
     P4_FXco ("CLEARSTACK",	p4_clearstack),
 
+    P4_INTO ("EXTENSIONS",      "FORTH"),
+    P4_FXco ("+UNDER",          p4_plus_under),
+
 /* smart */
     P4_FXco ("EXECUTES",	p4_executes),
 
