@@ -4,8 +4,8 @@
  *  Copyright (C) Tektronix, Inc. 1998 - 2001. All rights reserved.
  *
  *  @author Tektronix CTE             @(#)  %derived_by: guidod %
- *  @version %version: 5.20 %
- *    (%date_modified: Fri Aug 31 13:43:43 2001 %)
+ *  @version %version: bln_mpt1!5.30 %
+ *    (%date_modified: Wed Aug 28 09:57:32 2002 %)
  *
  *  @description
  *              The Terminal Driver for K12xx-2 Emulation System,
@@ -18,7 +18,7 @@
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec: term-k12.c~5.20:csrc:bln_12xx!2 % $ ";
+"@(#) $Id: %full_filespec: term-k12.c~bln_mpt1!5.30:csrc:bln_12xx!2 % $ ";
 #endif
 
 #define _P4_SOURCE 1
@@ -45,6 +45,7 @@ static char* id __attribute__((unused)) =
 #endif
 #include <sysSymTbl.h>
 #include <pfe/logging.h>
+#include <pfe/def-restore.h>
 
 #include <pfe/term-k12.h>
 int tQSend (u32_t qId, void* hdr, u32_t hdrlen, 
@@ -55,6 +56,14 @@ int tQSend (u32_t qId, void* hdr, u32_t hdrlen,
 #include <taskLib.h>
 #include <inetLib.h>
 
+#ifdef _TERMTOFRONT_ON_ABORT
+# if 0
+# include <K12/gemapi.h>
+# else
+# include <gemapi.h>
+# endif
+#endif
+
 /* two defines stolen from gemrpc.h which is not included here */
 #ifndef FORTH_MAX_MESSAGE_SIZE
 #define FORTH_MAX_MESSAGE_SIZE     0x200
@@ -154,6 +163,26 @@ void k12LogString (int mask, char* hail, char* p, int size)
     
 }  
 
+k12_status_t
+write_link (k12_emu_type_t* emu, k12_emu_event_t link, int type,
+            s8_t* buf, u32_t n)
+{
+    /* messages have the format seen on the stdin-link. If you know the
+     * the stdin-link id of a pfe-box then you can connect the output of
+     * one pfe box with the input of another box if you dare to.
+     */
+    register k12_status_t e;
+    auto k12_emu_msg_subhead_t* msg;
+    e = k12EmuLowBufferGet(emu, n + sizeof(*msg), (s8_t**)(&msg));
+    if (e) return e;
+    msg->type = type; /* K12_EMU_DATA_REQ or K12_EMU_DATA_IND */
+    msg->para0 = n; (void*)msg->handle = emu;
+    msg->para1 = 0; msg->para2 = 0;
+    memcpy (& msg[1], buf, n);
+    return k12EmuLowEventSend (emu, link, (s8_t*)(msg), n + sizeof(*msg), 
+                               K12_EMU_NOOPT);
+}
+
 /* ************************************************ */
 /* output thread */
 
@@ -196,6 +225,9 @@ p4_TX_task (k12_priv* p)
         if (n && p->tx_qid) 
         {
             do {
+                if (p->tx_link) 
+                    write_link (p->emu, p->tx_link, K12_EMU_DATA_REQ, buf, n);
+
                 if (tQSend (p->tx_qid, 0, 0, buf, n, OUTPUT_TQSEND_TIMEOUT) 
                   == OK)
                 {
@@ -215,6 +247,8 @@ p4_TX_task (k12_priv* p)
                     n = rngBufGet (p->tx_ring, buf, OUTBUF_SIZE);
             } while (n);
         } else { /* offline */
+            if (p->tx_link) 
+                write_link (p->emu, p->tx_link, K12_EMU_DATA_REQ, buf, n);
             if (p->tx_logfile)
                 write (p->tx_logfile, buf, n);
             if (n > 32)
@@ -345,9 +379,8 @@ tx_task_spawn (k12_priv* k12p)
 {
     int prio=0;
     char* taskname = taskName(taskIdSelf());
-    
 #    ifdef P4_UPPER_REGS /* i960 */
-    _p4_thread_save();
+    P4_CALLER_MKSAVED
 #    endif
     
     /* avoiding prio inversion by making the output_thread of
@@ -373,7 +406,7 @@ tx_task_spawn (k12_priv* k12p)
                                    (int) k12p, 0, 0, 0, 0, 0, 0, 0, 0, 0);
     }
 #    ifdef P4_UPPER_REGS
-    _p4_thread_load(); 
+    P4_CALLER_RESTORE
 #    endif
 }
 
@@ -395,6 +428,62 @@ tx_task_check (k12_priv* k12p)
     }
 }
 
+#ifdef _TERMTOFRONT_ON_ABORT
+/* this is a copy from the k12termtofront.module */
+static FCode (p4_k12TermToFront_on_abort)
+{
+# if 0
+    extern status_t k12GemStubTerminalWnd(uint32_t ulGemId,     
+                               char *pEmulName,      /* emulation name */
+                               uint32_t ulAction );  /* action: hide/pop up */
+# endif
+    extern status_t k12EmuLowEmulNameGet(k12_emu_type_t *EmulId, 
+                                         char **EmulName);
+
+    register int e;
+    register k12_emu_type_t * emul_id;
+    auto uint32_t stackid;
+    auto char* emulname;
+    
+    emul_id = P4_K12_PRIV(p4TH)->emu;
+
+    e = k12EmuLowPipeIdGet (emul_id, &stackid);
+    if (e) { P4_warn1 ("k12EmuLowPipeIdGet failed (%x)", e); return; }
+
+    e = k12EmuLowEmulNameGet (emul_id, &emulname);
+    if (e) { P4_warn1 ("k12EmuLowNameGet failed (%x)", e); return; }
+
+    e = k12GemStubTerminalWnd (stackid, emulname, K12_EMU_GEM_TERM_WND_POPUP);
+    if (e) { P4_warn1 ("k12GemStubTerminalWnd failed (%x)", e); return; }
+}
+#endif
+
+/*
+ * the abort-cleanups are also run with COLD - this function here should
+ * actullay hook into a THROW list and then be executed.
+ */
+static FCode(p4_paintred_on_abort)
+{
+    k12_priv* k12p = K12PRIV(p4TH);
+
+    if (k12p->validation_req)
+        return; /* in setup-phase : QUIT calls ABORT[*] followed by QUERY */
+
+# ifdef _TERMTOFRONT_ON_ABORT
+    if (k12p->state == K12_EMU_NOT_LOADED)
+    { FX (p4_k12TermToFront_on_abort); }
+# endif
+
+    k12p->state = K12_EMU_ERROR_CASE;
+    if (PFE.tib && ! k12p->answering)
+    {
+        strncpy (k12p->answerbuf, PFE.tib, K12_ANSWERBUF_SIZE);
+        k12p->answerwarning = strlen (k12p->answerbuf);
+    }else{
+        k12p->answerwarning = 0;
+    }
+}
+
 static int
 c_prepare_terminal (void)
 {
@@ -440,6 +529,8 @@ c_prepare_terminal (void)
             rngDelete (k12p->tx_ring); return 0;
         }
     }
+
+    PFE.abort[2] = PFX(p4_paintred_on_abort); /* abusing abort_float vector */
     
     k12p->rx_dataSAP = DATA_SAP;
     return 1;
@@ -470,18 +561,30 @@ c_cleanup_terminal (void)
 
 void p4_set_fkey_name (int num, char* name);
 
+#define FORTHTYPE(_TYPE_) (0x4e00 < (_TYPE_) && (_TYPE_) < 0x5000)
+
 /* handle incoming events until a terminal' key-event is received */
 static int 
 p4_k12_wait_for_stdin ()
 {
     register k12_priv* k12p = K12PRIV(p4TH);
-    k12p->answering = FALSE;
-    k12p->frm_putback = 0;
 
-    if (k12p->state == K12_EMU_NOT_LOADED)
+    if (k12p->answering && k12p->frm_putback == '\n')
+    {   /*              && k12p->frm_input == k12p->rx_link */
+        k12p->rx_data = "\n";
+        k12p->rx_datalen = 1;
+        k12p->rx_dataIN = 0;
+        k12p->frm_putback = 0;
+        return 1;
+    }else{
+        k12p->answering = FALSE;  
+        k12p->frm_putback = 0;
+        /*fallthrough*/
+    }
+
+    if (k12p->state == K12_EMU_NOT_LOADED && ! k12p->validation_req)
         k12p->state = K12_EMU_IDLE;
     
-#   define _sizeof_subhead_t (sizeof(k12_emu_msg_subhead_t))  
     while (1)
     {
         if (k12p->frm_putback) { k12p->frm_putback = 0; }
@@ -495,16 +598,24 @@ p4_k12_wait_for_stdin ()
               "got>", "event via %i len %i\n", 
               k12p->frm_input, k12p->frm_datalen);  
         }
-        if (k12p->frm_input == k12p->rx_dataSAP)
+        if (k12p->rx_dataSAP == k12p->frm_input
+            || (k12p->rx_link && k12p->rx_link == k12p->frm_input))
         {  
             register k12_emu_msg_subhead_t* msg = (void*)k12p->frm_data;
             switch (msg->type) 
             {
+             case K12_EMU_VALIDATION_REQ:
+                 if (k12p->validation_req)
+                     k12p->validation_req --;
+                 if (k12p->state == K12_EMU_NOT_LOADED)
+                     k12p->state = K12_EMU_IDLE;
+                 break;
              case K12_EMU_DATA_REQ: 
-                 K12LogMsg2 ( K12_LOG_DEBUG_L(11), "got>", "K12_EMU_DATA_REQ(%i) len=%i\n",
-                   msg->type, k12p->frm_datalen -_sizeof_subhead_t); 
-                 k12p->rx_data = k12p->frm_data +_sizeof_subhead_t;
-                 k12p->rx_datalen = k12p->frm_datalen -_sizeof_subhead_t;
+                 K12LogMsg2 ( K12_LOG_DEBUG_L(11), "got>", 
+                              "K12_EMU_DATA_REQ(%i) len=%i\n",
+                   msg->type, k12p->frm_datalen - sizeof(*msg)); 
+                 k12p->rx_data = k12p->frm_data + sizeof(*msg);
+                 k12p->rx_datalen = k12p->frm_datalen - sizeof(*msg);
                  k12p->rx_dataIN = 0;
                  k12LogString(K12_LOG_DEBUG_L(10), 
                    "got>", k12p->rx_data, k12p->rx_datalen);
@@ -525,23 +636,27 @@ p4_k12_wait_for_stdin ()
                          return 0;
                  }else{
                      /* there are new chars to be handled (will go to TIB) */
+
+                     /* see new p4_paintred_on_abort - paint good now */
+                     if (k12p->state == K12_EMU_ERROR_CASE ||
+                         k12p->state == K12_EMU_WARNING)
+                         k12p->state = K12_EMU_IDLE;
                      return 1;
                  }
                  break;
              case K12_EMU_XDAT_REQ: /* command/answer facility */
                  K12LogMsg2 (K12_LOG_DEBUG_L(11),
                    "got>", "K12_EMU_XDAT_REQ(%i) len=%i\n",
-                   msg->type, k12p->frm_datalen -_sizeof_subhead_t); 
-                 k12p->rx_data = k12p->frm_data +_sizeof_subhead_t;
-                 k12p->rx_datalen = k12p->frm_datalen -_sizeof_subhead_t;
+                   msg->type, k12p->frm_datalen - sizeof(*msg)); 
+                 k12p->rx_data = k12p->frm_data + sizeof(*msg);
+                 k12p->rx_datalen = k12p->frm_datalen - sizeof(*msg);
                  k12p->rx_dataIN = 0;
                  k12LogString (K12_LOG_DEBUG_L(10), 
                    "cmd>", k12p->rx_data, k12p->rx_datalen);
                  
                  k12p->answering = TRUE;
                  if (!k12p->rx_datalen || !k12p->rx_data[0])
-                 {
-                     /* empty input */
+                 { /* empty input - old fixup to frm_putback problems */
                      k12p->rx_data = "\n";
                      k12p->rx_datalen = 1;
                      k12p->rx_dataIN = 0;
@@ -549,6 +664,10 @@ p4_k12_wait_for_stdin ()
                  /* -> start command/answer facility */
                  strcpy (k12p->answerbuf, "\\ ");
                  k12p->answeridx = strlen (k12p->answerbuf);
+
+                 /* see new p4_paintred_on_abort - paint partly good now */
+                 if (k12p->state == K12_EMU_ERROR_CASE)
+                     k12p->state = K12_EMU_WARNING;
                  return 1;
              case K12_EMU_CONN_REQ: 
                  K12LogMsg2 (K12_LOG_DEBUG_L(11),
@@ -604,27 +723,60 @@ p4_k12_wait_for_stdin ()
                    "got>"," K12_EMU_UnKnOwN(%i) handle=%i;\n",
                    msg->type, msg->handle);
 
-                 if (k12p->eventHook 
-                   && 0x4e00 < msg->type && msg->type < 0x5000)
-                 { (*k12p->eventHook)(
-                                      k12p->frm_input, 
-                                      k12p->frm_data, 
-                                      k12p->frm_datalen);
+                 if (k12p->eventHook && FORTHTYPE(msg->type))
+                 { (*k12p->eventHook)(k12p->frm_input, 
+                                      k12p->frm_data, k12p->frm_datalen);
+                 }
+            }
+        }
+        else if (k12p->qx_link && k12p->qx_link == k12p->frm_input)
+        {
+            register k12_emu_msg_subhead_t* msg = (void*)k12p->frm_data;
+            switch (msg->type) 
+            {
+             case K12_EMU_DATA_REQ: /* command/answer facility */
+                 K12LogMsg2 (K12_LOG_DEBUG_L(11),
+                   "got>", "K12_EMU_DATA_REQ(%i) len=%i to-forth\n",
+                   msg->type, k12p->frm_datalen - sizeof(*msg)); 
+                 k12p->rx_data = k12p->frm_data + sizeof(*msg);
+                 k12p->rx_datalen = k12p->frm_datalen - sizeof(*msg);
+                 k12p->rx_dataIN = 0;
+                 k12LogString (K12_LOG_DEBUG_L(10), 
+                   "cmd>", k12p->rx_data, k12p->rx_datalen);
+                 
+                 k12p->answering = TRUE;
+                 /* -> start command/answer facility */
+                 k12p->answeridx = 0; 
+                 k12p->frm_putback = '\n';
+                 return 1;
+             default: 
+                 K12LogMsg2 (K12_LOG_DEBUG_L(11),
+                   "got>"," K12_EMU_UnKnOwN(%i) handle=%i;\n",
+                   msg->type, msg->handle);
+
+                 if (k12p->eventHook && FORTHTYPE(msg->type))
+                 { (*k12p->eventHook)(k12p->frm_input, 
+                                      k12p->frm_data, k12p->frm_datalen);
                  }
             }
         }else{
-            if (k12p->eventHook) 
+            if (K12_EMU_VALIDATION_REQ == 
+                ((k12_emu_msg_subhead_t*)k12p->frm_data)->type)
             {
-                if ( (*k12p->eventHook)(
-                                        k12p->frm_input, 
-                                        k12p->frm_data, 
-                                        k12p->frm_datalen))
-                {
-                    continue;
-                }
+                 if (k12p->validation_req)
+                     k12p->validation_req --;
+                 if (k12p->state == K12_EMU_NOT_LOADED)
+                     k12p->state = K12_EMU_IDLE;
+            }
+            else if (k12p->eventHook) 
+            {
+                if ( (*k12p->eventHook)(k12p->frm_input, 
+                                        k12p->frm_data, k12p->frm_datalen))
+                { continue; } /* else unhandled message */
             }else{
                 K12LogMsg1 (K12_LOG_WARNING,
-                  "got>", "unknown input - skipped event via %i\n", k12p->frm_input);
+                            "got>", "unknown input - skipped event via %i\n", 
+                            k12p->frm_input);
             }
         }
     }
@@ -753,15 +905,6 @@ c_getkey (void)
 /* Output to screen, control with termcap:                              */
 /************************************************************************/
 
-static void 
-c_putchar (char c)
-{
-    k12_priv* k12p = K12PRIV(p4TH);
-
-    k12p->buffer[(int)(k12p->bufidx++)] = c;
-    k12p->buffer[(int)(k12p->bufidx)] = 0;
-}
-
 static void
 c_put_flush (void)
 {
@@ -786,29 +929,97 @@ c_put_flush (void)
     tx_task_check (k12p);  
     semGive (k12p->tx_sem);
     
-    if (k12p->answering                        /* command/answer facility */
-      && k12p->answeridx < K12_ANSWERBUF_SIZE-1) 
-    {
+    if (! k12p->answering) goto done; 
+    /* -------------------------------------- command/answer facility ----- */
+    if (k12p->qx_link)
+    { /* line mode on qx_link for terminal output scanning as for TMS */
+        register int i;
+        if (!isprint(*k12p->buffer) && !isspace(*k12p->buffer)) /* esc-seq */
+        {
+            if (k12p->answeridx)
+            {
+                write_link (k12p->emu, k12p->qx_link, K12_EMU_DATA_IND,
+                            k12p->answerbuf, k12p->answeridx);
+                k12p->answeridx = 0;
+            }
+
+            write_link (k12p->emu, k12p->qx_link, K12_EMU_DATA_IND,
+                        k12p->buffer, k12p->bufidx);
+            goto done;
+        }
+
+        for (i=0; i < k12p->bufidx; i++)
+        {
+            register unsigned char c;
+            if (k12p->answeridx >= K12_ANSWERBUF_SIZE-1)
+            {
+                write_link (k12p->emu, k12p->qx_link, K12_EMU_DATA_IND,
+                            k12p->answerbuf, k12p->answeridx);
+                k12p->answeridx = 0;
+            }
+            
+            c = k12p->buffer[i];
+            if (isprint (c))
+            {
+                k12p->answerbuf[k12p->answeridx++] = c;
+            }else
+            {
+                if (k12p->answeridx)
+                {
+                    write_link (k12p->emu, k12p->qx_link, K12_EMU_DATA_IND,
+                                k12p->answerbuf, k12p->answeridx);
+                    k12p->answeridx = 0;
+                }
+                if (c == '\n')
+                    write_link (k12p->emu, k12p->qx_link, K12_EMU_DATA_IND,
+                                "\n", 1);
+                else
+                    k12p->answerbuf[k12p->answeridx++] = c;
+            }
+        }
+
+        if (k12p->answeridx && !isprint(*k12p->answerbuf))
+        { /* send controls always, including newlines at end of line */
+            write_link (k12p->emu, k12p->qx_link, K12_EMU_DATA_IND,
+                        k12p->answerbuf, k12p->answeridx);
+            k12p->answeridx = 0;
+        }
+        // goto done;
+    } 
+    else  if (k12p->answeridx < K12_ANSWERBUF_SIZE-1) 
+    { /* answer to emu-config commandline only when no qx_link command mode */
         register int i;
         register unsigned char c;
         for (i=0; i < k12p->bufidx; i++)
         {
             if (k12p->answeridx >= K12_ANSWERBUF_SIZE-1)
                 break;
-           
+            
             c = k12p->buffer[i];
-            if (isprint (c) && !iscntrl(c))
+            if (isprint (c) && ! iscntrl (c))
                 k12p->answerbuf[k12p->answeridx++] = c;
             else
                 k12p->answerbuf[k12p->answeridx++] = '\t'; 
         }
         k12p->answerbuf[k12p->answeridx] = '\0';
+        k12p->answerwarning = 0;
     }
 
+ done:
     /* ->bufidx == 0 will mark it as "flushed" -> see wait_for_stdin */
     k12p->bufidx = 0; *k12p->buffer = 0;
 }
 
+static void 
+c_putchar (char c)
+{
+    k12_priv* k12p = K12PRIV(p4TH);
+
+    k12p->buffer[(int)(k12p->bufidx++)] = c;
+    if (k12p->bufidx >= sizeof (k12p->buffer)) { c_put_flush (); }
+    k12p->buffer[(int)(k12p->bufidx)] = 0;
+}
+
 static void				/* output character and */
 c_putc_noflush (char c)		/* trace the cursor position */
 {
