@@ -5,8 +5,8 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE              @(#) %derived_by: guidod %
- *  @version %version: bln_mpt1!5.15 %
- *    (%date_modified: Tue Oct 02 22:01:00 2001 %)
+ *  @version %version: bln_mpt1!5.16 %
+ *    (%date_modified: Tue Sep 10 12:31:47 2002 %)
  *
  *  @description
  *                      Terminal driver for UNIX-like systems using
@@ -19,7 +19,7 @@
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  term-lib.c~bln_mpt1!5.15:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  term-lib.c~bln_mpt1!5.16:csrc:bln_12xx!1 % $";
 #endif
 
 #define _P4_SOURCE 1
@@ -31,10 +31,46 @@ static char* id __attribute__((unused)) =
 #include <string.h>		/* strlen() */
 #include <errno.h>		/* EINTR */
 
+#ifdef PFE_HAVE_UNISTD_H 
+#include <unistd.h>
+#endif
+#ifdef PFE_HAVE_FCNTL_H
+#include <fcntl.h>
+#endif
+
 #include <pfe/term-sub.h>
 #include <pfe/_nonansi.h>
 #include <pfe/_missing.h>
 
+#if ! defined PFE_HAVE_FCNTL && defined F_GETFL && defined F_SETFL
+#define PFE_HAVE_FCNTL 1 /* should have been done by configure... */
+#endif
+
+#ifdef PFE_HAVE_FCNTL
+#define P4_fcntl_F_GETFL(_FD_,_ARG_) (fcntl((_FD_),F_GETFL,(_ARG_)))
+#define P4_fcntl_F_SETFL(_FD_,_ARG_) (fcntl((_FD_),F_SETFL,(_ARG_)))
+#else
+int P4_fcntl_F_GETFL(int fd, int arg)
+{
+    static seen;
+    if (! seen)
+    {
+        P4_warn2 ("no fcntl - ignored F_GETFL %i on fd %i", arg, fd);
+        seen++;
+    }
+    return 0;
+}
+int P4_fcntl_F_SETFL(int fd, int arg)
+{
+    static seen;
+    if (! seen)
+    {
+        P4_warn2 ("no fcntl - ignored F_SETFL %i on fd %i", arg, fd);
+        seen++;
+    }
+}
+#endif
+
 /*-
  * Table of contents
  *
@@ -110,24 +146,20 @@ static void c_tput (int);
 # endif
 #endif
 
-static int saved_fcntl[3] = {0};
+static int saved_fcntl[3] = {0}; /* fixme: in multithreaded, use only once */
 
 static void			/* set file associated with fd to */
 waitchar (int fd)		/* "wait for character" */
 {
-  int flags;
-
-  flags = fcntl (fd, F_GETFL, 0);
-  fcntl (fd, F_SETFL, flags & ~O_NONBLOCK);
+    int flags = P4_fcntl_F_GETFL (fd, 0); /* fcntl (fd, F_GETFL, 0); */
+    P4_fcntl_F_SETFL (fd, flags & ~O_NONBLOCK);
 }
 
 static void			/* set file associated with fd to */
 no_waitchar (int fd)		/* "don't wait for character" */
 {
-  int flags;
-
-  flags = fcntl (fd, F_GETFL, 0);
-  fcntl (fd, F_SETFL, flags | O_NONBLOCK);
+    int flags = P4_fcntl_F_GETFL (fd, 0);  /* fcntl (fd, F_GETFL, 0); */
+    P4_fcntl_F_SETFL (fd, flags | O_NONBLOCK);
 }
 
 /*
@@ -240,7 +272,8 @@ c_interrupt_key (char ch)
 static void
 tty_interactive (void)
 {
-  struct termios tty = tty_system;
+  /* struct termios tty = tty_system; */
+  struct termios tty; memcpy(&tty, &tty_system, sizeof(tty));
 
   if (! isatty (STDIN_FILENO)) return;
 
@@ -726,6 +759,8 @@ p4_term_struct p4_term_ios =
 #undef TRUE			/* These confuse Sun's cc in curses.h */
 #undef FALSE
 #include <sys/termio.h>		/* Solaris' curses.h seems to need this */
+#elif defined HAVE_TERMIO_H
+#include <termio.h>             /* AIX 4.1 needs this */
 #endif
 #include <curses.h>
 #include <term.h>
@@ -848,24 +883,24 @@ p4_term_struct p4_term_ios =
 static int				/* Prepares usage of all other functions. */
 c_prepare_terminal (void)		/* Call only once at startup. */
 {
-  int fd;
+    int fd;
   
-#if defined ASSUME_DUMBTERM
-  PFE.control_string = p4_dumbterm_controls;
-  PFE.rawkey_string  = p4_dumbterm_rawkeys;
-#elif defined ASSUME_VT100
-  PFE.control_string = p4_vt100_controls;
-  PFE.rawkey_string  = p4_vt100_rawkeys;
-#endif
+#  if defined ASSUME_DUMBTERM
+    PFE.control_string = p4_dumbterm_controls;
+    PFE.rawkey_string  = p4_dumbterm_rawkeys;
+#  elif defined ASSUME_VT100
+    PFE.control_string = p4_vt100_controls;
+    PFE.rawkey_string  = p4_vt100_rawkeys;
+#  endif
   
 
-  /* save state before all changes */
-  for (fd = 0; fd < 3; fd++)
-    saved_fcntl[fd] = fcntl (fd, F_GETFL, 0);
-  tty_save ();
+    /* save state before all changes */
+    for (fd = 0; fd < 3; fd++)
+        saved_fcntl[fd] = P4_fcntl_F_GETFL (fd, 0); /* fcntl(fd, F_GETFL, 0) */
+    tty_save ();
 
-  /* initialize termcap-stuff */
-  return query_database ();
+    /* initialize termcap-stuff */
+    return query_database ();
 }
 
 static void
@@ -884,15 +919,15 @@ c_interactive_terminal (void)	/* `cbreak' mode used by p4th */
 static void			/* resets terminal state to */
 c_system_terminal (void)		/* as it was before */
 {
-  int fd;
+    int fd;
 
-  if (! isatty (STDIN_FILENO)) return;
+    if (! isatty (STDIN_FILENO)) return;
 
-  fflush (stdout);
-  tty_restore ();
-  t_puts (keypad_local, 0);
-  for (fd = 0; fd < 3; fd++)
-    fcntl (fd, F_SETFL, saved_fcntl[fd]);
+    fflush (stdout);
+    tty_restore ();
+    t_puts (keypad_local, 0);
+    for (fd = 0; fd < 3; fd++)
+        P4_fcntl_F_SETFL (fd, saved_fcntl[fd]); /* fcntl (fd, F_SETFL, ..) */
 }
 
 /*
