@@ -5,8 +5,8 @@
  *
  *  @see     GNU LGPL
  *  @author  Guido Draheim            @(#) %derived_by: guidod %
- *  @version %version: 1.24 %
- *    (%date_modified: Mon Sep 10 18:04:33 2001 %)
+ *  @version %version: bln_mpt1!1.28 %
+ *    (%date_modified: Tue Aug 13 15:38:46 2002 %)
  *
  *  @description
  *              	Terminal driver for win32 wincon.h API as
@@ -19,7 +19,7 @@
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  term-wincon.c~1.24:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  term-wincon.c~bln_mpt1!1.28:csrc:bln_12xx!1 % $";
 #endif
 
 #ifndef _REENTRANT
@@ -427,6 +427,7 @@ c_putc_noflush (char c)
 	 * scroll the buffer up. 
 	 */
 
+#if 0 /* FIXME: !! win32-window erased multiple times... */
 	if ((screenInfo.dwSize.Y-1) == screenInfo.dwCursorPosition.Y) 
 	{ 
 	    /* ScrollScreenBuffer(pfeTerm->hStdout, 1); */
@@ -450,12 +451,13 @@ c_putc_noflush (char c)
 	/* Otherwise, advance the cursor to the next line. */
 
 	else screenInfo.dwCursorPosition.Y += 1; 
- 
+
 	if (! SetConsoleCursorPosition(
 	    pfeTerm->hStdout, 
 	    screenInfo.dwCursorPosition))
 	    goto failed;
 	return;
+#endif
     failed:
         if (pfeTerm->isNoConsole)
             WriteFile (pfeTerm->hStdout, &c, 1, &ignore, 0);
@@ -583,7 +585,11 @@ c_clrscr (void)
     CONSOLE_SCREEN_BUFFER_INFO screenInfo;
     if (! GetConsoleScreenBufferInfo(pfeTerm->hStdout, &screenInfo)) 
 	return;
-    c_gotoxy (0, 0);
+    /* note: we blank the screen buffer first, and move the cursor to 0,0
+     * just then. Thus we avoid screen flicker. (compare with c_gotoxy)
+     */
+    screenInfo.dwCursorPosition.X = 0; 
+    screenInfo.dwCursorPosition.Y = 0; 
     FillConsoleOutputCharacter (
 	pfeTerm->hStdout,
 	' ',
@@ -596,6 +602,9 @@ c_clrscr (void)
 	screenInfo.dwSize.X*screenInfo.dwSize.Y, 
 	screenInfo.dwCursorPosition,
 	&ignore);
+    SetConsoleCursorPosition(
+	pfeTerm->hStdout, 
+	screenInfo.dwCursorPosition);
 }
 
 static void
@@ -733,7 +742,7 @@ p4_term_struct p4_term_wincon =
 /* NOT static - export to debugger task list */
 void* p4_term_wincon_k12_thread (p4_threadP p4)
 {
-#  ifdef p4TH
+#  ifdef P4_REGTH
     p4TH = p4;
 #  endif
     {
