@@ -5,8 +5,8 @@
  *
  *  @see     GNU LGPL
  *  @author  Tektronix CTE              @(#) %derived_by: guidod %
- *  @version %version: bln_mpt1!1.34 %
- *    (%date_modified: Tue Oct 02 17:11:17 2001 %)
+ *  @version %version: bln_mpt1!1.58 %
+ *    (%date_modified: Tue Feb 25 13:22:45 2003 %)
  *
  *  @description
  *              This wordset adds some additional primitives that
@@ -18,7 +18,7 @@
 /*@{*/
 #if defined(__version_control__) && defined(__GNUC__)
 static char* id __attribute__((unused)) = 
-"@(#) $Id: %full_filespec:  useful-ext.c~bln_mpt1!1.34:csrc:bln_12xx!1 % $";
+"@(#) $Id: %full_filespec:  useful-ext.c~bln_mpt1!1.58:csrc:bln_12xx!1 % $";
 #endif
  
 #define _P4_SOURCE 1
@@ -30,6 +30,7 @@ static char* id __attribute__((unused)) =
 
 #include <pfe/dict-sub.h>
 #include <pfe/logging.h>
+#include <pfe/def-restore.h>
 
 /** 
  * see => >COMPILE  and => POSTPONE
@@ -59,7 +60,7 @@ FCode (p4_to_compile)
 #define P4_PAREN_MAGIC P4_MAGIC_('P','(',')','X')
 
 extern FCode (p4_tick);
-/** ($ ( [word] -- cs-token ) compile-only
+/** "($" ( [word] -- cs-token ) compile-only
  *  takes the execution token of the following word and
  *  saves it on the compile-stack. The correspondig closing
  *  => ) will then feed it into => >COMPILE - so this pair
@@ -75,7 +76,7 @@ FCode (p4_prefix_begin)
     FX_PUSH (P4_PAREN_MAGIC);
 }
 
-/** ) ( cs-token -- )
+/** ")" ( cs-token -- )
  * takes the execution-token from => ($ and compiles
  * it using => >COMPILE
  */
@@ -85,7 +86,7 @@ FCode (p4_prefix_end)
     FX (p4_to_compile);
 }
 
-/** )) ( cs-token cs-token -- )
+/** "))" ( cs-token cs-token -- )
  * takes two execution-tokens from two of => ($ and compiles
  * them on after another using => >COMPILE
  simulate:
@@ -181,7 +182,7 @@ p4sprintf (char* s)
               argv[4], argv[5], argv[6], argv[7]);
         }else{
 #         ifdef P4_UPPER_REGS /* i960 */
-            _p4_thread_save();
+            P4_CALLER_MKSAVED
 #         endif
             printed = sprintf (s, formbuf, 
               argv[0], argv[1], argv[2], argv[3], 
@@ -190,7 +191,7 @@ p4sprintf (char* s)
               argv[12], argv[13], argv[14], argv[15]);
             
 #         ifdef P4_UPPER_REGS
-            _p4_thread_load(); 
+            P4_CALLER_RESTORE
 #         endif
             if (argn > 13)
             { P4_warn1 ("quite many args for a printf (%i)", argn); }
@@ -233,11 +234,11 @@ FCode (p4_printf)
 
 void
 p4_forget_loadf(void)
-{
-    char* dp = (char*) WP_PFA[0];
+{   FX_USE_BODY_ADDR {
+    char* dp = (char*) FX_POP_BODY_ADDR[0];
 
     p4_forget (dp);
-}
+}}
 
 /** LOADF ( "filename" -- )
  *  loads a file just like => INCLUDE but does also put
@@ -273,7 +274,7 @@ p4_loadf_locate(p4xt xt)
         while (p)
         {
             p4xt cfa = p4_name_from(p);
-            if (*cfa == p4_forget_loadf
+            if (*P4_TO_CODE(cfa) == p4_forget_loadf
               &&  cfa > xt && xt > *(p4xt*)P4_TO_BODY(cfa)) 
                 return p;
             
@@ -283,7 +284,7 @@ p4_loadf_locate(p4xt xt)
     return 0;
 }
 
-/** (LOADF-LOCATE) ( xt -- nfa )
+/** "(LOADF-LOCATE)" ( xt -- nfa )
  * the implementation of => LOADF-LOCATE
  */
 FCode(p4_paren_loadf_locate)
@@ -329,22 +330,32 @@ FCode(p4_loadf_locate)
 */
 /**@{*/
 
-/** (;AND)
+/** "(;AND)" ( -- )
  * compiled by => ;AND
  */
-FCode (p4_semicolon_and_execution)
+FCode_XE (p4_semicolon_and_execution)
 {
+    FX_USE_CODE_ADDR;
     FX (p4_semicolon_execution);
     /* cannot use it in P4COMPILES directly, since it would prevent
        decompiler from acting on at that place 
     */
+    FX_USE_CODE_EXIT;
 }
 
+#if !defined PFE_SBR_CALL_THREADING
+#define FX_COMPILE1_p4_semicolon_and FX_COMPILE(p4_semicolon_and)
+#define FX_COMPILE2_p4_semicolon_and FX_COMPILE(p4_semicolon_and)
+#else
+#define FX_COMPILE1_p4_semicolon_and FX_COMPILE1_p4_semicolon
+#define FX_COMPILE2_p4_semicolon_and FX_COMPILE2_p4_semicolon
+#endif
+
 /* the 3 in MAKE-begin style is invariant, see debug.c */
 #define P4_MAKE0_STYLE 3, 1, 0,  0, 4  /* almost like IF */
 #define P4_MAKE1_STYLE 1, 0, -4, 1, 0  /* almost like THEN */
 
-/** ;AND ( -- )
+/** ";AND" ( -- )
  * For the code piece between => MAKE and => ;AND , this word
  * will do just an => EXIT . For the code outside of
  * the => MAKE construct a branch-around must be resolved then.
@@ -361,10 +372,10 @@ FCode (p4_semicolon_and)
 
     if (PFE.locals)
     {
-        FX_COMPILE (p4_semicolon_and); /* FX_COMPILE2 (p4_semicolon_and); */
+        FX_COMPILE2_p4_semicolon_and;
         PFE.locals = NULL;
     }else{
-        FX_COMPILE (p4_semicolon_and); /* FX_COMPILE1 (p4_semicolon_and); */
+        FX_COMPILE1_p4_semicolon_and;
     }
 
     /*
@@ -382,25 +393,29 @@ FCode (p4_semicolon_and)
 P4COMPILES (p4_semicolon_and, p4_semicolon_and_execution,
         	P4_SKIPS_NOTHING, P4_MAKE1_STYLE);
 
-/** ((MAKE-))
+/** "((MAKE-))" ( -- )
  * compiled by => MAKE
  */
-FCode (p4_make_to_local_execution)
+FCode_XE (p4_make_to_local_execution)
 {
+    FX_USE_CODE_ADDR;
     FX_PUSH (IP+2);             /* push following colon-RT, ie. CFA */
     FX (p4_to_local_execution); /* let TO put it into local */
     FX_BRANCH;                  /* and branch over */
+    FX_USE_CODE_EXIT;
 }
 
-/** ((MAKE))
+/** "((MAKE))" ( -- )
  * compiled by => MAKE
  */
-FCode (p4_make_to_execution)
+FCode_XE (p4_make_to_execution)
 {
     extern FCode(p4_is_execution);
+    FX_USE_CODE_ADDR;
     FX_PUSH (IP+2);             /* push following colon-RT, ie. CFA */
     FX (p4_is_execution);       /* let IS put it into defer */
     FX_BRANCH;                  /* and branch over */
+    FX_USE_CODE_EXIT;
 }
 
 /** DOER ( word -- )
@@ -410,6 +425,7 @@ FCode (p4_make_to_execution)
  * should be replaced with =>"DEFER" =>"IS", which is easy since
  * the =>"DEFER" and =>"DOER" point to the same internal runtime.
  */
+extern void FXCode (p4_defer);
 
 /** MAKE ( [word] -- ) ... ;AND
  * make a seperated piece of code between => MAKE and => ;AND 
@@ -457,30 +473,99 @@ P4COMPILES2(p4_make, p4_make_to_execution, p4_make_to_local_execution,
 
 /**@}*/
 
-/** OFFSET: ( offset "name" -- )
- *  an older word for =>"FIELD-OFFSET", please use =>"FIELD-OFFSET"
- *  or =>"FIELD".
+/** OFFSET-RT ( value -- value+offset )
+ *  this runtime will add the body-value to the value at top-of-stack.
+ *  used heavily in structure access words, compiled by => /FIELD
  */
-FCode (p4_offset_RT)
+FCode_RT (p4_offset_RT)
 {
-    *SP += WP_PFA[0];
+    FX_USE_BODY_ADDR;
+    *SP += FX_POP_BODY_ADDR[0];
 }
 
-/** FIELD-OFFSET ( offset "name" -- )
+/** +CONSTANT ( offset "name" -- )
  * create a new offsetword. The word is created and upon execution
- * it add the offset, ie. compiling runtime:
+ * it adds the offset, ie. compiling the => OFFSET-RT runtime:
        ( address -- address+offset )
+ * This word is just a convenience word, just use the word => +FIELD 
+ * directly and choose a => DROP to flag the end of a current
+ * offset-field declaration series. See also => /FIELD series to
+ * declare simple structures which end with a final => CONSTANT to
+ * memorize the complete size. The => /FIELD style is more traditional.
  */
-FCode (p4_field_offset)
+FCode (p4_offset_constant)
 {
-    extern FCode(p4_offset_RT);
-
-    FX_HEADER;
-    FX_RUNTIME1 (p4_field_offset);
+    FX_RUNTIME_HEADER;
+    FX_RUNTIME1 (p4_offset_constant);
     FX_UCOMMA (*SP); FX_DROP;
 }
-P4RUNTIME1(p4_field_offset, p4_offset_RT);
+P4RUNTIME1(p4_offset_constant, p4_offset_RT);
 
+/** +FIELD ( offset "name" -- offset )
+ * created a new name with an => OFFSET-RT runtime using the given offset. 
+ * Leave the offset-value untouched, so it can be modified with words
+ * like => CHAR+ and => CELL+ and => SFLOAT+ ; This word is the simplest way 
+ * to declared structure access words in forth - the two => STRUCT modules 
+ * contain a more elaborate series of words. Use this one like:
+ 0                        ( a fresh definition is started )
+ +FIELD zapp.a+ CHAR+     ( zero offset from the base of the struct )
+ +FIELD zapp.b+ CELL+     ( no alignment, starts off at 1 from base )
+ +FIELD zapp+   DROP      ( store size of complete zap structure )
+
+ 0 zapp+                  ( extend the zap structure )
+ +FIELD zappx.c+ CELL+    ( a new field )
+ +FIELD zappx+   DROP     ( and save it again )
+
+ CREATE zapp1  0 zapp+ ALLOT ( a way to allocate a strucutre )
+
+ zapp2 zapp.b+ @         ( read a value from the field )
+ 16 zapp2 zapp.b+ !      ( store a value in there )
+
+ * this form is not the traditional form used in forth, it is however
+ * quite simple. Use the simplefield declaration with => /FIELD to
+ * be compatible with traditional styles that build on top of sizeof
+ * constants in forth (which are not part of the ANS Forth standard).
+ */
+FCode (p4_plus_field)
+{
+    FX_RUNTIME_HEADER;
+    FX_RUNTIME1 (p4_offset_constant);
+    FX_UCOMMA (SP[0]);
+}
+
+/** /FIELD ( offset size "name" -- offset+size )
+ * created a new => +FIELD name with an => OFFSET-RT
+ * of offset. Then add the size value to the offset so that
+ * the next => /FIELD declaration will start at the end of the
+ * field currently declared. This word is the simplest way to
+ * declared structure access words in forth - the two => STRUCT modules 
+ * contain a more elaborate series of words. This one is used like:
+ 0                        ( a fresh definition is started )
+ /CHAR /FIELD ->zapp.a    ( zero offset from the base of the struct )
+ /CELL /FIELD ->zapp.b    ( no alignment, starts off at 1 from base )
+ CONSTANT /zapp           ( store size of complete zap structure )
+
+ /zapp                    ( extend the zap structure )
+ /CELL /FIELD ->zappx.c   ( a new field )
+ CONSTANT /zappx          ( and save it again )
+
+ CREATE zapp1 /zapp ALLOT ( a way to allocate a strucutre )
+ /zapp BUFFER: zapp2      ( another way to do it, semi-standard )
+
+ zapp2 ->zapp.b @         ( read a value from the field )
+ 16 zapp2 ->zapp.b !      ( store a value in there )
+
+ * compare also with => /CHAR => /WCHAR => /CELL => /DCELL
+ * and use => +FIELD as the lowlevel word, can simulate as
+ : /FIELD SWAP +FIELD + ;
+ */
+FCode (p4_slash_field)
+{
+    FX_RUNTIME_HEADER;
+    FX_RUNTIME1 (p4_offset_constant);
+    FX_UCOMMA (SP[1]);
+    SP[1] += SP[0]; FX_DROP;
+}
 
 
 /** [NOT] ( a -- a' )
@@ -576,49 +661,6 @@ FCode(p4_replace_in)
     }
 }
 
-/* ----------------------------------------------------------------- */
-    
-extern FCode(p4_defer); /* -> DOER */
-
-/** ALIAS ( xt "name" -- )
- * create a defer word that is initialized with the given x-token.
- */
-FCode (p4_alias)
-{
-    FX_HEADER;
-    FX_RUNTIME1 (p4_defer); /* fixme? p4_alias_RT */
-    FX_XCOMMA (0); /* DOES-CODE field (later may be used for chain link)*/
-    FX_XCOMMA (FX_POP); /* set DOES-BODY here */
-}
-
-/** ALIAS-ATEXIT ( xt "name" -- )
- * create a defer word that is initialized with the given x-token.
- */
-FCode (p4_alias_atexit)
-{
-    FX_HEADER_(PFE.atexit_wl); /* <-- the difference with => ALIAS */
-    FX_RUNTIME1 (p4_defer); /* fixme? p4_alias_atexit_RT */
-    FX_XCOMMA (0); 
-    FX_XCOMMA (FX_POP);
-
-#ifdef PFE_WITH_FFA
-    *_FFA(LAST) |= P4xONxDESTROY; /* fixme: p4_alias_atexit_RT !! */
-#endif
-}
-
-/** [VOCABULARY] ( "name" -- )
- * create an immediate vocabulary. Provides for basic 
- * modularization.
- : [VOCABULARY] VOCABULARY IMMEDIATE ;
- */
-FCode (p4_bracket_vocabulary)
-{
-    extern FCode (p4_vocabulary);
-
-    FX (p4_vocabulary);
-    *_FFA(LAST) |= P4xIMMEDIATE;
-}
-
 /* ------------------------------------------------------------------- 
  * hex string
  */
@@ -683,6 +725,19 @@ P4COMPILES (p4_x_quote, p4_c_quote_execution,
 
 /* ------------------------------------------------------------------- */
 
+/** [VOCABULARY] ( "name" -- )
+ * create an immediate vocabulary. Provides for basic 
+ * modularization.
+ : [VOCABULARY] VOCABULARY IMMEDIATE ;
+ */
+FCode (p4_bracket_vocabulary)
+{
+    extern FCode (p4_vocabulary);
+
+    FX (p4_vocabulary);
+    *_FFA(LAST) |= P4xIMMEDIATE;
+}
+
 /** [POSSIBLY] ( [name] -- ?? )
  * check if the name exists, and execute it immediatly
  * if found. Derived from POSSIBLY as seen in other forth systems.
@@ -744,7 +799,7 @@ FCode (p4_context_Q)
 
 /** DEFS-ARE-CASE-SENSITIVE ( -- ) 
  * accesses => CURRENT which is generally the last wordlist that the
- * => DEFINTIONS shall go in. sets there a flag in the vocabulary-definition
+ * => DEFINITIONS shall go in. sets there a flag in the vocabulary-definition
  * so that words are matched case-sensitive. 
  example: 
     VOCABULARY MY-VOC  MY-VOC DEFINITIONS DEFS-ARE-CASE-SENSITIVE
@@ -793,7 +848,7 @@ FCode (p4_defs_are_searched_also)
  * OBSOLETE!! use DEFS-ARE-SEARCHED-ALSO
  */
 
-/* [EXECUTE] ( [word] -- )
+/** [EXECUTE] ( [word] -- )
  * ticks the following word, and executes it - even in compiling mode.
  : [EXECUTE] ' EXECUTE ;
  */
@@ -802,16 +857,15 @@ FCode (p4_bracket_execute)
     p4_call (p4_tick_cfa(FX_VOID));
 }
 
-/* !NO ( -- false )
+/** !NO ( -- false )
  * a synonym for => FALSE
  !NO SMART-WORDS!
  */
 
-/* !USE ( -- false )
+/** !USE ( -- true )
  * a synonym for => TRUE
  !USE SMART-WORDS!
  */
-   
 
 P4_LISTWORDS (useful) =
 {
@@ -828,8 +882,15 @@ P4_LISTWORDS (useful) =
     P4_SXco ("MAKE",			p4_make),
     P4_SXco (";AND",			p4_semicolon_and),
     P4_IXco ("[NOT]",			p4_bracket_not),
-    P4_RTco ("FIELD-OFFSET",		p4_field_offset),  
-    P4_xOLD ("OFFSET:",			"FIELD-OFFSET"),  
+
+    P4_RTco ("+CONSTANT",		p4_offset_constant),  
+    P4_FNYM ("FIELD-OFFSET",		"+CONSTANT"),  
+    P4_FNYM ("OFFSET:",			"+CONSTANT"),  
+    P4_FXco ("+FIELD",                  p4_plus_field),
+    P4_FXco ("/FIELD",                  p4_slash_field),
+    P4_OCoN ("/CHAR",                   sizeof(p4char)),
+    P4_OCoN ("/WCHAR",                  sizeof(short)),
+
     P4_FXco ("REPLACE-IN",		p4_replace_in),
     P4_FXco ("(LOADF-LOCATE)",		p4_paren_loadf_locate),
     P4_FXco ("LOADF-LOCATE",		p4_loadf_locate),
@@ -839,8 +900,6 @@ P4_LISTWORDS (useful) =
 # ifdef WITH_NO_FFA
     P4_OCoN ("#WITH-NO-FFA",		WITH_NO_FFA+100),
 # endif
-    P4_FXco ("ALIAS-ATEXIT",		p4_alias_atexit),
-    P4_FXco ("ALIAS",			p4_alias),
     P4_SXco ("X\"",			p4_x_quote),
     P4_IXco ("[POSSIBLY]",		p4_bracket_possibly),
     P4_FXco ("[VOCABULARY]",		p4_bracket_vocabulary),
@@ -854,6 +913,7 @@ P4_LISTWORDS (useful) =
     P4_xOLD ("SEARCH-ALSO-VOC",         "DEFS-ARE-SEARCHED-ALSO"),
     P4_FNYM ("!NO",			"FALSE"),
     P4_FNYM ("!USE",                    "TRUE"),
+    P4_SNYM ("|(",                      "("),
 };
 P4_COUNTWORDS (useful, "Useful kernel extensions");
 
